/* 
/ GGRAPH_aux.c
/
/ RAW image helpers
/
/ version 1.0, 2010 July 20
/
/ Author: Sandro Furieri a.furieri@lqt.it
/
/ Copyright (C) 2009  Alessandro Furieri
/
/    This program is free software: you can redistribute it and/or modify
/    it under the terms of the GNU Lesser General Public License as published by
/    the Free Software Foundation, either version 3 of the License, or
/    (at your option) any later version.
/
/    This program is distributed in the hope that it will be useful,
/    but WITHOUT ANY WARRANTY; without even the implied warranty of
/    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/    GNU Lesser General Public License for more details.
/
/    You should have received a copy of the GNU Lesser General Public License
/    along with this program.  If not, see <http://www.gnu.org/licenses/>.
/
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <math.h>

#ifdef _WIN32
#include <io.h>
#else
#include <unistd.h>
#endif

#include <geotiff.h>
#include <xtiffio.h>
#include <geo_normalize.h>
#include <geovalues.h>
#include <tiffio.h>

#include "gaiagraphics.h"
#include "gaiagraphics_internals.h"

GGRAPH_DECLARE int
gGraphImageInfosFromMemBuf (const void *mem_buf, int mem_buf_size,
			    int image_type, const void **infos_handle)
{
/* retrieving Image infos from a memory block containing an encoded image */
    gGraphImageInfosPtr infos = NULL;
    int ret;

    *infos_handle = NULL;
    switch (image_type)
      {
      case GGRAPH_IMAGE_GIF:
	  ret =
	      gg_image_infos_from_gif (mem_buf_size, mem_buf,
				       GG_TARGET_IS_MEMORY, &infos);
	  break;
      case GGRAPH_IMAGE_PNG:
	  ret =
	      gg_image_infos_from_png (mem_buf_size, mem_buf,
				       GG_TARGET_IS_MEMORY, &infos);
	  break;
      case GGRAPH_IMAGE_JPEG:
	  ret =
	      gg_image_infos_from_jpeg (mem_buf_size, mem_buf,
					GG_TARGET_IS_MEMORY, &infos);
	  break;
      case GGRAPH_IMAGE_TIFF:
	  ret =
	      gg_image_infos_from_tiff (mem_buf_size, mem_buf,
					GG_TARGET_IS_MEMORY, &infos);
	  break;
      case GGRAPH_IMAGE_GEOTIFF:
	  ret =
	      gg_image_infos_from_geo_tiff (mem_buf_size, mem_buf,
					    GG_TARGET_IS_MEMORY, &infos);
	  break;
      }
    if (ret != GGRAPH_OK)
	return ret;

    *infos_handle = infos;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageInfosFromFile (const char *path, int image_type,
			  const void **infos_handle)
{
/* retrieving Image infos from file */
    FILE *in = NULL;
    gGraphImageInfosPtr infos = NULL;
    int ret;

    *infos_handle = NULL;

/* attempting to open the image file */
    in = fopen (path, "rb");
    if (in == NULL)
	return GGRAPH_FILE_OPEN_ERROR;
    switch (image_type)
      {
      case GGRAPH_IMAGE_GIF:
	  ret = gg_image_infos_from_gif (0, in, GG_TARGET_IS_FILE, &infos);
	  break;
      case GGRAPH_IMAGE_PNG:
	  ret = gg_image_infos_from_png (0, in, GG_TARGET_IS_FILE, &infos);
	  break;
      case GGRAPH_IMAGE_JPEG:
	  ret = gg_image_infos_from_jpeg (0, in, GG_TARGET_IS_FILE, &infos);
	  break;
      case GGRAPH_IMAGE_TIFF:
	  ret = gg_image_infos_from_tiff (0, in, GG_TARGET_IS_FILE, &infos);
	  break;
      case GGRAPH_IMAGE_GEOTIFF:
	  ret = gg_image_infos_from_geo_tiff (0, in, GG_TARGET_IS_FILE, &infos);
	  break;
      };
    fclose (in);
    if (ret != GGRAPH_OK)
	return ret;

    *infos_handle = infos;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageFromMemBuf (const void *mem_buf, int mem_buf_size, int image_type,
		       const void **image_handle, int scale)
{
/* decompressing a memory block containing an encoded image */
    gGraphImagePtr img = NULL;
    int ret;

    *image_handle = NULL;
    switch (image_type)
      {
      case GGRAPH_IMAGE_GIF:
	  ret =
	      gg_image_from_gif (mem_buf_size, mem_buf, GG_TARGET_IS_MEMORY,
				 &img);
	  break;
      case GGRAPH_IMAGE_PNG:
	  ret =
	      gg_image_from_png (mem_buf_size, mem_buf, GG_TARGET_IS_MEMORY,
				 &img, scale);
	  break;
      case GGRAPH_IMAGE_JPEG:
	  ret =
	      gg_image_from_jpeg (mem_buf_size, mem_buf, GG_TARGET_IS_MEMORY,
				  &img, scale);
	  break;
      case GGRAPH_IMAGE_TIFF:
	  ret =
	      gg_image_from_tiff (mem_buf_size, mem_buf, GG_TARGET_IS_MEMORY,
				  &img);
	  break;
      case GGRAPH_IMAGE_GEOTIFF:
	  ret =
	      gg_image_from_geo_tiff (mem_buf_size, mem_buf,
				      GG_TARGET_IS_MEMORY, &img);
	  break;
      }
    if (ret != GGRAPH_OK)
	return ret;

    *image_handle = img;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageFromFile (const char *path, int image_type,
		     const void **image_handle, int scale)
{
/* reading an image from file */
    FILE *in = NULL;
    gGraphImagePtr img = NULL;
    int ret;

    *image_handle = NULL;

/* attempting to open the image file */
    in = fopen (path, "rb");
    if (in == NULL)
	return GGRAPH_FILE_OPEN_ERROR;
    switch (image_type)
      {
      case GGRAPH_IMAGE_GIF:
	  ret = gg_image_from_gif (0, in, GG_TARGET_IS_FILE, &img);
	  break;
      case GGRAPH_IMAGE_PNG:
	  ret = gg_image_from_png (0, in, GG_TARGET_IS_FILE, &img, scale);
	  break;
      case GGRAPH_IMAGE_JPEG:
	  ret = gg_image_from_jpeg (0, in, GG_TARGET_IS_FILE, &img, scale);
	  break;
      case GGRAPH_IMAGE_TIFF:
	  ret = gg_image_from_tiff (0, in, GG_TARGET_IS_FILE, &img);
	  break;
      case GGRAPH_IMAGE_GEOTIFF:
	  ret = gg_image_from_geo_tiff (0, in, GG_TARGET_IS_FILE, &img);
	  break;
      };
    fclose (in);
    if (ret != GGRAPH_OK)
	return ret;

    *image_handle = img;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageToJpegFile (const void *ptr, const char *path, int quality)
{
/* exporting an image into a JPEG compressed file */
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    int ret;
    FILE *out = NULL;

    if (!img)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

/* opening the output image file */
    out = fopen (path, "wb");
    if (out == NULL)
	return GGRAPH_FILE_OPEN_ERROR;
/* compressing as JPEG */
    ret = gg_image_to_jpeg (img, NULL, NULL, out, GG_TARGET_IS_FILE, quality);
    fclose (out);
    if (ret != GGRAPH_OK)
      {
	  /* some unexpected error occurred */
	  unlink (path);
	  return ret;
      }

    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageToJpegMemBuf (const void *ptr, void **mem_buf, int *mem_buf_size,
			 int quality)
{
/* exporting an image into a JPEG compressed memory buffer */
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    void *buf = NULL;
    int size;
    int ret;

    *mem_buf = NULL;
    *mem_buf_size = 0;
    if (img == NULL)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

/* compressing as JPEG */
    ret =
	gg_image_to_jpeg (img, &buf, &size, NULL, GG_TARGET_IS_MEMORY, quality);
    if (ret != GGRAPH_OK)
	return ret;

/* exporting the memory buffer */
    *mem_buf = buf;
    *mem_buf_size = size;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageToPngFile (const void *ptr, const char *path, int compression_level,
		      int quantization_factor, int interlaced)
{
/* exporting an image into a PNG compressed file */
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    int ret;
    FILE *out = NULL;

    if (!img)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

/* opening the output image file */
    out = fopen (path, "wb");
    if (out == NULL)
	return GGRAPH_FILE_OPEN_ERROR;
/* compressing as PNG */
    ret =
	gg_image_to_png (img, NULL, NULL, out, GG_TARGET_IS_FILE,
			 compression_level, quantization_factor, interlaced);
    fclose (out);
    if (ret != GGRAPH_OK)
      {
	  /* some unexpected error occurred */
	  unlink (path);
	  return ret;
      }

    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageToPngMemBuf (const void *ptr, void **mem_buf, int *mem_buf_size,
			int compression_level, int quantization_factor,
			int interlaced)
{
/* exporting an image into a PNG compressed memory buffer */
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    void *buf = NULL;
    int size;
    int ret;

    *mem_buf = NULL;
    *mem_buf_size = 0;
    if (img == NULL)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

/* compressing as PNG */
    ret =
	gg_image_to_png (img, &buf, &size, NULL, GG_TARGET_IS_MEMORY,
			 compression_level, quantization_factor, interlaced);
    if (ret != GGRAPH_OK)
	return ret;

/* exporting the memory buffer */
    *mem_buf = buf;
    *mem_buf_size = size;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageToGifFile (const void *ptr, const char *path)
{
/* exporting an image into a GIF compressed file */
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    int ret;
    FILE *out = NULL;

    if (!img)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

/* opening the output image file */
    out = fopen (path, "wb");
    if (out == NULL)
	return GGRAPH_FILE_OPEN_ERROR;
/* compressing as GIF */
    ret = gg_image_to_gif (img, NULL, NULL, out, GG_TARGET_IS_FILE);
    fclose (out);
    if (ret != GGRAPH_OK)
      {
	  /* some unexpected error occurred */
	  unlink (path);
	  return ret;
      }

    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageToGifMemBuf (const void *ptr, void **mem_buf, int *mem_buf_size)
{
/* exporting an image into a GIF compressed memory buffer */
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    void *buf = NULL;
    int size;
    int ret;

    *mem_buf = NULL;
    *mem_buf_size = 0;
    if (img == NULL)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

/* compressing as GIF */
    ret = gg_image_to_gif (img, &buf, &size, NULL, GG_TARGET_IS_MEMORY);
    if (ret != GGRAPH_OK)
	return ret;

/* exporting the memory buffer */
    *mem_buf = buf;
    *mem_buf_size = size;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageToTiffFile (const void *ptr, const char *path, int tiff_layout,
		       int tile_width, int tile_height, int rows_per_strip,
		       int compression)
{
/* exporting an image into a TIFF formatted file */
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    int ret;
    FILE *out;
    int actual_layout = GGRAPH_TIFF_LAYOUT_TILES;
    int actual_tile_width = 256;
    int actual_tile_height = 256;
    int actual_rows_per_strip = 1;

    if (img == NULL)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

/* setting TIFF options */
    if (tiff_layout == GGRAPH_TIFF_LAYOUT_STRIPS)
      {
	  actual_layout = GGRAPH_TIFF_LAYOUT_STRIPS;
	  if (rows_per_strip != GGRAPH_TIFF_DEFAULT)
	      actual_rows_per_strip = rows_per_strip;
	  if (actual_rows_per_strip < 1)
	      actual_rows_per_strip = 1;
	  if (actual_rows_per_strip > 1024)
	      actual_rows_per_strip = 1024;
	  if (actual_rows_per_strip > img->height)
	      actual_rows_per_strip = img->height;
      }
    else
      {
	  if (tile_width != GGRAPH_TIFF_DEFAULT)
	      actual_tile_width = tile_width;
	  if (tile_height != GGRAPH_TIFF_DEFAULT)
	      actual_tile_height = tile_height;
	  if (actual_tile_width < 64)
	      actual_tile_width = 64;
	  if (actual_tile_height < 64)
	      actual_tile_height = 64;
	  if (actual_tile_width > 1024)
	      actual_tile_width = 1024;
	  if (actual_tile_height > 1024)
	      actual_tile_height = 1024;
      }

/* opening the output image file */
    out = fopen (path, "wb");
    if (out == NULL)
	return GGRAPH_FILE_OPEN_ERROR;
/* compressing as TIFF */
    if (actual_layout == GGRAPH_TIFF_LAYOUT_STRIPS)
      {
	  if (img->pixel_format == GG_PIXEL_GRID)
	      ret =
		  gg_image_to_tiff_grid_strip (img, NULL, NULL, out,
					       GG_TARGET_IS_FILE,
					       actual_rows_per_strip,
					       compression);
	  else if (gg_is_image_monochrome_ready (img) == GGRAPH_TRUE)
	      ret =
		  gg_image_to_tiff_fax_strip (img, NULL, NULL, out,
					      GG_TARGET_IS_FILE,
					      actual_rows_per_strip,
					      compression);
	  else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
	      ret =
		  gg_image_to_tiff_grayscale_strip (img, NULL, NULL, out,
						    GG_TARGET_IS_FILE,
						    actual_rows_per_strip,
						    compression);
	  else if (img->pixel_format == GG_PIXEL_PALETTE)
	      ret =
		  gg_image_to_tiff_palette_strip (img, NULL, NULL, out,
						  GG_TARGET_IS_FILE,
						  actual_rows_per_strip,
						  compression);
	  else
	      ret =
		  gg_image_to_tiff_rgb_strip (img, NULL, NULL, out,
					      GG_TARGET_IS_FILE,
					      actual_rows_per_strip,
					      compression);
      }
    else
      {
	  if (img->pixel_format == GG_PIXEL_GRID)
	      ret =
		  gg_image_to_tiff_grid_tile (img, NULL, NULL, out,
					      GG_TARGET_IS_FILE,
					      actual_tile_width,
					      actual_tile_height, compression);
	  else if (gg_is_image_monochrome_ready (img) == GGRAPH_TRUE)
	      ret =
		  gg_image_to_tiff_fax_tile (img, NULL, NULL, out,
					     GG_TARGET_IS_FILE,
					     actual_tile_width,
					     actual_tile_height, compression);
	  else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
	      ret =
		  gg_image_to_tiff_grayscale_tile (img, NULL, NULL, out,
						   GG_TARGET_IS_FILE,
						   actual_tile_width,
						   actual_tile_height,
						   compression);
	  else if (img->pixel_format == GG_PIXEL_PALETTE)
	      ret =
		  gg_image_to_tiff_palette_tile (img, NULL, NULL, out,
						 GG_TARGET_IS_FILE,
						 actual_tile_width,
						 actual_tile_height,
						 compression);
	  else
	      ret =
		  gg_image_to_tiff_rgb_tile (img, NULL, NULL, out,
					     GG_TARGET_IS_FILE,
					     actual_tile_width,
					     actual_tile_height, compression);
      }
    fclose (out);
    if (ret != GGRAPH_OK)
      {
	  /* some unexpected error occurred */
	  unlink (path);
	  return ret;
      }

    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageToTiffMemBuf (const void *ptr, void **mem_buf, int *mem_buf_size,
			 int tiff_layout, int tile_width, int tile_height,
			 int rows_per_strip, int compression)
{
/* exporting an image into a TIFF compressed memory buffer */
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    void *buf = NULL;
    int size;
    int ret;
    int actual_layout = GGRAPH_TIFF_LAYOUT_TILES;
    int actual_tile_width = 256;
    int actual_tile_height = 256;
    int actual_rows_per_strip = 1;

    *mem_buf = NULL;
    *mem_buf_size = 0;
    if (img == NULL)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

/* setting TIFF options */
    if (tiff_layout == GGRAPH_TIFF_LAYOUT_STRIPS)
      {
	  actual_layout = GGRAPH_TIFF_LAYOUT_STRIPS;
	  if (rows_per_strip != GGRAPH_TIFF_DEFAULT)
	      actual_rows_per_strip = rows_per_strip;
	  if (actual_rows_per_strip < 1)
	      actual_rows_per_strip = 1;
	  if (actual_rows_per_strip > 1024)
	      actual_rows_per_strip = 1024;
	  if (actual_rows_per_strip > img->height)
	      actual_rows_per_strip = img->height;
      }
    else
      {
	  if (tile_width != GGRAPH_TIFF_DEFAULT)
	      actual_tile_width = tile_width;
	  if (tile_height != GGRAPH_TIFF_DEFAULT)
	      actual_tile_height = tile_height;
	  if (actual_tile_width < 64)
	      actual_tile_width = 64;
	  if (actual_tile_height < 64)
	      actual_tile_height = 64;
	  if (actual_tile_width > 1024)
	      actual_tile_width = 1024;
	  if (actual_tile_height > 1024)
	      actual_tile_height = 1024;
      }

/* compressing as TIFF */
    if (actual_layout == GGRAPH_TIFF_LAYOUT_STRIPS)
      {
	  if (img->pixel_format == GG_PIXEL_GRID)
	      ret =
		  gg_image_to_tiff_grid_strip (img, &buf, &size, NULL,
					       GG_TARGET_IS_MEMORY,
					       actual_rows_per_strip,
					       compression);
	  else if (gg_is_image_monochrome_ready (img) == GGRAPH_TRUE)
	      ret =
		  gg_image_to_tiff_fax_strip (img, &buf, &size, NULL,
					      GG_TARGET_IS_MEMORY,
					      actual_rows_per_strip,
					      compression);
	  else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
	      ret =
		  gg_image_to_tiff_grayscale_strip (img, &buf, &size, NULL,
						    GG_TARGET_IS_MEMORY,
						    actual_rows_per_strip,
						    compression);
	  else if (img->pixel_format == GG_PIXEL_PALETTE)
	      ret =
		  gg_image_to_tiff_palette_strip (img, &buf, &size, NULL,
						  GG_TARGET_IS_MEMORY,
						  actual_rows_per_strip,
						  compression);
	  else
	      ret =
		  gg_image_to_tiff_rgb_strip (img, &buf, &size, NULL,
					      GG_TARGET_IS_MEMORY,
					      actual_rows_per_strip,
					      compression);
      }
    else
      {
	  if (img->pixel_format == GG_PIXEL_GRID)
	      ret =
		  gg_image_to_tiff_grid_tile (img, &buf, &size, NULL,
					      GG_TARGET_IS_MEMORY,
					      actual_tile_width,
					      actual_tile_height, compression);
	  else if (gg_is_image_monochrome_ready (img) == GGRAPH_TRUE)
	      ret =
		  gg_image_to_tiff_fax_tile (img, &buf, &size, NULL,
					     GG_TARGET_IS_MEMORY,
					     actual_tile_width,
					     actual_tile_height, compression);
	  else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
	      ret =
		  gg_image_to_tiff_grayscale_tile (img, &buf, &size, NULL,
						   GG_TARGET_IS_MEMORY,
						   actual_tile_width,
						   actual_tile_height,
						   compression);
	  else if (img->pixel_format == GG_PIXEL_PALETTE)
	      ret =
		  gg_image_to_tiff_palette_tile (img, &buf, &size, NULL,
						 GG_TARGET_IS_MEMORY,
						 actual_tile_width,
						 actual_tile_height,
						 compression);
	  else
	      ret =
		  gg_image_to_tiff_rgb_tile (img, &buf, &size, NULL,
					     GG_TARGET_IS_MEMORY,
					     actual_tile_width,
					     actual_tile_height, compression);
      }
    if (ret != GGRAPH_OK)
	return ret;

/* exporting the memory buffer */
    *mem_buf = buf;
    *mem_buf_size = size;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageToGeoTiffFile (const void *ptr, const char *path, int tiff_layout,
			  int tile_width, int tile_height, int rows_per_strip,
			  int compression)
{
/* exporting an image into a GeoTIFF formatted file */
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    int ret;
    FILE *out;
    int actual_layout = GGRAPH_TIFF_LAYOUT_TILES;
    int actual_tile_width = 256;
    int actual_tile_height = 256;
    int actual_rows_per_strip = 1;

    if (img == NULL)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

/* setting TIFF options */
    if (tiff_layout == GGRAPH_TIFF_LAYOUT_STRIPS)
      {
	  actual_layout = GGRAPH_TIFF_LAYOUT_STRIPS;
	  if (rows_per_strip != GGRAPH_TIFF_DEFAULT)
	      actual_rows_per_strip = rows_per_strip;
	  if (actual_rows_per_strip < 1)
	      actual_rows_per_strip = 1;
	  if (actual_rows_per_strip > 1024)
	      actual_rows_per_strip = 1024;
	  if (actual_rows_per_strip > img->height)
	      actual_rows_per_strip = img->height;
      }
    else
      {
	  if (tile_width != GGRAPH_TIFF_DEFAULT)
	      actual_tile_width = tile_width;
	  if (tile_height != GGRAPH_TIFF_DEFAULT)
	      actual_tile_height = tile_height;
	  if (actual_tile_width < 64)
	      actual_tile_width = 64;
	  if (actual_tile_height < 64)
	      actual_tile_height = 64;
	  if (actual_tile_width > 1024)
	      actual_tile_width = 1024;
	  if (actual_tile_height > 1024)
	      actual_tile_height = 1024;
      }

/* opening the output image file */
    out = fopen (path, "wb");
    if (out == NULL)
	return GGRAPH_FILE_OPEN_ERROR;
/* compressing as TIFF */
    if (actual_layout == GGRAPH_TIFF_LAYOUT_STRIPS)
      {
	  if (img->pixel_format == GG_PIXEL_GRID)
	      ret =
		  gg_image_to_geo_tiff_grid_strip (img, NULL, NULL, out,
						   GG_TARGET_IS_FILE,
						   actual_rows_per_strip,
						   compression);
	  else if (gg_is_image_monochrome_ready (img) == GGRAPH_TRUE)
	      ret =
		  gg_image_to_geo_tiff_fax_strip (img, NULL, NULL, out,
						  GG_TARGET_IS_FILE,
						  actual_rows_per_strip,
						  compression);
	  else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
	      ret =
		  gg_image_to_geo_tiff_grayscale_strip (img, NULL, NULL, out,
							GG_TARGET_IS_FILE,
							actual_rows_per_strip,
							compression);
	  else if (img->pixel_format == GG_PIXEL_PALETTE)
	      ret =
		  gg_image_to_geo_tiff_palette_strip (img, NULL, NULL, out,
						      GG_TARGET_IS_FILE,
						      actual_rows_per_strip,
						      compression);
	  else
	      ret =
		  gg_image_to_geo_tiff_rgb_strip (img, NULL, NULL, out,
						  GG_TARGET_IS_FILE,
						  actual_rows_per_strip,
						  compression);
      }
    else
      {
	  if (img->pixel_format == GG_PIXEL_GRID)
	      ret =
		  gg_image_to_geo_tiff_grid_tile (img, NULL, NULL, out,
						  GG_TARGET_IS_FILE,
						  actual_tile_width,
						  actual_tile_height,
						  compression);
	  else if (gg_is_image_monochrome_ready (img) == GGRAPH_TRUE)
	      ret =
		  gg_image_to_geo_tiff_fax_tile (img, NULL, NULL, out,
						 GG_TARGET_IS_FILE,
						 actual_tile_width,
						 actual_tile_height,
						 compression);
	  else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
	      ret =
		  gg_image_to_geo_tiff_grayscale_tile (img, NULL, NULL, out,
						       GG_TARGET_IS_FILE,
						       actual_tile_width,
						       actual_tile_height,
						       compression);
	  else if (img->pixel_format == GG_PIXEL_PALETTE)
	      ret =
		  gg_image_to_geo_tiff_palette_tile (img, NULL, NULL, out,
						     GG_TARGET_IS_FILE,
						     actual_tile_width,
						     actual_tile_height,
						     compression);
	  else
	      ret =
		  gg_image_to_geo_tiff_rgb_tile (img, NULL, NULL, out,
						 GG_TARGET_IS_FILE,
						 actual_tile_width,
						 actual_tile_height,
						 compression);
      }
    fclose (out);
    if (ret != GGRAPH_OK)
      {
	  /* some unexpected error occurred */
	  unlink (path);
	  return ret;
      }

    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageToGeoTiffMemBuf (const void *ptr, void **mem_buf, int *mem_buf_size,
			    int tiff_layout, int tile_width, int tile_height,
			    int rows_per_strip, int compression)
{
/* exporting an image into a GeoTIFF compressed memory buffer */
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    void *buf = NULL;
    int size;
    int ret;
    int actual_layout = GGRAPH_TIFF_LAYOUT_TILES;
    int actual_tile_width = 256;
    int actual_tile_height = 256;
    int actual_rows_per_strip = 1;

    *mem_buf = NULL;
    *mem_buf_size = 0;
    if (img == NULL)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

/* setting TIFF options */
    if (tiff_layout == GGRAPH_TIFF_LAYOUT_STRIPS)
      {
	  actual_layout = GGRAPH_TIFF_LAYOUT_STRIPS;
	  if (rows_per_strip != GGRAPH_TIFF_DEFAULT)
	      actual_rows_per_strip = rows_per_strip;
	  if (actual_rows_per_strip < 1)
	      actual_rows_per_strip = 1;
	  if (actual_rows_per_strip > 1024)
	      actual_rows_per_strip = 1024;
	  if (actual_rows_per_strip > img->height)
	      actual_rows_per_strip = img->height;
      }
    else
      {
	  if (tile_width != GGRAPH_TIFF_DEFAULT)
	      actual_tile_width = tile_width;
	  if (tile_height != GGRAPH_TIFF_DEFAULT)
	      actual_tile_height = tile_height;
	  if (actual_tile_width < 64)
	      actual_tile_width = 64;
	  if (actual_tile_height < 64)
	      actual_tile_height = 64;
	  if (actual_tile_width > 1024)
	      actual_tile_width = 1024;
	  if (actual_tile_height > 1024)
	      actual_tile_height = 1024;
      }

/* compressing as TIFF */
    if (actual_layout == GGRAPH_TIFF_LAYOUT_STRIPS)
      {
	  if (img->pixel_format == GG_PIXEL_GRID)
	      ret =
		  gg_image_to_geo_tiff_grid_strip (img, &buf, &size, NULL,
						   GG_TARGET_IS_MEMORY,
						   actual_rows_per_strip,
						   compression);
	  else if (gg_is_image_monochrome_ready (img) == GGRAPH_TRUE)
	      ret =
		  gg_image_to_geo_tiff_fax_strip (img, &buf, &size, NULL,
						  GG_TARGET_IS_MEMORY,
						  actual_rows_per_strip,
						  compression);
	  else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
	      ret =
		  gg_image_to_geo_tiff_grayscale_strip (img, &buf, &size, NULL,
							GG_TARGET_IS_MEMORY,
							actual_rows_per_strip,
							compression);
	  else if (img->pixel_format == GG_PIXEL_PALETTE)
	      ret =
		  gg_image_to_geo_tiff_palette_strip (img, &buf, &size, NULL,
						      GG_TARGET_IS_MEMORY,
						      actual_rows_per_strip,
						      compression);
	  else
	      ret =
		  gg_image_to_geo_tiff_rgb_strip (img, &buf, &size, NULL,
						  GG_TARGET_IS_MEMORY,
						  actual_rows_per_strip,
						  compression);
      }
    else
      {
	  if (img->pixel_format == GG_PIXEL_GRID)
	      ret =
		  gg_image_to_geo_tiff_grid_tile (img, &buf, &size, NULL,
						  GG_TARGET_IS_MEMORY,
						  actual_tile_width,
						  actual_tile_height,
						  compression);
	  else if (gg_is_image_monochrome_ready (img) == GGRAPH_TRUE)
	      ret =
		  gg_image_to_geo_tiff_fax_tile (img, &buf, &size, NULL,
						 GG_TARGET_IS_MEMORY,
						 actual_tile_width,
						 actual_tile_height,
						 compression);
	  else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
	      ret =
		  gg_image_to_geo_tiff_grayscale_tile (img, &buf, &size, NULL,
						       GG_TARGET_IS_MEMORY,
						       actual_tile_width,
						       actual_tile_height,
						       compression);
	  else if (img->pixel_format == GG_PIXEL_PALETTE)
	      ret =
		  gg_image_to_geo_tiff_palette_tile (img, &buf, &size, NULL,
						     GG_TARGET_IS_MEMORY,
						     actual_tile_width,
						     actual_tile_height,
						     compression);
	  else
	      ret =
		  gg_image_to_geo_tiff_rgb_tile (img, &buf, &size, NULL,
						 GG_TARGET_IS_MEMORY,
						 actual_tile_width,
						 actual_tile_height,
						 compression);
      }
    if (ret != GGRAPH_OK)
	return ret;

/* exporting the memory buffer */
    *mem_buf = buf;
    *mem_buf_size = size;
    return GGRAPH_OK;
}

GGRAPH_DECLARE void
gGraphDestroyImage (const void *ptr)
{
/* destroyng an image object */
    gGraphImagePtr img = (gGraphImagePtr) ptr;

    if (img == NULL)
	return;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return;
    gg_image_destroy (img);
}

GGRAPH_DECLARE void
gGraphDestroyImageInfos (const void *ptr)
{
/* destroyng an image INFOS object */
    gGraphImageInfosPtr img = (gGraphImageInfosPtr) ptr;

    if (img == NULL)
	return;
    if (img->signature != GG_IMAGE_INFOS_MAGIC_SIGNATURE)
	return;
    gg_image_infos_destroy (img);
}

GGRAPH_DECLARE int
gGraphGetImageInfos (const void *ptr, int *width, int *height, int *colorspace,
		     int *max_palette, int *bits_per_sample,
		     int *samples_per_pixel, int *sample_format,
		     int *tile_width, int *tile_height, int *rows_per_strip,
		     int *compression, int *scale_1_2, int *scale_1_4,
		     int *scale_1_8)
{
/* querying image INFOSt */
    gGraphImageInfosPtr infos = (gGraphImageInfosPtr) ptr;
    gGraphImagePtr img = (gGraphImagePtr) ptr;

    if (ptr == NULL)
	return GGRAPH_ERROR;
    if (infos->signature == GG_IMAGE_INFOS_MAGIC_SIGNATURE)
      {
	  /* form the INFO struct */
	  *width = infos->width;
	  *height = infos->height;
	  switch (infos->pixel_format)
	    {
	    case GG_PIXEL_RGB:
	    case GG_PIXEL_BGR:
	    case GG_PIXEL_GDI:
		*colorspace = GGRAPH_COLORSPACE_TRUECOLOR;
		break;
	    case GG_PIXEL_RGBA:
	    case GG_PIXEL_ARGB:
	    case GG_PIXEL_BGRA:
		*colorspace = GGRAPH_COLORSPACE_TRUECOLOR;
		break;
	    case GG_PIXEL_GRAYSCALE:
		*colorspace = GGRAPH_COLORSPACE_GRAYSCALE;
		break;
	    case GG_PIXEL_PALETTE:
		*colorspace = GGRAPH_COLORSPACE_PALETTE;
		break;
	    case GG_PIXEL_GRID:
		*colorspace = GGRAPH_COLORSPACE_GRID;
		break;
	    default:
		*colorspace = GGRAPH_COLORSPACE_UNKNOWN;
		break;
	    };
	  *bits_per_sample = infos->bits_per_sample;
	  *samples_per_pixel = infos->samples_per_pixel;
	  *sample_format = infos->sample_format;
	  *tile_width = infos->tile_width;
	  *tile_height = infos->tile_height;
	  *rows_per_strip = infos->rows_per_strip;
	  *compression = infos->compression;
	  *scale_1_2 = infos->scale_1_2;
	  *scale_1_4 = infos->scale_1_4;
	  *scale_1_8 = infos->scale_1_8;
	  return GGRAPH_OK;
      }
    else if (img->signature == GG_IMAGE_MAGIC_SIGNATURE)
      {
	  /* from the IMAGE struct */
	  *width = img->width;
	  *height = img->height;
	  switch (img->pixel_format)
	    {
	    case GG_PIXEL_RGB:
	    case GG_PIXEL_BGR:
	    case GG_PIXEL_GDI:
		*colorspace = GGRAPH_COLORSPACE_TRUECOLOR;
		break;
	    case GG_PIXEL_RGBA:
	    case GG_PIXEL_ARGB:
	    case GG_PIXEL_BGRA:
		*colorspace = GGRAPH_COLORSPACE_TRUECOLOR;
		break;
	    case GG_PIXEL_GRAYSCALE:
		*colorspace = GGRAPH_COLORSPACE_GRAYSCALE;
		break;
	    case GG_PIXEL_PALETTE:
		*colorspace = GGRAPH_COLORSPACE_PALETTE;
		break;
	    case GG_PIXEL_GRID:
		*colorspace = GGRAPH_COLORSPACE_GRID;
		break;
	    default:
		*colorspace = GGRAPH_COLORSPACE_UNKNOWN;
		break;
	    };
	  *max_palette = img->max_palette;
	  *bits_per_sample = img->bits_per_sample;
	  *samples_per_pixel = img->samples_per_pixel;
	  *sample_format = img->sample_format;
	  *tile_width = img->tile_width;
	  *tile_height = img->tile_height;
	  *rows_per_strip = img->rows_per_strip;
	  *compression = img->compression;
	  return GGRAPH_OK;
      }
    return GGRAPH_ERROR;
}

GGRAPH_DECLARE int
gGraphGetImageDims (const void *ptr, int *width, int *height)
{
/* querying image Width / Height */
    gGraphImagePtr img = (gGraphImagePtr) ptr;

    if (img == NULL)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;
    *width = img->width;
    *height = img->height;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphGetImageSize (const void *ptr, int *size)
{
/* querying image size */
    gGraphImagePtr img = (gGraphImagePtr) ptr;

    if (img == NULL)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

    *size = img->height * img->scanline_width;
    if (gg_is_image_monochrome_ready (img) == GGRAPH_TRUE)
	*size /= 8;

    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageColorSpaceOptimize (const void *ptr)
{
/* attempting to optimize the actual color space */
    int current_color_space;
    gGraphImagePtr img = (gGraphImagePtr) ptr;

    if (img == NULL)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

    if (img->pixel_format == GG_PIXEL_PALETTE)
      {
	  if (gg_is_image_monochrome_ready (img) == GGRAPH_TRUE)
	      current_color_space = GGRAPH_COLORSPACE_MONOCHROME;
	  else
	      current_color_space = GGRAPH_COLORSPACE_PALETTE;
      }
    else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
	current_color_space = GGRAPH_COLORSPACE_GRAYSCALE;
    else
	current_color_space = GGRAPH_COLORSPACE_TRUECOLOR;

    if (current_color_space == GGRAPH_COLORSPACE_MONOCHROME)
      {
	  /* already optimized */
	  return GGRAPH_OK;
      }
    if (gg_is_image_monochrome (img) == GGRAPH_TRUE)
      {
	  /* attempting to optimize into MONOCHROME */
	  return gg_convert_image_to_monochrome (img);
      }
    if (gg_is_image_grayscale (img) == GGRAPH_TRUE)
      {
	  /* attempting to optimize into GRAYSCALE */
	  return gg_convert_image_to_grayscale (img);
      }
    if (current_color_space == GGRAPH_COLORSPACE_PALETTE)
      {
	  /* already optimized */
	  return GGRAPH_OK;
      }
    if (gg_is_image_palette256 (img) == GGRAPH_TRUE)
      {
	  /* attempting to optimize into PALETTE */
	  return gg_convert_image_to_palette (img);
      }
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageResampleAsMonochrome (const void *ptr)
{
/* applying BILEVEL quantization */
    gGraphImagePtr img = (gGraphImagePtr) ptr;

    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

    return gg_convert_image_to_monochrome (img);
}

GGRAPH_DECLARE int
gGraphImageResampleAsPalette (const void *ptr, int num_colors)
{
/* applying quantization in order to fit into PALETTE colorspace */
    gGraphImagePtr img = (gGraphImagePtr) ptr;

    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

    return gg_image_resample_as_palette (img, num_colors);
}

GGRAPH_DECLARE int
gGraphImageResampleAsGrayscale (const void *ptr)
{
/* converting into GRAYSCALE */
    gGraphImagePtr img = (gGraphImagePtr) ptr;

    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

    return gg_convert_image_to_grayscale (img);
}

GGRAPH_DECLARE int
gGraphImageResampleAsPhotographic (const void *ptr)
{
/* converting into GRAYSCALE or RGB colorspace */
    gGraphImagePtr img = (gGraphImagePtr) ptr;

    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

    if (img->pixel_format == GG_PIXEL_GRAYSCALE
	|| img->pixel_format == GG_PIXEL_RGB)
	return GGRAPH_OK;

    if (gg_is_image_grayscale (img) == GGRAPH_TRUE)
      {
	  /* attempting to convert into GRAYSCALE */
	  return gg_convert_image_to_grayscale (img);
      }

    return gg_convert_image_to_rgb (img);
}

GGRAPH_DECLARE int
gGraphImageGuessFormat (const void *mem_buf, int mem_buf_size)
{
/* attempting to guess the Image type from its Magic Number */
    const unsigned char *p = mem_buf;
    if (mem_buf_size > 6)
      {
	  if (p[0] == 'G' && p[1] == 'I' && p[2] == 'F' && p[3] == '8'
	      && (p[4] == '7' || p[4] == '9') && p[5] == 'a')
	      return GGRAPH_IMAGE_GIF;
	  if (p[0] == 0x89 && p[1] == 'P' && p[2] == 'N' && p[3] == 'G'
	      && p[4] == 0x0d && p[5] == 0x0a)
	      return GGRAPH_IMAGE_PNG;
      }
    if (mem_buf_size > 4)
      {
	  if (p[0] == 'M' && p[1] == 'M' && p[2] == 0x00 && p[3] == '*')
	      return GGRAPH_IMAGE_TIFF;
	  if (p[0] == 'I' && p[1] == 'I' && p[2] == '*' && p[3] == 0x00)
	      return GGRAPH_IMAGE_TIFF;
      }
    if (mem_buf_size > 2)
      {
	  if (p[0] == 0xff && p[1] == 0xd8)
	      return GGRAPH_IMAGE_JPEG;
      }
    return GGRAPH_IMAGE_UNKNOWN;
}

GGRAPH_DECLARE int
gGraphFileImageGuessFormat (const char *path, int *type)
{
/* attempting to guess the FileImage type from its Magic Number */
    FILE *in = NULL;
    unsigned char mem_buf[10];
    int mem_buf_size;

    *type = GGRAPH_IMAGE_UNKNOWN;
/* attempting to open the image file */
    in = fopen (path, "rb");
    if (in == NULL)
	return GGRAPH_FILE_OPEN_ERROR;
    mem_buf_size = fread (mem_buf, 1, 10, in);
    if (mem_buf_size < 10)
      {
	  fclose (in);
	  return GGRAPH_FILE_READ_ERROR;
      }
    fclose (in);
    *type = gGraphImageGuessFormat (mem_buf, 10);
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageIsGeoReferenced (const void *ptr)
{
/* checking if an Image is georeferenced or not */
    gGraphImagePtr img = (gGraphImagePtr) ptr;

    if (img == NULL)
	return GGRAPH_ERROR;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_ERROR;

    if (img->is_georeferenced)
	return GGRAPH_OK;
    return GGRAPH_ERROR;
}

GGRAPH_DECLARE int
gGraphImageSetGeoReference (const void *ptr, int srid, const char *srs_name,
			    const char *proj4text, double upper_left_x,
			    double upper_left_y, double pixel_x_size,
			    double pixel_y_size)
{
/* setting the georeferencing infos for some Image */
    int len;
    char *SrsName = NULL;
    char *Proj4Text = NULL;
    gGraphImagePtr img = (gGraphImagePtr) ptr;

    if (img == NULL)
	return GGRAPH_IMAGE_UNKNOWN;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_IMAGE_UNKNOWN;

    if (srs_name)
      {
	  len = strlen (srs_name);
	  if (len > 0)
	    {
		SrsName = malloc (len + 1);
		if (!SrsName)
		    return GGRAPH_ERROR;
		strcpy (SrsName, srs_name);
	    }
      }
    if (proj4text)
      {
	  len = strlen (proj4text);
	  if (len > 0)
	    {
		Proj4Text = malloc (len + 1);
		if (!Proj4Text)
		  {
		      free (SrsName);
		      return GGRAPH_ERROR;
		  }
		strcpy (Proj4Text, proj4text);
	    }
      }

    if (img->srs_name)
	free (img->srs_name);
    if (img->proj4text)
	free (img->proj4text);

    img->is_georeferenced = 1;
    img->srid = srid;
    img->srs_name = SrsName;
    img->proj4text = Proj4Text;
    img->upper_left_x = upper_left_x;
    img->upper_left_y = upper_left_y;
    img->pixel_x_size = pixel_x_size;
    img->pixel_y_size = pixel_y_size;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageGetGeoReference (const void *ptr, int *srid, char *srs_name,
			    char *proj4text, double *upper_left_x,
			    double *upper_left_y, double *pixel_x_size,
			    double *pixel_y_size)
{
/* retrieving the georeferencing infos from some Image */
    gGraphImagePtr img = (gGraphImagePtr) ptr;

    if (img == NULL)
	return GGRAPH_IMAGE_UNKNOWN;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_IMAGE_UNKNOWN;
    if (img->is_georeferenced == 0)
	return GGRAPH_ERROR;

    *srid = img->srid;
    if (img->srs_name)
	strcpy (srs_name, img->srs_name);
    else
	*srs_name = '\0';
    if (img->proj4text)
	strcpy (proj4text, img->proj4text);
    else
	*proj4text = '\0';
    *upper_left_x = img->upper_left_x;
    *upper_left_y = img->upper_left_y;
    *pixel_x_size = img->pixel_x_size;
    *pixel_y_size = img->pixel_y_size;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageInfosGetGeoReference (const void *ptr, int *srid, char *srs_name,
				 char *proj4text, double *upper_left_x,
				 double *upper_left_y, double *pixel_x_size,
				 double *pixel_y_size)
{
/* retrieving the georeferencing infos from some Image */
    gGraphImageInfosPtr img = (gGraphImageInfosPtr) ptr;

    if (img == NULL)
	return GGRAPH_IMAGE_UNKNOWN;
    if (img->signature != GG_IMAGE_INFOS_MAGIC_SIGNATURE)
	return GGRAPH_IMAGE_UNKNOWN;
    if (img->is_georeferenced == 0)
	return GGRAPH_ERROR;

    *srid = img->srid;
    if (img->srs_name)
	strcpy (srs_name, img->srs_name);
    else
	*srs_name = '\0';
    if (img->proj4text)
	strcpy (proj4text, img->proj4text);
    else
	*proj4text = '\0';
    *upper_left_x = img->upper_left_x;
    *upper_left_y = img->upper_left_y;
    *pixel_x_size = img->pixel_x_size;
    *pixel_y_size = img->pixel_y_size;
    return GGRAPH_OK;
}

static int
token_to_double (const char *str, double *value)
{
/* attempting to convert from string to double */
    const char *p = str;
    int error = 0;
    int digits = 0;
    int comma = 0;
    int sign = 0;
    while (*p != '\0')
      {
	  if (*p >= '0' && *p <= '9')
	      digits++;
	  else if (*p == '.')
	      comma++;
	  else if (*p == '-' || *p == '+')
	    {
		if (digits == 0)
		    sign++;
		else
		    error = 1;
	    }
	  else
	      error = 1;
	  p++;
      }
    if (digits > 0 && sign < 2 && comma < 2)
      {
	  *value = atof (str);
	  return 1;
      }
    return 0;
}

GGRAPH_DECLARE int
gGraphGetWorldFilePath (const char *main_path, char *world_file_path)
{
/* generating a World File path corresponding to the Image path */
    char suffix[1024];
    char string[1024];
    int i;
    int len;

    strcpy (string, main_path);
    len = strlen (string) - 1;
    for (i = len - 1; i >= 0; i--)
      {
	  if (string[i] == '.')
	    {
		strcpy (suffix, string + i);
		string[i] = '\0';
		if (strcasecmp (suffix, ".jpg") == 0)
		  {
		      strcpy (world_file_path, string);
		      strcat (world_file_path, ".jgw");
		      return GGRAPH_OK;
		  }
		if (strcasecmp (suffix, ".png") == 0)
		  {
		      strcpy (world_file_path, string);
		      strcat (world_file_path, ".pgw");
		      return GGRAPH_OK;
		  }
		if (strcasecmp (suffix, ".gif") == 0)
		  {
		      strcpy (world_file_path, string);
		      strcat (world_file_path, ".gfw");
		      return GGRAPH_OK;
		  }
		if (strcasecmp (suffix, ".tif") == 0)
		  {
		      strcpy (world_file_path, string);
		      strcat (world_file_path, ".tfw");
		      return GGRAPH_OK;
		  }
	    }
      }
    *world_file_path = '\0';
    return GGRAPH_ERROR;
}

GGRAPH_DECLARE int
gGraphReadWorldFile (const char *path, double *ul_x, double *ul_y,
		     double *x_size, double *y_size)
{
/* attempting to read georeferencing infos from some World File */
    FILE *in = NULL;
    double upper_left_x = DBL_MAX;
    double upper_left_y = DBL_MAX;
    double pixel_x = DBL_MAX;
    double pixel_y = DBL_MAX;
    double value;
    int c;
    int i;
    char wf_buf[1024];
    char *p_wf;
    int wf_row;

    in = fopen (path, "rb");
    if (in == NULL)
	return GGRAPH_FILE_OPEN_ERROR;

    wf_row = 0;
    p_wf = wf_buf;
    while ((c = getc (in)) != EOF)
      {
	  /* parsing the Worldfile */
	  if (c == '\r')
	      continue;
	  if (c == '\n')
	    {
		*p_wf = '\0';
		for (i = 0; i < (int) strlen (wf_buf); i++)
		  {
		      /* replacing any COMMA (badly formatted decimal separator) */
		      if (wf_buf[i] == ',')
			  wf_buf[i] = '.';
		  }
		switch (wf_row)
		  {
		  case 0:
		      if (!token_to_double (wf_buf, &value))
			  goto stop;
		      else
			  pixel_x = value;
		      break;
		  case 3:
		      if (!token_to_double (wf_buf, &value))
			  goto stop;
		      else
			  pixel_y = value * -1.0;
		      break;
		  case 4:
		      if (!token_to_double (wf_buf, &value))
			  goto stop;
		      else
			  upper_left_x = value;
		      break;
		  case 5:
		      if (!token_to_double (wf_buf, &value))
			  goto stop;
		      else
			  upper_left_y = value;
		      break;
		  };
		p_wf = wf_buf;
		wf_row++;
		continue;
	    }
	  *p_wf++ = c;
	  if (p_wf - wf_buf > 1024)
	      goto stop;
      }
    fclose (in);

    if (pixel_x == DBL_MAX || pixel_y == DBL_MAX || upper_left_x == DBL_MAX
	|| upper_left_y == DBL_MAX)
	return GGRAPH_ERROR;

    *ul_x = upper_left_x;
    *ul_y = upper_left_y;
    *x_size = pixel_x;
    *y_size = pixel_y;
    return GGRAPH_OK;

  stop:
    fclose (in);
    return GGRAPH_ERROR;
}

GGRAPH_DECLARE int
gGraphImageResizeNormal (const void *orig, const void **dest, int width,
			 int height)
{
/* generating a resized image [Normal Quality] */
    gGraphImagePtr img2 = NULL;
    gGraphImagePtr img = (gGraphImagePtr) orig;

    *dest = NULL;
    if (img == NULL)
	return GGRAPH_ERROR;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_ERROR;

    img2 =
	gg_image_create (img->pixel_format, width, height, img->bits_per_sample,
			 img->samples_per_pixel, img->sample_format,
			 img->srs_name, img->proj4text);
    if (!img2)
	return GGRAPH_INSUFFICIENT_MEMORY;
    gg_image_resize (img2, img);
    gg_image_clone_georeferencing (img2, img);
    *dest = img2;

    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageResizeHighQuality (const void *orig, const void **dest, int width,
			      int height)
{
/* generating a resized image [High Quality] */
    gGraphImagePtr img2 = NULL;
    gGraphImagePtr img = (gGraphImagePtr) orig;

    *dest = NULL;
    if (img == NULL)
	return GGRAPH_ERROR;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_ERROR;

    img2 =
	gg_image_create (img->pixel_format, width, height, img->bits_per_sample,
			 img->samples_per_pixel, img->sample_format,
			 img->srs_name, img->proj4text);
    if (!img2)
	return GGRAPH_INSUFFICIENT_MEMORY;
    gg_make_thumbnail (img2, img);
    gg_image_clone_georeferencing (img2, img);
    *dest = img2;

    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageSubSet (const void *orig, const void **dest, int upper_left_x,
		   int upper_left_y, int width, int height)
{
/* generating an Image SubSet */
    gGraphImagePtr img2 = NULL;
    gGraphImagePtr img = (gGraphImagePtr) orig;

    *dest = NULL;
    if (img == NULL)
	return GGRAPH_ERROR;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_ERROR;

    img2 =
	gg_image_create (img->pixel_format, width, height, img->bits_per_sample,
			 img->samples_per_pixel, img->sample_format,
			 img->srs_name, img->proj4text);
    if (!img2)
	return GGRAPH_INSUFFICIENT_MEMORY;
    gg_image_sub_set (img2, img, upper_left_x, upper_left_y);
    *dest = img2;

    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageBufferReferenceRGB (const void *ptr, unsigned char **buffer)
{
/*
/ this function create a separate copy of the internal image buffer
/ if required color-space conversion will be applied
/ and the original image will be unaffected
*/
    unsigned char *buf;
    unsigned char *p_in;
    unsigned char *p_out;
    int scanline_width;
    int x;
    int y;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char gray;
    unsigned char index;
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    *buffer = NULL;
    if (img == NULL)
	return GGRAPH_ERROR;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_ERROR;

    scanline_width = img->width * 3;
    buf = malloc (scanline_width * img->height);
    if (!buf)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = buf + (y * scanline_width);
	  for (x = 0; x < img->width; x++)
	    {
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      *p_out++ = *p_in++;	/* red */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* blue */
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      *p_out++ = *p_in++;	/* red */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* blue */
		      p_in++;	/* discarding Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;
		      *p_out++ = *p_in++;	/* red */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* blue */
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      *p_out++ = red;
		      *p_out++ = green;
		      *p_out++ = blue;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      p_in++;	/* discarding Alpha */
		      *p_out++ = red;
		      *p_out++ = green;
		      *p_out++ = blue;
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      *p_out++ = red;
		      *p_out++ = green;
		      *p_out++ = blue;
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      gray = *p_in++;
		      *p_out++ = gray;
		      *p_out++ = gray;
		      *p_out++ = gray;
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      index = *p_in++;
		      *p_out++ = img->palette_red[index];
		      *p_out++ = img->palette_green[index];
		      *p_out++ = img->palette_blue[index];
		  }
	    }
      }
    *buffer = buf;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageBufferReferenceRGBA (const void *ptr, unsigned char **buffer)
{
/*
/ this function create a separate copy of the internal image buffer
/ if required color-space conversion will be applied
/ and the original image will be unaffected
*/
    unsigned char *buf;
    unsigned char *p_in;
    unsigned char *p_out;
    int scanline_width;
    int x;
    int y;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char alpha;
    unsigned char gray;
    unsigned char index;
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    *buffer = NULL;
    if (img == NULL)
	return GGRAPH_ERROR;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_ERROR;

    scanline_width = img->width * 4;
    buf = malloc (scanline_width * img->height);
    if (!buf)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = buf + (y * scanline_width);
	  for (x = 0; x < img->width; x++)
	    {
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      *p_out++ = *p_in++;	/* red */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* blue */
		      *p_out++ = 255;	/* opaque Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      *p_out++ = *p_in++;	/* red */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* blue */
		      *p_out++ = *p_in++;	/* Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      alpha = *p_in++;
		      *p_out++ = *p_in++;	/* red */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* blue */
		      *p_out++ = alpha;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      *p_out++ = red;
		      *p_out++ = green;
		      *p_out++ = blue;
		      *p_out++ = 255;	/* opaque Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      alpha = *p_in++;
		      *p_out++ = red;
		      *p_out++ = green;
		      *p_out++ = blue;
		      *p_out++ = alpha;
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      *p_out++ = red;
		      *p_out++ = green;
		      *p_out++ = blue;
		      *p_out++ = 255;	/* opaque Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      gray = *p_in++;
		      *p_out++ = gray;
		      *p_out++ = gray;
		      *p_out++ = gray;
		      *p_out++ = 255;	/* opaque Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      index = *p_in++;
		      *p_out++ = img->palette_red[index];
		      *p_out++ = img->palette_green[index];
		      *p_out++ = img->palette_blue[index];
		      *p_out++ = 255;	/* opaque Alpha */
		  }
	    }
      }
    *buffer = buf;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageBufferReferenceARGB (const void *ptr, unsigned char **buffer)
{
/*
/ this function create a separate copy of the internal image buffer
/ if required color-space conversion will be applied
/ and the original image will be unaffected
*/
    unsigned char *buf;
    unsigned char *p_in;
    unsigned char *p_out;
    int scanline_width;
    int x;
    int y;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char alpha;
    unsigned char gray;
    unsigned char index;
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    *buffer = NULL;
    if (img == NULL)
	return GGRAPH_ERROR;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_ERROR;

    scanline_width = img->width * 4;
    buf = malloc (scanline_width * img->height);
    if (!buf)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = buf + (y * scanline_width);
	  for (x = 0; x < img->width; x++)
	    {
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      *p_out++ = 255;	/* opaque Alpha */
		      *p_out++ = *p_in++;	/* red */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* blue */
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      alpha = *p_in++;
		      *p_out++ = *p_in++;	/* red */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* blue */
		      *p_in++ = alpha;
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      *p_out++ = *p_in++;	/* Alpha */
		      *p_out++ = *p_in++;	/* red */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* blue */
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      *p_out++ = 255;	/* opaque Alpha */
		      *p_out++ = red;
		      *p_out++ = green;
		      *p_out++ = blue;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      *p_out++ = *p_in++;	/* Alpha */
		      *p_out++ = red;
		      *p_out++ = green;
		      *p_out++ = blue;
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      *p_out++ = 255;	/* opaque Alpha */
		      *p_out++ = red;
		      *p_out++ = green;
		      *p_out++ = blue;
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      gray = *p_in++;
		      *p_out++ = 255;	/* opaque Alpha */
		      *p_out++ = gray;
		      *p_out++ = gray;
		      *p_out++ = gray;
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      index = *p_in++;
		      *p_out++ = 255;	/* opaque Alpha */
		      *p_out++ = img->palette_red[index];
		      *p_out++ = img->palette_green[index];
		      *p_out++ = img->palette_blue[index];
		  }
	    }
      }
    *buffer = buf;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageBufferReferenceBGR (const void *ptr, unsigned char **buffer)
{
/*
/ this function create a separate copy of the internal image buffer
/ if required color-space conversion will be applied
/ and the original image will be unaffected
*/
    unsigned char *buf;
    unsigned char *p_in;
    unsigned char *p_out;
    int scanline_width;
    int x;
    int y;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char gray;
    unsigned char index;
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    *buffer = NULL;
    if (img == NULL)
	return GGRAPH_ERROR;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_ERROR;

    scanline_width = img->width * 3;
    buf = malloc (scanline_width * img->height);
    if (!buf)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = buf + (y * scanline_width);
	  for (x = 0; x < img->width; x++)
	    {
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      *p_out++ = blue;
		      *p_out++ = green;
		      *p_out++ = red;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      p_in++;	/* discarding Alpha */
		      *p_out++ = blue;
		      *p_out++ = green;
		      *p_out++ = red;
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      *p_out++ = blue;
		      *p_out++ = green;
		      *p_out++ = red;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      *p_out++ = *p_in++;	/* blue */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* red */
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      *p_out++ = *p_in++;	/* blue */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* red */
		      p_in++;	/* discarding Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      *p_out++ = *p_in++;	/* blue */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* red */
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      gray = *p_in++;
		      *p_out++ = gray;
		      *p_out++ = gray;
		      *p_out++ = gray;
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      index = *p_in++;
		      *p_out++ = img->palette_blue[index];
		      *p_out++ = img->palette_green[index];
		      *p_out++ = img->palette_red[index];
		  }
	    }
      }
    *buffer = buf;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageBufferReferenceBGRA (const void *ptr, unsigned char **buffer)
{
/*
/ this function create a separate copy of the internal image buffer
/ if required color-space conversion will be applied
/ and the original image will be unaffected
*/
    unsigned char *buf;
    unsigned char *p_in;
    unsigned char *p_out;
    int scanline_width;
    int x;
    int y;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char alpha;
    unsigned char gray;
    unsigned char index;
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    *buffer = NULL;
    if (img == NULL)
	return GGRAPH_ERROR;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_ERROR;

    scanline_width = img->width * 4;
    buf = malloc (scanline_width * img->height);
    if (!buf)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = buf + (y * scanline_width);
	  for (x = 0; x < img->width; x++)
	    {
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      *p_out++ = blue;
		      *p_out++ = green;
		      *p_out++ = red;
		      *p_out++ = 255;	/* opaque Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      *p_out++ = blue;
		      *p_out++ = green;
		      *p_out++ = red;
		      *p_out++ = *p_in++;	/* Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      alpha = *p_in++;
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      *p_out++ = blue;
		      *p_out++ = green;
		      *p_out++ = red;
		      *p_out++ = alpha;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      *p_out++ = *p_in++;	/* blue */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* red */
		      *p_out++ = 255;	/* opaque Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      *p_out++ = *p_in++;	/* blue */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* red */
		      *p_out++ = *p_in++;	/* Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      *p_out++ = *p_in++;	/* blue */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* red */
		      *p_out++ = 255;	/* opaque Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      gray = *p_in++;
		      *p_out++ = gray;
		      *p_out++ = gray;
		      *p_out++ = gray;
		      *p_out++ = 255;	/* opaque Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      index = *p_in++;
		      *p_out++ = img->palette_blue[index];
		      *p_out++ = img->palette_green[index];
		      *p_out++ = img->palette_red[index];
		      *p_out++ = 255;	/* opaque Alpha */
		  }
	    }
      }
    *buffer = buf;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageBufferReferenceGDI (const void *ptr, unsigned char **buffer)
{
/*
/ this function create a separate copy of the internal image buffer
/ if required color-space conversion will be applied
/ and the original image will be unaffected
*/
    unsigned char *buf;
    unsigned char *p_in;
    unsigned char *p_out;
    int scanline_width;
    int x;
    int y;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char gray;
    unsigned char index;
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    *buffer = NULL;
    if (img == NULL)
	return GGRAPH_ERROR;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_ERROR;

    scanline_width = img->width * 3;
    buf = malloc (scanline_width * img->height);
    if (!buf)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = buf + (y * scanline_width);
	  for (x = 0; x < img->width; x++)
	    {
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      *p_out++ = blue;
		      *p_out++ = green;
		      *p_out++ = red;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      p_in++;	/* discarding Alpha */
		      *p_out++ = blue;
		      *p_out++ = green;
		      *p_out++ = red;
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      *p_out++ = blue;
		      *p_out++ = green;
		      *p_out++ = red;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      *p_out++ = *p_in++;	/* blue */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* red */
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      *p_out++ = *p_in++;	/* blue */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* red */
		      p_in++;	/* discarding Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      *p_out++ = *p_in++;	/* blue */
		      *p_out++ = *p_in++;	/* green */
		      *p_out++ = *p_in++;	/* red */
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      gray = *p_in++;
		      *p_out++ = gray;
		      *p_out++ = gray;
		      *p_out++ = gray;
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      index = *p_in++;
		      *p_out++ = img->palette_blue[index];
		      *p_out++ = img->palette_green[index];
		      *p_out++ = img->palette_red[index];
		  }
	    }
      }
    *buffer = buf;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageToRawMemBuf (const void *ptr, void **mem_buf, int *mem_buf_size)
{
/* exporting an image into a RAW un-compressed memory buffer */
    gGraphImagePtr img = (gGraphImagePtr) ptr;
    void *buf = NULL;
    int size;
    int ret;

    *mem_buf = NULL;
    *mem_buf_size = 0;
    if (img == NULL)
	return GGRAPH_INVALID_IMAGE;
    if (img->signature != GG_IMAGE_MAGIC_SIGNATURE)
	return GGRAPH_INVALID_IMAGE;

/* raw image generation */
    ret = gg_image_to_raw (img, &buf, &size);
    if (ret != GGRAPH_OK)
	return ret;

/* exporting the memory buffer */
    *mem_buf = buf;
    *mem_buf_size = size;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphImageFromRawMemBuf (const void *mem_buf, int mem_buf_size,
			  const void **image_handle, int scale)
{
/* importing an image from a RAW un-compressed memory buffer */
    gGraphImagePtr img = NULL;
    int ret;

    *image_handle = NULL;
    ret = gg_image_from_raw (mem_buf_size, mem_buf, &img, scale);
    if (ret != GGRAPH_OK)
	return ret;

    *image_handle = img;
    return GGRAPH_OK;
}

GGRAPH_DECLARE int
gGraphIsRawImage (const void *mem_buf, int mem_buf_size)
{
/* checks if this one is a RAW un-compressed memory buffer */
    unsigned char *p = (unsigned char *) mem_buf;
    unsigned char type;
    if (mem_buf_size < 14)
	return GGRAPH_ERROR;

/* checking the magic signature */
    if (*p++ != 0xaf)
	return GGRAPH_ERROR;
    if (*p++ != 0xfa)
	return GGRAPH_ERROR;
    if (*p++ != 0x80)
	return GGRAPH_ERROR;
    if (*p++ != 0x86)
	return GGRAPH_ERROR;

    type = *p;
    if (type == 0xee || type == 0xaa || type == 0xbb || type == 0xcc
	|| type == 0xdd || type == 0x11)
	;
    else
	return GGRAPH_ERROR;
    p = (unsigned char *) mem_buf + mem_buf_size - 4;
    if (*p++ != 0x80)
	return GGRAPH_ERROR;
    if (*p++ != 0x86)
	return GGRAPH_ERROR;
    if (*p++ != 0xfa)
	return GGRAPH_ERROR;
    if (*p++ != 0xaf)
	return GGRAPH_ERROR;

    return GGRAPH_OK;
}
