/* 
/ GGRAPH_image.c
/
/ image methods
/
/ version 1.0, 2010 July 20
/
/ Author: Sandro Furieri a.furieri@lqt.it
/
/ Copyright (C) 2010  Alessandro Furieri
/
/    This program is free software: you can redistribute it and/or modify
/    it under the terms of the GNU Lesser General Public License as published by
/    the Free Software Foundation, either version 3 of the License, or
/    (at your option) any later version.
/
/    This program is distributed in the hope that it will be useful,
/    but WITHOUT ANY WARRANTY; without even the implied warranty of
/    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/    GNU Lesser General Public License for more details.
/
/    You should have received a copy of the GNU Lesser General Public License
/    along with this program.  If not, see <http://www.gnu.org/licenses/>.
/
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>

#include "gaiagraphics.h"
#include "gaiagraphics_internals.h"

extern gGraphImageInfosPtr
gg_image_infos_create (int pixel_format, int width, int height,
		       int bits_per_sample, int samples_per_pixel,
		       int sample_format, const char *srs_name,
		       const char *proj4text)
{
/* creating an image infos struct */
    gGraphImageInfosPtr img;
    int len;
    char *SrsName = NULL;
    char *Proj4Text = NULL;

/* checking PIXEL_FORMAT */
    if (pixel_format == GG_PIXEL_RGB || pixel_format == GG_PIXEL_RGBA
	|| pixel_format == GG_PIXEL_ARGB || pixel_format == GG_PIXEL_BGR
	|| pixel_format == GG_PIXEL_BGRA || pixel_format == GG_PIXEL_GDI
	|| pixel_format == GG_PIXEL_GRAYSCALE
	|| pixel_format == GG_PIXEL_PALETTE || pixel_format == GG_PIXEL_GRID
	|| pixel_format == GG_PIXEL_UNKNOWN)
	;
    else
	return NULL;

    if (srs_name)
      {
	  len = strlen (srs_name);
	  SrsName = malloc (len + 1);
	  if (!SrsName)
	      return NULL;
	  strcpy (SrsName, srs_name);
      }
    if (proj4text)
      {
	  len = strlen (proj4text);
	  Proj4Text = malloc (len + 1);
	  if (!Proj4Text)
	    {
		if (SrsName)
		    free (SrsName);
		return NULL;
	    }
	  strcpy (Proj4Text, proj4text);
      }

/* allocating the image INFOS struct */
    img = malloc (sizeof (gGraphImage));
    if (!img)
	return NULL;

    img->signature = GG_IMAGE_INFOS_MAGIC_SIGNATURE;
    img->width = width;
    img->height = height;
    img->bits_per_sample = bits_per_sample;
    img->samples_per_pixel = samples_per_pixel;
    img->sample_format = sample_format;
    img->pixel_format = pixel_format;
    img->max_palette = 0;
    img->is_transparent = 0;
    img->tile_width = -1;
    img->tile_height = -1;
    img->rows_per_strip = -1;
    img->compression = GGRAPH_TIFF_COMPRESSION_NONE;
    img->scale_1_2 = 0;
    img->scale_1_4 = 0;
    img->scale_1_8 = 0;
    img->is_georeferenced = 0;
    img->srid = -1;
    img->srs_name = SrsName;
    img->proj4text = Proj4Text;
    img->upper_left_x = DBL_MAX;
    img->upper_left_y = DBL_MAX;
    img->pixel_x_size = 0.0;
    img->pixel_y_size = 0.0;

    return img;
}

extern void
gg_image_infos_destroy (gGraphImageInfosPtr img)
{
/* destroying an image INFOS struct */
    if (!img)
	return;
    if (img->srs_name)
	free (img->srs_name);
    if (img->proj4text)
	free (img->proj4text);
    free (img);
}

extern gGraphImagePtr
gg_image_create (int pixel_format, int width, int height, int bits_per_sample,
		 int samples_per_pixel, int sample_format, const char *srs_name,
		 const char *proj4text)
{
/* creating a generic image */
    gGraphImagePtr img;
    int len;
    char *SrsName = NULL;
    char *Proj4Text = NULL;

/* checking PIXEL_FORMAT */
    if (pixel_format == GG_PIXEL_RGB || pixel_format == GG_PIXEL_RGBA
	|| pixel_format == GG_PIXEL_ARGB || pixel_format == GG_PIXEL_BGR
	|| pixel_format == GG_PIXEL_BGRA || pixel_format == GG_PIXEL_GDI
	|| pixel_format == GG_PIXEL_GRAYSCALE
	|| pixel_format == GG_PIXEL_PALETTE || pixel_format == GG_PIXEL_GRID)
	;
    else
	return NULL;
    if (sample_format == GGRAPH_SAMPLE_UINT
	|| sample_format == GGRAPH_SAMPLE_INT
	|| sample_format == GGRAPH_SAMPLE_FLOAT)
	;
    else
	return NULL;

    if (srs_name)
      {
	  len = strlen (srs_name);
	  if (len > 0)
	    {
		SrsName = malloc (len + 1);
		if (!SrsName)
		    return NULL;
		strcpy (SrsName, srs_name);
	    }
      }
    if (proj4text)
      {
	  len = strlen (proj4text);
	  if (len > 0)
	    {
		Proj4Text = malloc (len + 1);
		if (!Proj4Text)
		  {
		      if (SrsName)
			  free (SrsName);
		      return NULL;
		  }
		strcpy (Proj4Text, proj4text);
	    }
      }

/* allocating the image struct */
    img = malloc (sizeof (gGraphImage));
    if (!img)
	return NULL;

    img->signature = GG_IMAGE_MAGIC_SIGNATURE;
    img->pixels = NULL;
    img->width = width;
    img->height = height;
    img->bits_per_sample = bits_per_sample;
    img->samples_per_pixel = samples_per_pixel;
    img->sample_format = sample_format;
    img->pixel_format = pixel_format;
    img->max_palette = 0;
    img->is_transparent = 0;
    img->tile_width = -1;
    img->tile_height = -1;
    img->rows_per_strip = -1;
    img->compression = GGRAPH_TIFF_COMPRESSION_NONE;
    img->is_georeferenced = 0;
    img->srid = -1;
    img->srs_name = SrsName;
    img->proj4text = Proj4Text;
    img->upper_left_x = DBL_MAX;
    img->upper_left_y = DBL_MAX;
    img->pixel_x_size = 0.0;
    img->pixel_y_size = 0.0;

/* computing the scanline size */
    if (pixel_format == GG_PIXEL_GDI)
      {
	  /* GDI required scanline padding */
	  int padding = 0;
	  while ((width * 3 + padding) % 4 != 0)
	      padding++;
	  img->scanline_width = (width * 3) + padding;
	  img->pixel_size = 3;
      }
    else if (pixel_format == GG_PIXEL_GRAYSCALE
	     || pixel_format == GG_PIXEL_PALETTE)
      {
	  img->scanline_width = width;
	  img->pixel_size = 1;
      }
    else if (pixel_format == GG_PIXEL_RGB || pixel_format == GG_PIXEL_BGR)
      {
	  img->scanline_width = width * 3;
	  img->pixel_size = 3;
      }
    else if (pixel_format == GG_PIXEL_GRID)
      {
	  if (sample_format == GGRAPH_SAMPLE_FLOAT)
	    {
		if (bits_per_sample == 32)
		  {
		      /* float */
		      img->scanline_width = width * 4;
		      img->pixel_size = 8;
		  }
		else
		  {
		      /* double */
		      img->scanline_width = width * 8;
		      img->pixel_size = 8;
		  }
	    }
	  else
	    {
		/* UINT or INT */
		if (bits_per_sample == 8)
		  {
		      /* 8bits int */
		      img->scanline_width = width;
		      img->pixel_size = 1;
		  }
		else if (bits_per_sample == 16)
		  {
		      /* 16bits int */
		      img->scanline_width = width * 2;
		      img->pixel_size = 2;
		  }
		else
		  {
		      /* 32bits int */
		      img->scanline_width = width * 4;
		      img->pixel_size = 4;
		  }
	    }
      }
    else
      {
	  /* RGBA, ARGB or BGRA */
	  img->scanline_width = width * 4;
	  img->pixel_size = 4;
      }

/* allocating the pixel buffer */
    img->pixels = malloc (img->scanline_width * height);
    if (!img->pixels)
      {
	  free (img);
	  return NULL;
      }
    return img;
}

extern void
gg_image_destroy (gGraphImagePtr img)
{
/* destroying a generic image */
    if (!img)
	return;
    if (img->pixels)
	free (img->pixels);
    if (img->srs_name)
	free (img->srs_name);
    if (img->proj4text)
	free (img->proj4text);
    free (img);
}

extern void
gg_image_clone_georeferencing (const gGraphImagePtr dst,
			       const gGraphImagePtr src)
{
/* adjusting georeferencing infos between two images */
    double size_x;
    double size_y;
    int len;
    char *SrsName = NULL;
    char *Proj4Text = NULL;
    if (src->is_georeferenced)
      {
	  if (src->srs_name)
	    {
		len = strlen (src->srs_name);
		if (len > 0)
		  {
		      SrsName = malloc (len + 1);
		      if (SrsName)
			  strcpy (SrsName, src->srs_name);
		  }
	    }
	  if (src->proj4text)
	    {
		len = strlen (src->proj4text);
		if (len > 0)
		  {
		      Proj4Text = malloc (len + 1);
		      if (Proj4Text)
			  strcpy (Proj4Text, src->proj4text);
		  }
	    }
      }

/* cleaning up destination georeferencing */
    dst->is_georeferenced = 0;
    dst->srid = -1;
    if (dst->srs_name)
	free (dst->srs_name);
    if (dst->proj4text)
	free (dst->proj4text);
    dst->srs_name = NULL;
    dst->proj4text = NULL;
    dst->upper_left_x = DBL_MAX;
    dst->upper_left_y = DBL_MAX;
    dst->pixel_x_size = 0.0;
    dst->pixel_y_size = 0.0;
    if (src->is_georeferenced == 0)
	return;

/* setting up destination georeferencing */
    dst->is_georeferenced = 1;
    dst->srid = src->srid;
    dst->srs_name = SrsName;
    dst->proj4text = Proj4Text;
    dst->upper_left_x = src->upper_left_x;
    dst->upper_left_y = src->upper_left_y;
    size_x = (double) (src->width) * src->pixel_x_size;
    size_y = (double) (src->height) * src->pixel_y_size;
    dst->pixel_x_size = size_x / (double) (dst->width);
    dst->pixel_y_size = size_y / (double) (dst->height);

}

static void
sub_set_georeferencing (const gGraphImagePtr dst, const gGraphImagePtr src,
			int upper_left_x, int upper_left_y)
{
/* adjusting georeferencing infos between two images [ImageSubSet] */
    int len;
    char *SrsName = NULL;
    char *Proj4Text = NULL;
    if (src->is_georeferenced)
      {
	  if (src->srs_name)
	    {
		len = strlen (src->srs_name);
		if (len > 0)
		  {
		      SrsName = malloc (len + 1);
		      if (SrsName)
			  strcpy (SrsName, src->srs_name);
		  }
	    }
	  if (src->proj4text)
	    {
		len = strlen (src->proj4text);
		if (len > 0)
		  {
		      Proj4Text = malloc (len + 1);
		      if (Proj4Text)
			  strcpy (Proj4Text, src->proj4text);
		  }
	    }
      }

/* cleaning up destination georeferencing */
    dst->is_georeferenced = 0;
    dst->srid = -1;
    if (dst->srs_name)
	free (dst->srs_name);
    if (dst->proj4text)
	free (dst->proj4text);
    dst->srs_name = NULL;
    dst->proj4text = NULL;
    dst->upper_left_x = DBL_MAX;
    dst->upper_left_y = DBL_MAX;
    dst->pixel_x_size = 0.0;
    dst->pixel_y_size = 0.0;
    if (src->is_georeferenced == 0)
	return;

/* setting up destination georeferencing */
    dst->is_georeferenced = 1;
    dst->srid = src->srid;
    dst->srs_name = SrsName;
    dst->proj4text = Proj4Text;
    dst->upper_left_x =
	src->upper_left_x + ((double) upper_left_x * src->pixel_x_size);
    dst->upper_left_y =
	src->upper_left_y - ((double) upper_left_y * src->pixel_y_size);
    dst->pixel_x_size = src->pixel_x_size;
    dst->pixel_y_size = src->pixel_y_size;

}

extern void
gg_image_sub_set (const gGraphImagePtr dst, const gGraphImagePtr src,
		  int upper_left_x, int upper_left_y)
{
/* creating an ImageSubSet */
    int x;
    int y;
    int dx;
    int dy = 0;
    int c;
    unsigned char *p_in;
    unsigned char *p_out;

    if (src->pixel_format != dst->pixel_format)
      {
	  /* an absolutely unexpected condition !!!! */
	  return;
      }

    for (y = upper_left_y; y < src->height; y++, dy++)
      {
	  if (dy >= dst->height)
	      break;
	  dx = 0;
	  p_in = src->pixels + (y * src->scanline_width);
	  p_in += upper_left_x * src->pixel_size;
	  p_out = dst->pixels + (dy * dst->scanline_width);
	  for (x = upper_left_x; x < src->width; x++, dx++)
	    {
		if (dx >= dst->width)
		    break;
		/* copying pixel's components */
		for (c = 0; c < src->pixel_size; c++)
		    *p_out++ = *p_in++;
	    }
      }
    if (src->pixel_format == GG_PIXEL_PALETTE)
      {
	  /* copying palette entries */
	  for (x = 0; x < src->max_palette; x++)
	    {
		dst->palette_red[x] = src->palette_red[x];
		dst->palette_green[x] = src->palette_green[x];
		dst->palette_blue[x] = src->palette_blue[x];
	    }
	  dst->max_palette = src->max_palette;
      }

/* computing georeferencing */
    sub_set_georeferencing (dst, src, upper_left_x, upper_left_y);
}

static unsigned char
to_grayscale (unsigned char r, unsigned char g, unsigned char b)
{
/* computing a grayscale value */
    double dval =
	(0.3 * (double) r) + (0.59 * (double) g) + (0.11 * (double) b);
    int ival = (int) dval;
    if (ival < 0)
	ival = 0;
    if (ival > 255)
	ival = 255;
    return (unsigned char) ival;
}

extern void
gg_image_fill (const gGraphImagePtr img, unsigned char r, unsigned char g,
	       unsigned char b, unsigned char alpha)
{
/* filling the image with given color */
    int x;
    int y;
    unsigned char *p;
    unsigned char gray;

/* resetting the palette anyway */
    img->max_palette = 1;
    img->palette_red[0] = r;
    img->palette_green[0] = g;
    img->palette_blue[0] = b;

/* precomputing a grayscale value anyway */
    if (r == g && g == b)
	gray = r;
    else
	gray = to_grayscale (r, g, b);

    for (y = 0; y < img->height; y++)
      {
	  /* setting pixels by scanline */
	  p = img->pixels + (y * img->scanline_width);
	  for (x = 0; x < img->width; x++)
	    {
		/* setting pixels */
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      *p++ = r;
		      *p++ = g;
		      *p++ = b;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      *p++ = r;
		      *p++ = g;
		      *p++ = b;
		      *p++ = alpha;
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      *p++ = alpha;
		      *p++ = r;
		      *p++ = g;
		      *p++ = b;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      *p++ = b;
		      *p++ = g;
		      *p++ = r;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      *p++ = b;
		      *p++ = g;
		      *p++ = r;
		      *p++ = alpha;
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      *p++ = b;
		      *p++ = g;
		      *p++ = r;
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		    *p++ = gray;
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		    *p++ = 0;
	    }
      }
}

extern unsigned char
gg_match_palette (const gGraphImagePtr img, unsigned char r, unsigned char g,
		  unsigned char b)
{
/* handling palette colors */
    int index;
    double min_dist = DBL_MAX;
    double dist;
    int min_index;
    for (index = 0; index < img->max_palette; index++)
      {
	  /* searching if already defined */
	  if (img->palette_red[index] == r && img->palette_green[index] == g
	      && img->palette_blue[index] == b)
	      return (unsigned char) index;
      }
    if (img->max_palette < 255)
      {
	  /* inserting a new palette entry */
	  unsigned char i = img->max_palette;
	  img->max_palette += 1;
	  img->palette_red[i] = r;
	  img->palette_green[i] = g;
	  img->palette_blue[i] = b;
	  return i;
      }
/* all right, the palette is already fully populated */
    for (index = 0; index < img->max_palette; index++)
      {
	  /* computing the minimal euclidean distance */
	  dist =
	      sqrt (((img->palette_red[index] - r) * (img->palette_red[index] -
						      r)) +
		    ((img->palette_green[index] -
		      g) * (img->palette_green[index] - g)) +
		    ((img->palette_blue[index] -
		      b) * (img->palette_blue[index] - b)));
	  if (dist < min_dist)
	    {
		min_dist = dist;
		min_index = index;
	    }
      }
    return (unsigned char) min_index;
}

static void
shrink_by (const gGraphImagePtr dst, const gGraphImagePtr src)
{
/*
/ this code is widely base upon the original wxWidgets gwxImage wxImage::ShrinkBy(() function
*/
    int xFactor = src->width / dst->width;
    int yFactor = src->height / dst->height;
    int x;
    int y;
    int x1;
    int y1;
    int y_offset;
    int x_offset;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char gray;
    unsigned char index;
    unsigned char *p;
    for (y = 0; y < dst->height; y++)
      {
	  for (x = 0; x < dst->width; x++)
	    {
		/* determine average */
		unsigned int avgRed = 0;
		unsigned int avgGreen = 0;
		unsigned int avgBlue = 0;
		unsigned int counter = 0;
		for (y1 = 0; y1 < yFactor; ++y1)
		  {
		      y_offset = (y * yFactor + y1);
		      for (x1 = 0; x1 < xFactor; ++x1)
			{
			    x_offset = (x * xFactor) + x1;
			    p = src->pixels + (y_offset * src->scanline_width);
			    if (src->pixel_format == GG_PIXEL_RGB)
			      {
				  p += x_offset * 3;
				  red = *p++;
				  green = *p++;
				  blue = *p;
			      }
			    else if (src->pixel_format == GG_PIXEL_RGBA)
			      {
				  p += x_offset * 4;
				  red = *p++;
				  green = *p++;
				  blue = *p;
			      }
			    else if (src->pixel_format == GG_PIXEL_ARGB)
			      {
				  p += x_offset * 4;
				  p++;
				  red = *p++;
				  green = *p++;
				  blue = *p;
			      }
			    else if (src->pixel_format == GG_PIXEL_BGR)
			      {
				  p += x_offset * 3;
				  blue = *p++;
				  green = *p++;
				  red = *p;
			      }
			    else if (src->pixel_format == GG_PIXEL_BGRA)
			      {
				  p += x_offset * 4;
				  blue = *p++;
				  green = *p++;
				  red = *p;
			      }
			    else if (src->pixel_format == GG_PIXEL_GDI)
			      {
				  p += x_offset * 3;
				  blue = *p++;
				  green = *p++;
				  red = *p;
			      }
			    else if (src->pixel_format == GG_PIXEL_GRAYSCALE)
			      {
				  p += x_offset;
				  red = *p;
				  green = *p;
				  blue = *p;
			      }
			    else if (src->pixel_format == GG_PIXEL_PALETTE)
			      {
				  p += x_offset;
				  red = src->palette_red[*p];
				  green = src->palette_green[*p];
				  blue = src->palette_blue[*p];
			      }
			    avgRed += red;
			    avgGreen += green;
			    avgBlue += blue;
			    counter++;
			}
		  }
		/* storing the pixel into destination */
		red = avgRed / counter;
		green = avgGreen / counter;
		blue = avgBlue / counter;
		p = dst->pixels + (y * dst->scanline_width);
		if (dst->pixel_format == GG_PIXEL_RGB)
		  {
		      p += x * 3;
		      *p++ = red;
		      *p++ = green;
		      *p = blue;
		  }
		else if (dst->pixel_format == GG_PIXEL_RGBA)
		  {
		      p += x * 4;
		      *p++ = red;
		      *p++ = green;
		      *p++ = blue;
		      *p = 255;	/* opaque */
		  }
		else if (dst->pixel_format == GG_PIXEL_ARGB)
		  {
		      p += x * 4;
		      *p++ = 255;	/* opaque */
		      *p++ = red;
		      *p++ = green;
		      *p = blue;
		  }
		else if (dst->pixel_format == GG_PIXEL_BGR)
		  {
		      p += x * 3;
		      *p++ = blue;
		      *p++ = green;
		      *p = red;
		  }
		else if (dst->pixel_format == GG_PIXEL_BGRA)
		  {
		      p += x * 4;
		      *p++ = blue;
		      *p++ = green;
		      *p++ = red;
		      *p = 255;	/* opaque */
		  }
		else if (dst->pixel_format == GG_PIXEL_GDI)
		  {
		      p += x * 3;
		      *p++ = blue;
		      *p++ = green;
		      *p = red;
		  }
		else if (dst->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      p += x;
		      if (red == green && green == blue)
			  gray = red;
		      else
			  gray = to_grayscale (red, green, blue);
		      *p = gray;
		  }
		else if (dst->pixel_format == GG_PIXEL_PALETTE)
		  {
		      p += x;
		      index = gg_match_palette (dst, red, green, blue);
		  }
	    }
      }
}

extern void
gg_image_resize (const gGraphImagePtr dst, const gGraphImagePtr src)
{
/*
/ this function builds an ordinary quality resized image, applying pixel replication
/
/ this code is widely base upon the original wxWidgets gwxImage wxImage::Scale(() function
/ wxIMAGE_QUALITY_NORMAL
*/
    int x_delta;
    int y_delta;
    int y;
    int j;
    int x;
    int i;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char gray;
    unsigned char index;
    unsigned char *p;

    if ((src->width % dst->width) == 0 && src->width >= dst->width
	&& (src->height % dst->height) == 0 && src->height >= dst->height)
      {
	  shrink_by (dst, src);
	  return;
      }
    x = src->width;
    y = src->height;
    x_delta = (x << 16) / dst->width;
    y_delta = (y << 16) / dst->height;
    y = 0;
    for (j = 0; j < dst->height; j++)
      {
	  x = 0;
	  for (i = 0; i < dst->width; i++)
	    {
		/* retrieving the origin pixel */
		p = src->pixels + ((y >> 16) * src->scanline_width);
		if (src->pixel_format == GG_PIXEL_RGB)
		  {
		      p += (x >> 16) * 3;
		      red = *p++;
		      green = *p++;
		      blue = *p;
		  }
		else if (src->pixel_format == GG_PIXEL_RGBA)
		  {
		      p += (x >> 16) * 4;
		      red = *p++;
		      green = *p++;
		      blue = *p;
		  }
		else if (src->pixel_format == GG_PIXEL_ARGB)
		  {
		      p += (x >> 16) * 4;
		      p++;
		      red = *p++;
		      green = *p++;
		      blue = *p;
		  }
		else if (src->pixel_format == GG_PIXEL_BGR)
		  {
		      p += (x >> 16) * 3;
		      blue = *p++;
		      green = *p++;
		      red = *p;
		  }
		else if (src->pixel_format == GG_PIXEL_BGRA)
		  {
		      p += (x >> 16) * 4;
		      blue = *p++;
		      green = *p++;
		      red = *p;
		  }
		else if (src->pixel_format == GG_PIXEL_GDI)
		  {
		      p += (x >> 16) * 3;
		      blue = *p++;
		      green = *p++;
		      red = *p;
		  }
		else if (src->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      p += (x >> 16);
		      red = *p;
		      green = *p;
		      blue = *p;
		  }
		else if (src->pixel_format == GG_PIXEL_PALETTE)
		  {
		      p += (x >> 16);
		      red = src->palette_red[*p];
		      green = src->palette_green[*p];
		      blue = src->palette_blue[*p];
		  }
		/* setting the destination pixel */
		p = dst->pixels + (j * dst->scanline_width);
		if (dst->pixel_format == GG_PIXEL_RGB)
		  {
		      p += i * 3;
		      *p++ = red;
		      *p++ = green;
		      *p = blue;
		  }
		else if (dst->pixel_format == GG_PIXEL_RGBA)
		  {
		      p += i * 4;
		      *p++ = red;
		      *p++ = green;
		      *p++ = blue;
		      *p = 255;	/* opaque */
		  }
		else if (dst->pixel_format == GG_PIXEL_ARGB)
		  {
		      p += i * 4;
		      *p++ = 255;	/* opaque */
		      *p++ = red;
		      *p++ = green;
		      *p = blue;
		  }
		else if (dst->pixel_format == GG_PIXEL_BGR)
		  {
		      p += i * 3;
		      *p++ = blue;
		      *p++ = green;
		      *p = red;
		  }
		else if (dst->pixel_format == GG_PIXEL_BGRA)
		  {
		      p += i * 4;
		      *p++ = blue;
		      *p++ = green;
		      *p++ = red;
		      *p = 255;	/* opaque */
		  }
		else if (dst->pixel_format == GG_PIXEL_GDI)
		  {
		      p += i * 3;
		      *p++ = blue;
		      *p++ = green;
		      *p = red;
		  }
		else if (dst->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      p += i;
		      if (red == green && green == blue)
			  gray = red;
		      else
			  gray = to_grayscale (red, green, blue);
		      *p = gray;
		  }
		else if (dst->pixel_format == GG_PIXEL_PALETTE)
		  {
		      p += i;
		      index = gg_match_palette (dst, red, green, blue);
		  }
		x += x_delta;
	    }
	  y += y_delta;
      }
}

#define floor2(exp) ((long) exp)

extern void
gg_make_thumbnail (const gGraphImagePtr thumbnail, const gGraphImagePtr image)
{
/*
/ this function builds an high quality thumbnail image, applying pixel interpolation
/
/ this code is widely base upon the original GD gdImageCopyResampled() function
*/
    int x, y;
    double sy1, sy2, sx1, sx2;
    unsigned char r;
    unsigned char g;
    unsigned char b;
    unsigned char gray;
    unsigned char index;
    unsigned char *p;
    for (y = 0; y < thumbnail->height; y++)
      {
	  sy1 =
	      ((double) y) * (double) image->height /
	      (double) thumbnail->height;
	  sy2 =
	      ((double) (y + 1)) * (double) image->height /
	      (double) thumbnail->height;
	  for (x = 0; x < thumbnail->width; x++)
	    {
		double sx, sy;
		double spixels = 0;
		double red = 0.0, green = 0.0, blue = 0.0;
		sx1 =
		    ((double) x) * (double) image->width /
		    (double) thumbnail->width;
		sx2 =
		    ((double) (x + 1)) * (double) image->width /
		    (double) thumbnail->width;
		sy = sy1;
		do
		  {
		      double yportion;
		      if (floor2 (sy) == floor2 (sy1))
			{
			    yportion = 1.0 - (sy - floor2 (sy));
			    if (yportion > sy2 - sy1)
			      {
				  yportion = sy2 - sy1;
			      }
			    sy = floor2 (sy);
			}
		      else if (sy == floor2 (sy2))
			{
			    yportion = sy2 - floor2 (sy2);
			}
		      else
			{
			    yportion = 1.0;
			}
		      sx = sx1;
		      do
			{
			    double xportion;
			    double pcontribution;
			    if (floor2 (sx) == floor2 (sx1))
			      {
				  xportion = 1.0 - (sx - floor2 (sx));
				  if (xportion > sx2 - sx1)
				    {
					xportion = sx2 - sx1;
				    }
				  sx = floor2 (sx);
			      }
			    else if (sx == floor2 (sx2))
			      {
				  xportion = sx2 - floor2 (sx2);
			      }
			    else
			      {
				  xportion = 1.0;
			      }
			    pcontribution = xportion * yportion;
			    /* retrieving the origin pixel */
			    p = image->pixels +
				((int) sy * image->scanline_width);
			    if (image->pixel_format == GG_PIXEL_RGB)
			      {
				  p += (int) sx *3;
				  r = *p++;
				  g = *p++;
				  b = *p;
			      }
			    else if (image->pixel_format == GG_PIXEL_RGBA)
			      {
				  p += (int) sx *4;
				  r = *p++;
				  g = *p++;
				  b = *p;
			      }
			    else if (image->pixel_format == GG_PIXEL_ARGB)
			      {
				  p += (int) sx *4;
				  p++;
				  r = *p++;
				  g = *p++;
				  b = *p;
			      }
			    else if (image->pixel_format == GG_PIXEL_BGR)
			      {
				  p += (int) sx *3;
				  b = *p++;
				  g = *p++;
				  r = *p;
			      }
			    else if (image->pixel_format == GG_PIXEL_BGRA)
			      {
				  p += (int) sx *4;
				  b = *p++;
				  g = *p++;
				  r = *p;
			      }
			    else if (image->pixel_format == GG_PIXEL_GDI)
			      {
				  p += (int) sx *3;
				  b = *p++;
				  g = *p++;
				  r = *p;
			      }
			    else if (image->pixel_format == GG_PIXEL_GRAYSCALE)
			      {
				  p += (int) sx;
				  r = *p;
				  g = *p;
				  b = *p;
			      }
			    else if (image->pixel_format == GG_PIXEL_PALETTE)
			      {
				  p += (int) sx;
				  r = image->palette_red[*p];
				  g = image->palette_green[*p];
				  b = image->palette_blue[*p];
			      }
			    red += r * pcontribution;
			    green += g * pcontribution;
			    blue += b * pcontribution;
			    spixels += xportion * yportion;
			    sx += 1.0;
			}
		      while (sx < sx2);
		      sy += 1.0;
		  }
		while (sy < sy2);
		if (spixels != 0.0)
		  {
		      red /= spixels;
		      green /= spixels;
		      blue /= spixels;
		  }
		if (red > 255.0)
		    red = 255.0;
		if (green > 255.0)
		    green = 255.0;
		if (blue > 255.0)
		    blue = 255.0;
		/* setting the destination pixel */
		p = thumbnail->pixels + (y * thumbnail->scanline_width);
		if (thumbnail->pixel_format == GG_PIXEL_RGB)
		  {
		      p += x * 3;
		      *p++ = (unsigned char) red;
		      *p++ = (unsigned char) green;
		      *p = (unsigned char) blue;
		  }
		else if (thumbnail->pixel_format == GG_PIXEL_RGBA)
		  {
		      p += x * 4;
		      *p++ = (unsigned char) red;
		      *p++ = (unsigned char) green;
		      *p++ = (unsigned char) blue;
		      *p = 255;	/* opaque */
		  }
		else if (thumbnail->pixel_format == GG_PIXEL_ARGB)
		  {
		      p += x * 4;
		      *p++ = 255;	/* opaque */
		      *p++ = (unsigned char) red;
		      *p++ = (unsigned char) green;
		      *p = (unsigned char) blue;
		  }
		else if (thumbnail->pixel_format == GG_PIXEL_BGR)
		  {
		      p += x * 3;
		      *p++ = (unsigned char) blue;
		      *p++ = (unsigned char) green;
		      *p = (unsigned char) red;
		  }
		else if (thumbnail->pixel_format == GG_PIXEL_BGRA)
		  {
		      p += x * 4;
		      *p++ = (unsigned char) blue;
		      *p++ = (unsigned char) green;
		      *p++ = (unsigned char) red;
		      *p = 255;	/* opaque */
		  }
		else if (thumbnail->pixel_format == GG_PIXEL_GDI)
		  {
		      p += x * 3;
		      *p++ = (unsigned char) blue;
		      *p++ = (unsigned char) green;
		      *p = (unsigned char) red;
		  }
		else if (thumbnail->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      p += x;
		      if (red == green && green == blue)
			  gray = (unsigned char) red;
		      else
			  gray =
			      to_grayscale ((unsigned char) red,
					    (unsigned char) green,
					    (unsigned char) blue);
		      *p = gray;
		  }
		else if (thumbnail->pixel_format == GG_PIXEL_PALETTE)
		  {
		      p += x;
		      index =
			  gg_match_palette (thumbnail, (unsigned char) red,
					    (unsigned char) green,
					    (unsigned char) blue);
		  }
	    }
      }
}

extern int
gg_convert_image_to_rgb (const gGraphImagePtr img)
{
/* converting this image to RGB */
    int x;
    int y;
    void *pixels;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char *p_in;
    unsigned char *p_out;

    if (img->pixel_format == GG_PIXEL_RGB)
	return GGRAPH_OK;

    pixels = malloc (img->width * img->height * 3);
    if (!pixels)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  /* processing any scanline */
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = pixels;
	  p_out += (y * img->width * 3);
	  for (x = 0; x < img->width; x++)
	    {
		/* retrieving the origin pixel */
		if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;	/* skipping alpha */
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      red = *p_in++;
		      green = red;
		      blue = red;

		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      int index = *p_in++;
		      red = img->palette_red[index];
		      green = img->palette_green[index];
		      blue = img->palette_blue[index];
		  }
		/* setting the destination pixel */
		*p_out++ = red;
		*p_out++ = green;
		*p_out++ = blue;
	    }
      }

    free (img->pixels);
    img->pixels = pixels;
    img->pixel_format = GG_PIXEL_RGB;
    img->scanline_width = img->width * 3;
    img->pixel_size = 3;
    img->max_palette = 0;
    return GGRAPH_OK;
}

extern int
gg_convert_image_to_rgba (const gGraphImagePtr img)
{
/* converting this image to RGBA */
    int x;
    int y;
    void *pixels;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char alpha;
    unsigned char *p_in;
    unsigned char *p_out;

    if (img->pixel_format == GG_PIXEL_RGBA)
	return GGRAPH_OK;

    pixels = malloc (img->width * img->height * 4);
    if (!pixels)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  /* processing any scanline */
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = pixels;
	  p_out += (y * img->width * 4);
	  for (x = 0; x < img->width; x++)
	    {
		/* retrieving the origin pixel */
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      alpha = *p_in++;
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      alpha = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      red = *p_in++;
		      green = red;
		      blue = red;
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      int index = *p_in++;
		      red = img->palette_red[index];
		      green = img->palette_green[index];
		      blue = img->palette_blue[index];
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		/* setting the destination pixel */
		*p_out++ = red;
		*p_out++ = green;
		*p_out++ = blue;
		*p_out++ = alpha;
	    }
      }

    free (img->pixels);
    img->pixels = pixels;
    img->pixel_format = GG_PIXEL_RGBA;
    img->scanline_width = img->width * 4;
    img->pixel_size = 4;
    img->max_palette = 0;
    return GGRAPH_OK;
}

extern int
gg_convert_image_to_argb (const gGraphImagePtr img)
{
/* converting this image to ARGB */
    int x;
    int y;
    void *pixels;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char alpha;
    unsigned char *p_in;
    unsigned char *p_out;

    if (img->pixel_format == GG_PIXEL_ARGB)
	return GGRAPH_OK;

    pixels = malloc (img->width * img->height * 4);
    if (!pixels)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  /* processing any scanline */
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = pixels;
	  p_out += (y * img->width * 4);
	  for (x = 0; x < img->width; x++)
	    {
		/* retrieving the origin pixel */
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      alpha = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      alpha = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      red = *p_in++;
		      green = red;
		      blue = red;
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      int index = *p_in++;
		      red = img->palette_red[index];
		      green = img->palette_green[index];
		      blue = img->palette_blue[index];
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		/* setting the destination pixel */
		*p_out++ = alpha;
		*p_out++ = red;
		*p_out++ = green;
		*p_out++ = blue;
	    }
      }

    free (img->pixels);
    img->pixels = pixels;
    img->pixel_format = GG_PIXEL_ARGB;
    img->scanline_width = img->width * 4;
    img->pixel_size = 4;
    img->max_palette = 0;
    return GGRAPH_OK;
}

extern int
gg_convert_image_to_bgr (const gGraphImagePtr img)
{
/* converting this image to BGR */
    int x;
    int y;
    void *pixels;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char *p_in;
    unsigned char *p_out;

    if (img->pixel_format == GG_PIXEL_BGR)
	return GGRAPH_OK;

    pixels = malloc (img->width * img->height * 3);
    if (!pixels)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  /* processing any scanline */
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = pixels;
	  p_out += (y * img->width * 3);
	  for (x = 0; x < img->width; x++)
	    {
		/* retrieving the origin pixel */
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;	/* skipping alpha */
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      red = *p_in++;
		      green = red;
		      blue = red;
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      int index = *p_in++;
		      red = img->palette_red[index];
		      green = img->palette_green[index];
		      blue = img->palette_blue[index];
		  }
		/* setting the destination pixel */
		*p_out++ = blue;
		*p_out++ = green;
		*p_out++ = red;
	    }
      }

    free (img->pixels);
    img->pixels = pixels;
    img->pixel_format = GG_PIXEL_BGR;
    img->scanline_width = img->width * 3;
    img->pixel_size = 3;
    img->max_palette = 0;
    return GGRAPH_OK;
}

extern int
gg_convert_image_to_bgra (const gGraphImagePtr img)
{
/* converting this image to BGRA */
    int x;
    int y;
    void *pixels;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char alpha;
    unsigned char *p_in;
    unsigned char *p_out;

    if (img->pixel_format == GG_PIXEL_ARGB)
	return GGRAPH_OK;

    pixels = malloc (img->width * img->height * 4);
    if (!pixels)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  /* processing any scanline */
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = pixels;
	  p_out += (y * img->width * 4);
	  for (x = 0; x < img->width; x++)
	    {
		/* retrieving the origin pixel */
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      alpha = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      alpha = *p_in++;
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      red = *p_in++;
		      green = red;
		      blue = red;
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      int index = *p_in++;
		      red = img->palette_red[index];
		      green = img->palette_green[index];
		      blue = img->palette_blue[index];
		      alpha = 255;
		      if (img->is_transparent && red == img->transparent_red
			  && green == img->transparent_green
			  && blue == img->transparent_blue)
			  alpha = 0;
		  }
		/* setting the destination pixel */
		*p_out++ = blue;
		*p_out++ = green;
		*p_out++ = red;
		*p_out++ = alpha;
	    }
      }

    free (img->pixels);
    img->pixels = pixels;
    img->pixel_format = GG_PIXEL_BGRA;
    img->scanline_width = img->width * 4;
    img->pixel_size = 4;
    img->max_palette = 0;
    return GGRAPH_OK;
}

extern int
gg_convert_image_to_gdi (const gGraphImagePtr img)
{
/* converting this image to GDI */
    int x;
    int y;
    void *pixels;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char *p_in;
    unsigned char *p_out;
    int padding;
    int scanline_width;

    if (img->pixel_format == GG_PIXEL_GDI)
	return GGRAPH_OK;

/* GDI required scanline padding */
    padding = 0;
    while ((img->width * 3 + padding) % 4 != 0)
	padding++;
    scanline_width = (img->width * 3) + padding;

    pixels = malloc (scanline_width * img->height * 3);
    if (!pixels)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  /* processing any scanline */
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = pixels;
	  p_out += (y * scanline_width);
	  for (x = 0; x < img->width; x++)
	    {
		/* retrieving the origin pixel */
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;	/* skipping alpha */
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      red = *p_in++;
		      green = red;
		      blue = red;
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      int index = *p_in++;
		      red = img->palette_red[index];
		      green = img->palette_green[index];
		      blue = img->palette_blue[index];
		  }
		/* setting the destination pixel */
		*p_out++ = blue;
		*p_out++ = green;
		*p_out++ = red;
	    }
      }

    free (img->pixels);
    img->pixels = pixels;
    img->pixel_format = GG_PIXEL_GDI;
    img->scanline_width = scanline_width;
    img->pixel_size = 3;
    img->max_palette = 0;
    return GGRAPH_OK;
}

extern int
gg_convert_image_to_grayscale (const gGraphImagePtr img)
{
/* converting this image to GRAYSCALE */
    int x;
    int y;
    void *pixels;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char gray;
    unsigned char *p_in;
    unsigned char *p_out;

    if (img->pixel_format == GG_PIXEL_GRAYSCALE)
	return GGRAPH_OK;

    pixels = malloc (img->width * img->height);
    if (!pixels)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  /* processing any scanline */
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = pixels;
	  p_out += (y * img->width);
	  for (x = 0; x < img->width; x++)
	    {
		/* retrieving the origin pixel */
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;	/* skipping alpha */
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      int index = *p_in++;
		      red = img->palette_red[index];
		      green = img->palette_green[index];
		      blue = img->palette_blue[index];
		  }
		/* setting the destination pixel */
		gray = to_grayscale (red, green, blue);
		*p_out++ = gray;
	    }
      }

    free (img->pixels);
    img->pixels = pixels;
    img->pixel_format = GG_PIXEL_GRAYSCALE;
    img->scanline_width = img->width;
    img->pixel_size = 1;
    img->max_palette = 0;
    return GGRAPH_OK;
}

extern int
gg_convert_image_to_palette (const gGraphImagePtr img)
{
/* converting this image to PALETTE */
    int x;
    int y;
    void *pixels;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char index;
    unsigned char *p_in;
    unsigned char *p_out;

    if (img->pixel_format == GG_PIXEL_PALETTE)
	return GGRAPH_OK;

    pixels = malloc (img->width * img->height);
    if (!pixels)
	return GGRAPH_INSUFFICIENT_MEMORY;
    img->max_palette = 0;
    for (y = 0; y < img->height; y++)
      {
	  /* processing any scanline */
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = pixels;
	  p_out += (y * img->width);
	  for (x = 0; x < img->width; x++)
	    {
		/* retrieving the origin pixel */
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;	/* skipping alpha */
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      red = *p_in++;
		      green = red;
		      blue = red;
		  }
		/* setting the destination pixel */
		index = gg_match_palette (img, red, green, blue);
		*p_out++ = index;
	    }
      }

    free (img->pixels);
    img->pixels = pixels;
    img->pixel_format = GG_PIXEL_PALETTE;
    img->scanline_width = img->width;
    img->pixel_size = 1;
    return GGRAPH_OK;
}

extern int
gg_convert_image_to_monochrome (const gGraphImagePtr img)
{
/* converting this image to MONOCHROME */
    int x;
    int y;
    void *pixels;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char index;
    unsigned char *p_in;
    unsigned char *p_out;

    if (img->pixel_format == GG_PIXEL_PALETTE
	&& gg_is_image_monochrome_ready (img) == GGRAPH_TRUE)
	return GGRAPH_OK;

    pixels = malloc (img->width * img->height);
    if (!pixels)
	return GGRAPH_INSUFFICIENT_MEMORY;
    for (y = 0; y < img->height; y++)
      {
	  /* processing any scanline */
	  p_in = img->pixels + (y * img->scanline_width);
	  p_out = pixels;
	  p_out += (y * img->width);
	  for (x = 0; x < img->width; x++)
	    {
		/* retrieving the origin pixel */
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;	/* skipping alpha */
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      red = *p_in++;
		      green = red;
		      blue = red;
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      int index = *p_in++;
		      red = img->palette_red[index];
		      green = img->palette_green[index];
		      blue = img->palette_blue[index];
		  }
		/* setting the destination pixel */
		if (red == 0 && green == 0 && blue == 0)
		    index = 0;
		else if (red == 255 && green == 255 && blue == 255)
		    index = 1;
		else
		  {
		      /* neither WHITE nor BLACK */
		      unsigned char gray = to_grayscale (red, green, blue);
		      if (gray < 128)
			{
			    /* defaulting to BLACK */
			    index = 0;
			}
		      else
			{
			    /* defaulting to WHITE */
			    index = 1;
			}
		  }
		*p_out++ = index;
	    }
      }

    free (img->pixels);
    img->pixels = pixels;
    img->pixel_format = GG_PIXEL_PALETTE;
    img->scanline_width = img->width;
    img->pixel_size = 1;
    img->max_palette = 2;
    img->palette_red[0] = 0;
    img->palette_green[0] = 0;
    img->palette_blue[0] = 0;
    img->palette_red[1] = 255;
    img->palette_green[1] = 255;
    img->palette_blue[1] = 255;
    return GGRAPH_OK;
}

extern int
gg_is_image_monochrome_ready (gGraphImagePtr img)
{
/* 
/ checking if this Image is into the Monochrome colorspace and ready to be used
/ expected: palette, 2 colors, black&white
*/
    int i;
    int black = 0;
    int white = 0;
    if (img->pixel_format != GG_PIXEL_PALETTE)
	return GGRAPH_FALSE;
    if (img->max_palette != 2)
	return GGRAPH_FALSE;
    for (i = 0; i < img->max_palette; i++)
      {
	  if (img->palette_red[i] == 0 && img->palette_green[i] == 0
	      && img->palette_blue[i] == 0)
	      black = 1;
	  if (img->palette_red[i] == 255 && img->palette_green[i] == 255
	      && img->palette_blue[i] == 255)
	      white = 1;
      }
    if (black && white)
	return GGRAPH_TRUE;
    return GGRAPH_FALSE;
}

extern int
gg_is_image_monochrome (gGraphImagePtr img)
{
/* checking if this Image is into the Monochrome colorspace */
    int x;
    int y;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char *p_in;
    for (y = 0; y < img->height; y++)
      {
	  p_in = img->pixels + (y * img->scanline_width);
	  for (x = 0; x < img->width; x++)
	    {
		/* retrieving a pixel */
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;	/* skipping alpha */
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		  {
		      red = *p_in++;
		      green = red;
		      blue = red;
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      int index = *p_in++;
		      red = img->palette_red[index];
		      green = img->palette_green[index];
		      blue = img->palette_blue[index];
		  }
		if (red == 0 && green == 0 && blue == 0)
		    continue;
		if (red == 255 && green == 255 && blue == 255)
		    continue;
		return GGRAPH_FALSE;
	    }
      }
    return GGRAPH_TRUE;
}

extern int
gg_is_image_grayscale (gGraphImagePtr img)
{
/* checking if this Image is into the GrayScale colorspace */
    int x;
    int y;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char *p_in;
    if (img->pixel_format == GG_PIXEL_GRAYSCALE)
	return GGRAPH_TRUE;
    for (y = 0; y < img->height; y++)
      {
	  p_in = img->pixels + (y * img->scanline_width);
	  for (x = 0; x < img->width; x++)
	    {
		/* retrieving a pixel */
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;	/* skipping alpha */
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      int index = *p_in++;
		      red = img->palette_red[index];
		      green = img->palette_green[index];
		      blue = img->palette_blue[index];
		  }
		if (red == green && green == blue)
		    continue;
		return GGRAPH_FALSE;
	    }
      }
    return GGRAPH_TRUE;
}

static int
palette_check (gGraphImagePtr img, unsigned char r, unsigned char g,
	       unsigned char b)
{
    unsigned char index;
    for (index = 0; index < img->max_palette; index++)
      {
	  /* searching if already defined */
	  if (img->palette_red[index] == r && img->palette_green[index] == g
	      && img->palette_blue[index] == b)
	      return GGRAPH_TRUE;
      }
    if (img->max_palette < 255)
      {
	  /* inserting a new palette entry */
	  unsigned char i = img->max_palette;
	  img->max_palette += 1;
	  img->palette_red[i] = r;
	  img->palette_green[i] = g;
	  img->palette_blue[i] = b;
	  return GGRAPH_TRUE;
      }
    return GGRAPH_FALSE;
}

extern int
gg_is_image_palette256 (gGraphImagePtr img)
{
/* checking if this Image may be represented using a 256 colors palette */
    int x;
    int y;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char *p_in;
    if (img->pixel_format == GG_PIXEL_PALETTE
	|| img->pixel_format == GG_PIXEL_GRAYSCALE)
	return GGRAPH_TRUE;
    img->max_palette = 0;
    for (y = 0; y < img->height; y++)
      {
	  p_in = img->pixels + (y * img->scanline_width);
	  for (x = 0; x < img->width; x++)
	    {
		/* retrieving a pixel */
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      red = *p_in++;
		      green = *p_in++;
		      blue = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;	/* skipping alpha */
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      p_in++;	/* skipping alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		  }
		if (palette_check (img, red, green, blue) == GGRAPH_FALSE)
		    return GGRAPH_FALSE;
	    }
      }
    return GGRAPH_TRUE;
}

extern int
gg_image_to_raw (const gGraphImagePtr img, void **mem_buf, int *mem_buf_size)
{
/* exporting a RAW image */
    unsigned char hi;
    unsigned char lo;
    unsigned char type;
    unsigned char *raw;
    unsigned char *p_raw;
    unsigned char *p_in;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char byte;
    unsigned char pixel;
    int pos;
    int x;
    int y;
    int scanline;
    int size = 14;		/* header and footer size */

    *mem_buf = NULL;
    *mem_buf_size = 0;
    switch (img->pixel_format)
      {
      case GG_PIXEL_GRAYSCALE:
	  size += img->width * img->height;
	  type = 0xee;
	  break;
      case GG_PIXEL_PALETTE:
	  if (img->max_palette <= 2)
	    {
		scanline = img->width / 8;
		if ((scanline * 8) < img->width)
		    scanline++;
		size += scanline * img->height;
		type = 0xaa;
	    }
	  else if (img->max_palette <= 4)
	    {
		scanline = img->width / 4;
		if ((scanline * 4) < img->width)
		    scanline++;
		size += scanline * img->height;
		type = 0xbb;
	    }
	  else if (img->max_palette <= 16)
	    {
		scanline = img->width / 2;
		if ((scanline * 2) < img->width)
		    scanline++;
		size += scanline * img->height;
		type = 0xcc;
	    }
	  else
	    {
		size += img->width * img->height;
		type = 0xdd;
	    }
	  size += img->max_palette * 3;	/* palette */
	  break;
      default:
	  size += img->width * img->height * 3;
	  type = 0x11;
	  break;
      };

    raw = malloc (size);
    if (!raw)
	return GGRAPH_INSUFFICIENT_MEMORY;

/* RAW image header */
    raw[0] = 0xaf;		/* 4 bytes - magic signature */
    raw[1] = 0xfa;
    raw[2] = 0x80;
    raw[3] = 0x86;
    raw[4] = type;
    hi = img->width / 256;	/* width */
    lo = img->width % 256;
    raw[5] = hi;
    raw[6] = lo;
    hi = img->height / 256;	/* height */
    lo = img->height % 256;
    raw[7] = hi;
    raw[8] = lo;
    if (img->pixel_format == GG_PIXEL_PALETTE)
	raw[9] = (img->max_palette - 1);	/* #palette entries */
    else
	raw[9] = 0x00;
    p_raw = raw + 10;

    if (img->pixel_format == GG_PIXEL_PALETTE)
      {
	  /* palette entries */
	  for (x = 0; x < img->max_palette; x++)
	    {
		*p_raw++ = img->palette_red[x];
		*p_raw++ = img->palette_green[x];
		*p_raw++ = img->palette_blue[x];
	    }
      }

    for (y = 0; y < img->height; y++)
      {
	  /* looping on image rows */
	  p_in = img->pixels + (y * img->scanline_width);
	  byte = 0x00;
	  pos = 0;
	  for (x = 0; x < img->width; x++)
	    {
		/* processing a scanline */
		if (img->pixel_format == GG_PIXEL_GRAYSCALE)
		    *p_raw++ = *p_in++;
		else if (img->pixel_format == GG_PIXEL_PALETTE)
		  {
		      if (img->max_palette <= 2)
			{
			    /* monochrome */
			    pixel = *p_in++;
			    if (pixel != 0)
			      {
				  switch (pos)
				    {
				    case 0:
					byte |= 0x80;
					break;
				    case 1:
					byte |= 0x40;
					break;
				    case 2:
					byte |= 0x20;
					break;
				    case 3:
					byte |= 0x10;
					break;
				    case 4:
					byte |= 0x08;
					break;
				    case 5:
					byte |= 0x04;
					break;
				    case 6:
					byte |= 0x02;
					break;
				    case 7:
					byte |= 0x01;
					break;
				    };
			      }
			    pos++;
			    if (pos > 7)
			      {
				  /* exporting an octet */
				  *p_raw++ = byte;
				  byte = 0x00;
				  pos = 0;
			      }
			}
		      else if (img->max_palette <= 4)
			{
			    /* 2 bits per sample */
			    pixel = *p_in++;
			    if (pixel != 0)
			      {
				  unsigned char mask = 0x00;
				  switch (pos)
				    {
				    case 0:
					switch (pixel)
					  {
					  case 1:
					      mask = 0x40;
					      break;
					  case 2:
					      mask = 0x80;
					      break;
					  case 3:
					      mask = 0xc0;
					      break;
					  };
					byte |= mask;
					break;
				    case 2:
					switch (pixel)
					  {
					  case 1:
					      mask = 0x10;
					      break;
					  case 2:
					      mask = 0x20;
					      break;
					  case 3:
					      mask = 0x30;
					      break;
					  };
					byte |= mask;
					break;
				    case 4:
					switch (pixel)
					  {
					  case 1:
					      mask = 0x04;
					      break;
					  case 2:
					      mask = 0x08;
					      break;
					  case 3:
					      mask = 0x0c;
					      break;
					  };
					byte |= mask;
					break;
				    case 6:
					switch (pixel)
					  {
					  case 1:
					      mask = 0x01;
					      break;
					  case 2:
					      mask = 0x02;
					      break;
					  case 3:
					      mask = 0x03;
					      break;
					  };
					byte |= mask;
					break;
				    };
			      }
			    pos += 2;
			    if (pos > 7)
			      {
				  /* exporting an octet */
				  *p_raw++ = byte;
				  byte = 0x00;
				  pos = 0;
			      }
			}
		      else if (img->max_palette <= 16)
			{
			    /* 4 bits per sample */
			    pixel = *p_in++;
			    if (pixel != 0)
			      {
				  unsigned char mask = 0x00;
				  switch (pos)
				    {
				    case 0:
					switch (pixel)
					  {
					  case 1:
					      mask = 0x10;
					      break;
					  case 2:
					      mask = 0x20;
					      break;
					  case 3:
					      mask = 0x30;
					      break;
					  case 4:
					      mask = 0x40;
					      break;
					  case 5:
					      mask = 0x50;
					      break;
					  case 6:
					      mask = 0x60;
					      break;
					  case 7:
					      mask = 0x70;
					      break;
					  case 8:
					      mask = 0x80;
					      break;
					  case 9:
					      mask = 0x90;
					      break;
					  case 10:
					      mask = 0xa0;
					      break;
					  case 11:
					      mask = 0xb0;
					      break;
					  case 12:
					      mask = 0xc0;
					      break;
					  case 13:
					      mask = 0xd0;
					      break;
					  case 14:
					      mask = 0xe0;
					      break;
					  case 15:
					      mask = 0xf0;
					      break;
					  };
					byte |= mask;
					break;
				    default:
					switch (pixel)
					  {
					  case 1:
					      mask = 0x01;
					      break;
					  case 2:
					      mask = 0x02;
					      break;
					  case 3:
					      mask = 0x03;
					      break;
					  case 4:
					      mask = 0x04;
					      break;
					  case 5:
					      mask = 0x05;
					      break;
					  case 6:
					      mask = 0x06;
					      break;
					  case 7:
					      mask = 0x07;
					      break;
					  case 8:
					      mask = 0x08;
					      break;
					  case 9:
					      mask = 0x09;
					      break;
					  case 10:
					      mask = 0x0a;
					      break;
					  case 11:
					      mask = 0x0b;
					      break;
					  case 12:
					      mask = 0x0c;
					      break;
					  case 13:
					      mask = 0x0d;
					      break;
					  case 14:
					      mask = 0x0e;
					      break;
					  case 15:
					      mask = 0x0f;
					      break;
					  };
					byte |= mask;
					break;
				    };
			      }
			    pos += 4;
			    if (pos > 7)
			      {
				  /* exporting an octet */
				  *p_raw++ = byte;
				  byte = 0x00;
				  pos = 0;
			      }
			}
		      else
			{
			    /* 8 bits per sample */
			    *p_raw++ = *p_in++;
			}
		  }
		else if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      *p_raw++ = *p_in++;	/* red */
		      *p_raw++ = *p_in++;	/* green */
		      *p_raw++ = *p_in++;	/* blue */
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      *p_raw++ = *p_in++;	/* red */
		      *p_raw++ = *p_in++;	/* green */
		      *p_raw++ = *p_in++;	/* blue */
		      p_in++;	/* discarding Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;	/* discarding Alpha */
		      *p_raw++ = *p_in++;	/* red */
		      *p_raw++ = *p_in++;	/* green */
		      *p_raw++ = *p_in++;	/* blue */
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      *p_raw++ = red;
		      *p_raw++ = green;
		      *p_raw++ = blue;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      p_in++;	/* discarding Alpha */
		      *p_raw++ = red;
		      *p_raw++ = green;
		      *p_raw++ = blue;
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      blue = *p_in++;
		      green = *p_in++;
		      red = *p_in++;
		      *p_raw++ = red;
		      *p_raw++ = green;
		      *p_raw++ = blue;
		  }
	    }
	  if (pos > 0)		/* exporting the last octet */
	      *p_raw++ = byte;
      }

/* RAW image footer */
    *p_raw++ = 0x80;		/* 4 bytes - magic signature */
    *p_raw++ = 0x86;
    *p_raw++ = 0xfa;
    *p_raw++ = 0xaf;

    *mem_buf = raw;
    *mem_buf_size = size;
    return GGRAPH_OK;
}

extern int
gg_image_from_raw (int mem_buf_size, const void *mem_buf,
		   gGraphImagePtr * image_handle, int scale)
{
/* importing a RAW image */
    gGraphImagePtr img = NULL;
    int x;
    int y;
    int width;
    int height;
    int max_palette;
    int pixel_format;
    int bits_per_sample;
    int samples_per_pixel;
    unsigned char *p = (unsigned char *) mem_buf;
    unsigned char type;
    int scanline_width;
    unsigned char byte;

    *image_handle = NULL;
    if (mem_buf_size < 14)
	return GGRAPH_ERROR;

/* checking the magic signature */
    if (*p++ != 0xaf)
	return GGRAPH_ERROR;
    if (*p++ != 0xfa)
	return GGRAPH_ERROR;
    if (*p++ != 0x80)
	return GGRAPH_ERROR;
    if (*p++ != 0x86)
	return GGRAPH_ERROR;
    type = *p;
    if (type == 0xee || type == 0xaa || type == 0xbb || type == 0xcc
	|| type == 0xdd || type == 0x11)
	;
    else
	return GGRAPH_ERROR;
    p = (unsigned char *) mem_buf + mem_buf_size - 4;
    if (*p++ != 0x80)
	return GGRAPH_ERROR;
    if (*p++ != 0x86)
	return GGRAPH_ERROR;
    if (*p++ != 0xfa)
	return GGRAPH_ERROR;
    if (*p++ != 0xaf)
	return GGRAPH_ERROR;

/* all right, seems to be a RAW image: going on */
    p = (unsigned char *) mem_buf;
    width = (p[5] * 256) + p[6];
    height = (p[7] * 256) + p[8];
    max_palette = p[9] + 1;
    switch (type)
      {
      case 0x11:
	  pixel_format = GG_PIXEL_RGB;
	  bits_per_sample = 8;
	  samples_per_pixel = 3;
	  break;
      case 0xee:
	  pixel_format = GG_PIXEL_GRAYSCALE;
	  bits_per_sample = 8;
	  samples_per_pixel = 1;
	  break;
      case 0xaa:
	  pixel_format = GG_PIXEL_PALETTE;
	  bits_per_sample = 1;
	  samples_per_pixel = 1;
	  break;
      case 0xbb:
	  pixel_format = GG_PIXEL_PALETTE;
	  bits_per_sample = 2;
	  samples_per_pixel = 1;
	  break;
      case 0xcc:
	  pixel_format = GG_PIXEL_PALETTE;
	  bits_per_sample = 4;
	  samples_per_pixel = 1;
	  break;
      case 0xdd:
	  pixel_format = GG_PIXEL_PALETTE;
	  bits_per_sample = 8;
	  samples_per_pixel = 1;
	  break;
      };

    img =
	gg_image_create (pixel_format, width, height, bits_per_sample,
			 samples_per_pixel, GGRAPH_SAMPLE_UINT, NULL, NULL);
    if (!img)
	return GGRAPH_INSUFFICIENT_MEMORY;

    p = (unsigned char *) mem_buf + 10;
    if (img->pixel_format == GG_PIXEL_PALETTE)
      {
	  /* loading the palette colors */
	  for (x = 0; x < max_palette; x++)
	    {
		img->palette_red[x] = *p++;
		img->palette_green[x] = *p++;
		img->palette_blue[x] = *p++;
	    }
      }

    for (y = 0; y < img->height; y++)
      {
	  unsigned char *p_out = img->pixels + (y * img->scanline_width);
	  if (img->pixel_format == GG_PIXEL_RGB)
	    {
		for (x = 0; x < img->width; x++)
		  {
		      *p_out++ = *p++;	/* red */
		      *p_out++ = *p++;	/* green */
		      *p_out++ = *p++;	/* blue */
		  }
	    }
	  else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
	    {
		for (x = 0; x < img->width; x++)
		    *p_out++ = *p++;
	    }
	  else if (img->pixel_format == GG_PIXEL_PALETTE)
	    {
		if (bits_per_sample == 1)
		  {
		      unsigned char *p_base = p_out;
		      scanline_width = img->width / 8;
		      if ((scanline_width * 8) < img->width)
			  scanline_width++;
		      for (x = 0; x < scanline_width; x++)
			{
			    byte = *p++;
			    if ((byte & 0x80) == 0x80)
				*p_out++ = 1;
			    else
				*p_out++ = 0;
			    if (p_out - p_base >= img->width)
				break;
			    if ((byte & 0x40) == 0x40)
				*p_out++ = 1;
			    else
				*p_out++ = 0;
			    if (p_out - p_base >= img->width)
				break;
			    if ((byte & 0x20) == 0x20)
				*p_out++ = 1;
			    else
				*p_out++ = 0;
			    if (p_out - p_base >= img->width)
				break;
			    if ((byte & 0x10) == 0x10)
				*p_out++ = 1;
			    else
				*p_out++ = 0;
			    if (p_out - p_base >= img->width)
				break;
			    if ((byte & 0x08) == 0x08)
				*p_out++ = 1;
			    else
				*p_out++ = 0;
			    if (p_out - p_base >= img->width)
				break;
			    if ((byte & 0x04) == 0x04)
				*p_out++ = 1;
			    else
				*p_out++ = 0;
			    if (p_out - p_base >= img->width)
				break;
			    if ((byte & 0x02) == 0x02)
				*p_out++ = 1;
			    else
				*p_out++ = 0;
			    if (p_out - p_base >= img->width)
				break;
			    if ((byte & 0x01) == 0x01)
				*p_out++ = 1;
			    else
				*p_out++ = 0;
			}
		  }
		else if (bits_per_sample == 2)
		  {
		      unsigned char *p_base = p_out;
		      scanline_width = img->width / 2;
		      if ((scanline_width * 2) < img->width)
			  scanline_width++;
		      for (x = 0; x < scanline_width; x++)
			{
			    byte = *p++;
			    switch (byte & 0xc0)
			      {
			      case 0xc0:
				  *p_out++ = 3;
				  break;
			      case 0x80:
				  *p_out++ = 2;
				  break;
			      case 0x40:
				  *p_out++ = 1;
				  break;
			      case 0x00:
				  *p_out++ = 0;
				  break;
			      };
			    if (p_out - p_base >= img->width)
				break;
			    switch (byte & 0x30)
			      {
			      case 0x30:
				  *p_out++ = 3;
				  break;
			      case 0x20:
				  *p_out++ = 2;
				  break;
			      case 0x10:
				  *p_out++ = 1;
				  break;
			      case 0x00:
				  *p_out++ = 0;
				  break;
			      };
			    if (p_out - p_base >= img->width)
				break;
			    switch (byte & 0x0c)
			      {
			      case 0x0c:
				  *p_out++ = 3;
				  break;
			      case 0x08:
				  *p_out++ = 2;
				  break;
			      case 0x04:
				  *p_out++ = 1;
				  break;
			      case 0x00:
				  *p_out++ = 0;
				  break;
			      };
			    if (p_out - p_base >= img->width)
				break;
			    switch (byte & 0x03)
			      {
			      case 0x03:
				  *p_out++ = 3;
				  break;
			      case 0x02:
				  *p_out++ = 2;
				  break;
			      case 0x01:
				  *p_out++ = 1;
				  break;
			      case 0x00:
				  *p_out++ = 0;
				  break;
			      };
			}
		  }
		else if (bits_per_sample == 4)
		  {
		      unsigned char *p_base = p_out;
		      scanline_width = img->width / 2;
		      if ((scanline_width * 2) < img->width)
			  scanline_width++;
		      for (x = 0; x < scanline_width; x++)
			{
			    byte = *p++;
			    switch (byte & 0xf0)
			      {
			      case 0xf0:
				  *p_out++ = 15;
				  break;
			      case 0xe0:
				  *p_out++ = 14;
				  break;
			      case 0xd0:
				  *p_out++ = 13;
				  break;
			      case 0xc0:
				  *p_out++ = 12;
				  break;
			      case 0xb0:
				  *p_out++ = 11;
				  break;
			      case 0xa0:
				  *p_out++ = 10;
				  break;
			      case 0x90:
				  *p_out++ = 9;
				  break;
			      case 0x80:
				  *p_out++ = 8;
				  break;
			      case 0x70:
				  *p_out++ = 7;
				  break;
			      case 0x60:
				  *p_out++ = 6;
				  break;
			      case 0x50:
				  *p_out++ = 5;
				  break;
			      case 0x40:
				  *p_out++ = 4;
				  break;
			      case 0x30:
				  *p_out++ = 3;
				  break;
			      case 0x20:
				  *p_out++ = 2;
				  break;
			      case 0x10:
				  *p_out++ = 1;
				  break;
			      case 0x00:
				  *p_out++ = 0;
				  break;
			      };
			    if (p_out - p_base >= img->width)
				break;
			    switch (byte & 0x0f)
			      {
			      case 0x0f:
				  *p_out++ = 15;
				  break;
			      case 0x0e:
				  *p_out++ = 14;
				  break;
			      case 0x0d:
				  *p_out++ = 13;
				  break;
			      case 0x0c:
				  *p_out++ = 12;
				  break;
			      case 0x0b:
				  *p_out++ = 11;
				  break;
			      case 0x0a:
				  *p_out++ = 10;
				  break;
			      case 0x09:
				  *p_out++ = 9;
				  break;
			      case 0x08:
				  *p_out++ = 8;
				  break;
			      case 0x07:
				  *p_out++ = 7;
				  break;
			      case 0x06:
				  *p_out++ = 6;
				  break;
			      case 0x05:
				  *p_out++ = 5;
				  break;
			      case 0x04:
				  *p_out++ = 4;
				  break;
			      case 0x03:
				  *p_out++ = 3;
				  break;
			      case 0x02:
				  *p_out++ = 2;
				  break;
			      case 0x01:
				  *p_out++ = 1;
				  break;
			      case 0x00:
				  *p_out++ = 0;
				  break;
			      };
			}
		  }
		else
		  {
		      for (x = 0; x < img->width; x++)
			  *p_out++ = *p++;
		  }
	    }
      }

    *image_handle = img;
    return GGRAPH_OK;
}
