/* 
/ GGRAPH_tiff.c
/
/ TIFF auxiliary helpers
/
/ version 1.0, 2010 July 20
/
/ Author: Sandro Furieri a.furieri@lqt.it
/
/ Copyright (C) 2010  Alessandro Furieri
/
/    This program is free software: you can redistribute it and/or modify
/    it under the terms of the GNU Lesser General Public License as published by
/    the Free Software Foundation, either version 3 of the License, or
/    (at your option) any later version.
/
/    This program is distributed in the hope that it will be useful,
/    but WITHOUT ANY WARRANTY; without even the implied warranty of
/    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/    GNU Lesser General Public License for more details.
/
/    You should have received a copy of the GNU Lesser General Public License
/    along with this program.  If not, see <http://www.gnu.org/licenses/>.
/
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include <geotiff.h>
#include <xtiffio.h>
#include <geo_tiffp.h>
#include <geo_keyp.h>
#include <geovalues.h>
#include <geo_normalize.h>

#include "gaiagraphics.h"
#include "gaiagraphics_internals.h"

struct memfile
{
/* a struct emulating a file [memory mapped] */
    FILE *file;
    unsigned char *buffer;
    int malloc_block;
    tsize_t size;
    tsize_t eof;
    toff_t current;
};

static void
memory_realloc (struct memfile *mem, tsize_t req_size)
{
/* expanding the allocated memory */
    unsigned char *new_buffer;
    tsize_t new_size = mem->size;
    while (new_size < req_size)
	new_size += mem->malloc_block;
    new_buffer = realloc (mem->buffer, new_size);
    if (!new_buffer)
	return;
    mem->buffer = new_buffer;
    mem->size = new_size;
}

static tsize_t
memory_readproc (thandle_t clientdata, tdata_t data, tsize_t size)
{
/* emulating the read()  function */
    struct memfile *mem = clientdata;
    tsize_t len;
    if (mem->current >= (toff_t) mem->eof)
	return 0;
    len = size;
    if ((mem->current + size) >= (toff_t) mem->eof)
	len = (tsize_t) (mem->eof - mem->current);
    memcpy (data, mem->buffer + mem->current, len);
    mem->current += len;
    return len;
}

static tsize_t
file_readproc (thandle_t clientdata, tdata_t data, tsize_t size)
{
/* read()  function */
    struct memfile *mem = clientdata;
    return fread (data, 1, size, mem->file);
}

static tsize_t
memory_writeproc (thandle_t clientdata, tdata_t data, tsize_t size)
{
/* emulating the write()  function */
    struct memfile *mem = clientdata;
    if ((mem->current + size) >= (toff_t) mem->size)
	memory_realloc (mem, mem->current + size);
    if ((mem->current + size) >= (toff_t) mem->size)
	return -1;
    memcpy (mem->buffer + mem->current, (unsigned char *) data, size);
    mem->current += size;
    if (mem->current > (toff_t) mem->eof)
	mem->eof = (tsize_t) (mem->current);
    return size;
}

static tsize_t
file_writeproc (thandle_t clientdata, tdata_t data, tsize_t size)
{
/* write()  function */
    struct memfile *mem = clientdata;
    size_t wr = fwrite (data, 1, size, mem->file);
    return wr;
}

static toff_t
memory_seekproc (thandle_t clientdata, toff_t offset, int whence)
{
/* emulating the lseek()  function */
    struct memfile *mem = clientdata;
    switch (whence)
      {
      case SEEK_CUR:
	  if ((int) (mem->current + offset) < 0)
	      return (toff_t) - 1;
	  mem->current += offset;
	  if ((toff_t) mem->eof < mem->current)
	      mem->eof = (tsize_t) (mem->current);
	  break;
      case SEEK_END:
	  if ((int) (mem->eof + offset) < 0)
	      return (toff_t) - 1;
	  mem->current = mem->eof + offset;
	  if ((toff_t) mem->eof < mem->current)
	      mem->eof = (tsize_t) (mem->current);
	  break;
      case SEEK_SET:
      default:
	  if ((int) offset < 0)
	      return (toff_t) - 1;
	  mem->current = offset;
	  if ((toff_t) mem->eof < mem->current)
	      mem->eof = (tsize_t) (mem->current);
	  break;
      };
    return mem->current;
}

static toff_t
file_seekproc (thandle_t clientdata, toff_t offset, int whence)
{
/* lseek()  function */
    struct memfile *mem = clientdata;
    if (fseek (mem->file, offset, whence) < 0)
	return (toff_t) - 1;
    return ftell (mem->file);
}

static int
closeproc (thandle_t clientdata)
{
/* emulating the close()  function */
    if (clientdata)
	return 0;		/* does absolutely nothing - required in order to suppress warnings */
    return 0;
}

static toff_t
memory_sizeproc (thandle_t clientdata)
{
/* returning the pseudo-file current size */
    struct memfile *mem = clientdata;
    return mem->eof;
}

static toff_t
file_sizeproc (thandle_t clientdata)
{
/* returning the file current size */
    struct memfile *mem = clientdata;
    toff_t size;
    long current = ftell (mem->file);
    if (fseek (mem->file, 0, SEEK_END) < 0)
	return 0;
    size = ftell (mem->file);
    if (fseek (mem->file, current, SEEK_SET) < 0)
	return 0;
    return size;
}

static int
mapproc (thandle_t clientdata, tdata_t * data, toff_t * offset)
{
    if (clientdata || data || offset)
	return 0;		/* does absolutely nothing - required in order to suppress warnings */
    return 0;
}

static void
unmapproc (thandle_t clientdata, tdata_t data, toff_t offset)
{
    if (clientdata || data || offset)
	return;			/* does absolutely nothing - required in order to suppress warnings */
    return;
}

static int
common_tiff_fax_strip (TIFF * out, const gGraphImagePtr img, int strip_rows)
{
/* common function writing a tiff raster [monochrome/strip] */
    int row;
    int col;
    tsize_t scan_bytes;
    tsize_t strip_bytes;
    tsize_t strip_effective;
    tstrip_t nstrip = 0;
    unsigned char *strip = NULL;
    unsigned char *strip_ptr;
    int pixel;
    unsigned char byte;
    int pos;

/* allocating the scan-line buffer */
    scan_bytes = TIFFScanlineSize (out);
    strip_bytes = scan_bytes * strip_rows;
    strip = (unsigned char *) _TIFFmalloc (strip_bytes);
    if (!strip)
	goto error;
    strip_ptr = strip;
    for (row = 0; row < img->height; row++)
      {
	  unsigned char *p_in = img->pixels + (row * img->scanline_width);
	  byte = 0x00;
	  pos = 0;
	  for (col = 0; col < img->width; col++)
	    {
		pixel = img->palette_red[*p_in++];	/* expected to be a PALETTE image */
		if (pixel == 0)
		  {
		      /* handling a black pixel */
		      switch (pos)
			{
			case 0:
			    byte |= 0x80;
			    break;
			case 1:
			    byte |= 0x40;
			    break;
			case 2:
			    byte |= 0x20;
			    break;
			case 3:
			    byte |= 0x10;
			    break;
			case 4:
			    byte |= 0x08;
			    break;
			case 5:
			    byte |= 0x04;
			    break;
			case 6:
			    byte |= 0x02;
			    break;
			case 7:
			    byte |= 0x01;
			    break;
			};
		  }
		pos++;
		if (pos > 7)
		  {
		      /* exporting an octet */
		      *strip_ptr++ = byte;
		      byte = 0x00;
		      pos = 0;
		  }
	    }
	  if (pos > 0)		/* exporting the last octet */
	      *strip_ptr++ = byte;
	  strip_effective = strip_ptr - strip;
	  if (strip_effective >= strip_bytes)
	    {
		/* writing a strip */
		if (TIFFWriteEncodedStrip (out, nstrip++, strip, strip_bytes) <
		    0)
		    goto error;
		strip_ptr = strip;
	    }
      }
    strip_effective = strip_ptr - strip;
    if (strip_effective)
      {
	  if (TIFFWriteEncodedStrip (out, nstrip++, strip, strip_effective) < 0)
	      goto error;
      }

    _TIFFfree (strip);
    return GGRAPH_OK;

  error:
    if (strip)
	_TIFFfree (strip);
    return GGRAPH_TIFF_CODEC_ERROR;
}

static int
common_tiff_fax_tile (TIFF * out, const gGraphImagePtr img, int tile_width,
		      int tile_height)
{
/* common function writing a tiff raster [monochrome/tile] */
    int row;
    int col;
    tsize_t scan_bytes;
    tsize_t tile_bytes;
    unsigned char *tile = NULL;
    unsigned char *tile_ptr;
    int pixel;
    uint32 tile_x;
    uint32 tile_y;
    int num_rows;
    int num_cols;
    unsigned char byte;
    int pos;

/* allocating the scan-line buffer */
    tile_bytes = TIFFTileSize (out);
    scan_bytes = tile_bytes / tile_height;
    tile = (unsigned char *) _TIFFmalloc (tile_bytes);
    if (!tile)
	goto error;
    for (tile_y = 0; tile_y < (uint32) (img->height); tile_y += tile_height)
      {
	  if (tile_y + tile_height > (uint32) (img->height))
	      num_rows = img->height - tile_y;
	  else
	      num_rows = tile_height;
	  for (tile_x = 0; tile_x < (uint32) (img->width); tile_x += tile_width)
	    {
		/* feeding a tile */
		if (tile_x + tile_width > (uint32) (img->width))
		    num_cols = img->width - tile_x;
		else
		    num_cols = tile_width;
		for (row = 0; row < num_rows; row++)
		  {
		      byte = 0x00;
		      pos = 0;
		      tile_ptr = tile + (scan_bytes * row);
		      for (col = 0; col < num_cols; col++)
			{
			    unsigned char *p_in =
				img->pixels +
				((tile_y + row) * img->scanline_width) +
				(tile_x + col);
			    pixel = img->palette_red[*p_in];	/* expected to be a PALETTE image */
			    if (pixel == 0)
			      {
				  /* handling a black pixel */
				  switch (pos)
				    {
				    case 0:
					byte |= 0x80;
					break;
				    case 1:
					byte |= 0x40;
					break;
				    case 2:
					byte |= 0x20;
					break;
				    case 3:
					byte |= 0x10;
					break;
				    case 4:
					byte |= 0x08;
					break;
				    case 5:
					byte |= 0x04;
					break;
				    case 6:
					byte |= 0x02;
					break;
				    case 7:
					byte |= 0x01;
					break;
				    };
			      }
			    pos++;
			    if (pos > 7)
			      {
				  /* exporting an octet */
				  *tile_ptr++ = byte;
				  byte = 0x00;
				  pos = 0;
			      }
			}
		      if (pos > 0)	/* exporting the last octet */
			  *tile_ptr++ = byte;
		  }
		if (TIFFWriteTile (out, tile, tile_x, tile_y, 0, 0) < 0)
		    goto error;
	    }
      }

    _TIFFfree (tile);
    return GGRAPH_OK;

  error:
    if (tile)
	_TIFFfree (tile);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_tiff_fax_strip (const gGraphImagePtr img, void **mem_buf,
			    int *mem_buf_size, FILE * file, int dest_type,
			    int rows_per_strip, int compression)
{
/* exporting a bi-level (monochrome) image [by strips] */
    TIFF *out = (TIFF *) 0;
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			      file_writeproc, file_seekproc, closeproc,
			      file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX3
	      || compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX4)
	      clientdata.malloc_block = (img->width * img->height) / 800;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 80;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			      memory_writeproc, memory_seekproc, closeproc,
			      memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 1);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_FILLORDER, FILLORDER_MSB2LSB);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISWHITE);
    if (compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX3)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_CCITTFAX3);
    else if (compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX4)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_CCITTFAX4);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_ROWSPERSTRIP, rows_per_strip);

    if (common_tiff_fax_strip (out, img, rows_per_strip) != GGRAPH_OK)
	goto error;

    TIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (out)
	TIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_tiff_fax_tile (const gGraphImagePtr img, void **mem_buf,
			   int *mem_buf_size, FILE * file, int dest_type,
			   int tile_width, int tile_height, int compression)
{
/* exporting a bi-level (monochrome) image [by tiles] */
    TIFF *out = (TIFF *) 0;
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			      file_writeproc, file_seekproc, closeproc,
			      file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX3
	      || compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX4)
	      clientdata.malloc_block = (img->width * img->height) / 800;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 80;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			      memory_writeproc, memory_seekproc, closeproc,
			      memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 1);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_FILLORDER, FILLORDER_MSB2LSB);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISWHITE);
    if (compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX3)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_CCITTFAX3);
    else if (compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX4)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_CCITTFAX4);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_TILEWIDTH, tile_width);
    TIFFSetField (out, TIFFTAG_TILELENGTH, tile_height);

    if (common_tiff_fax_tile (out, img, tile_width, tile_height) != GGRAPH_OK)
	goto error;

    TIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (out)
	TIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

static int
is_projected_srs (const char *proj4text)
{
/* checks if this one is a PCS SRS */
    if (proj4text == NULL)
	return 1;
    if (strlen (proj4text) > 14)
      {
	  if (strncmp (proj4text, "+proj=longlat ", 14) == 0)
	      return 0;
      }
    return 1;
}

extern int
gg_image_to_geo_tiff_fax_strip (const gGraphImagePtr img, void **mem_buf,
				int *mem_buf_size, FILE * file, int dest_type,
				int rows_per_strip, int compression)
{
/* exporting a bi-level (monochrome) image [by strips] */
    TIFF *out = (TIFF *) 0;
    GTIF *gtif = (GTIF *) 0;
    double tiepoint[6];
    double pixsize[3];
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (img->is_georeferenced == 0)
	return GGRAPH_MISSING_GEO_DEFS;

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			       file_writeproc, file_seekproc, closeproc,
			       file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX3
	      || compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX4)
	      clientdata.malloc_block = (img->width * img->height) / 800;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 80;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			       memory_writeproc, memory_seekproc, closeproc,
			       memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;
    gtif = GTIFNew (out);
    if (!gtif)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 1);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_FILLORDER, FILLORDER_MSB2LSB);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISWHITE);
    if (compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX3)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_CCITTFAX3);
    else if (compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX4)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_CCITTFAX4);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_ROWSPERSTRIP, rows_per_strip);

/* setting up the GeoTIFF Tags */
    pixsize[0] = img->pixel_x_size;
    pixsize[1] = img->pixel_y_size;
    pixsize[2] = 0.0;
    TIFFSetField (out, GTIFF_PIXELSCALE, 3, pixsize);
    tiepoint[0] = 0.0;
    tiepoint[1] = 0.0;
    tiepoint[2] = 0.0;
    tiepoint[3] = img->upper_left_x;
    tiepoint[4] = img->upper_left_y;
    tiepoint[5] = 0.0;
    TIFFSetField (out, GTIFF_TIEPOINTS, 6, tiepoint);
    if (img->srs_name != NULL)
	TIFFSetField (out, GTIFF_ASCIIPARAMS, img->srs_name);
    if (img->proj4text != NULL)
	GTIFSetFromProj4 (gtif, img->proj4text);
    if (img->srs_name != NULL)
	GTIFKeySet (gtif, GTCitationGeoKey, TYPE_ASCII, 0, img->srs_name);
    if (is_projected_srs (img->proj4text))
	GTIFKeySet (gtif, ProjectedCSTypeGeoKey, TYPE_SHORT, 1, img->srid);
    GTIFWriteKeys (gtif);

    if (common_tiff_fax_strip (out, img, rows_per_strip) != GGRAPH_OK)
	goto error;

    GTIFFree (gtif);
    XTIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (gtif)
	GTIFFree (gtif);
    if (out)
	XTIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_geo_tiff_fax_tile (const gGraphImagePtr img, void **mem_buf,
			       int *mem_buf_size, FILE * file, int dest_type,
			       int tile_width, int tile_height, int compression)
{
/* exporting a bi-level (monochrome) image [by tiles] */
    TIFF *out = (TIFF *) 0;
    GTIF *gtif = (GTIF *) 0;
    double tiepoint[6];
    double pixsize[3];
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (img->is_georeferenced == 0)
	return GGRAPH_MISSING_GEO_DEFS;

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			       file_writeproc, file_seekproc, closeproc,
			       file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX3
	      || compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX4)
	      clientdata.malloc_block = (img->width * img->height) / 800;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 80;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			       memory_writeproc, memory_seekproc, closeproc,
			       memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;
    gtif = GTIFNew (out);
    if (!gtif)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 1);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_FILLORDER, FILLORDER_MSB2LSB);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISWHITE);
    if (compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX3)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_CCITTFAX3);
    else if (compression == GGRAPH_TIFF_COMPRESSION_CCITTFAX4)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_CCITTFAX4);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_TILEWIDTH, tile_width);
    TIFFSetField (out, TIFFTAG_TILELENGTH, tile_height);

/* setting up the GeoTIFF Tags */
    pixsize[0] = img->pixel_x_size;
    pixsize[1] = img->pixel_y_size;
    pixsize[2] = 0.0;
    TIFFSetField (out, GTIFF_PIXELSCALE, 3, pixsize);
    tiepoint[0] = 0.0;
    tiepoint[1] = 0.0;
    tiepoint[2] = 0.0;
    tiepoint[3] = img->upper_left_x;
    tiepoint[4] = img->upper_left_y;
    tiepoint[5] = 0.0;
    TIFFSetField (out, GTIFF_TIEPOINTS, 6, tiepoint);
    if (img->srs_name != NULL)
	TIFFSetField (out, GTIFF_ASCIIPARAMS, img->srs_name);
    if (img->proj4text != NULL)
	GTIFSetFromProj4 (gtif, img->proj4text);
    if (img->srs_name != NULL)
	GTIFKeySet (gtif, GTCitationGeoKey, TYPE_ASCII, 0, img->srs_name);
    if (is_projected_srs (img->proj4text))
	GTIFKeySet (gtif, ProjectedCSTypeGeoKey, TYPE_SHORT, 1, img->srid);
    GTIFWriteKeys (gtif);

    if (common_tiff_fax_tile (out, img, tile_width, tile_height) != GGRAPH_OK)
	goto error;

    GTIFFree (gtif);
    XTIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (gtif)
	GTIFFree (gtif);
    if (out)
	XTIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

static int
common_tiff_palette_strip (TIFF * out, const gGraphImagePtr img)
{
/* common function writing a tiff raster [palette/strip] */
    int row;
    int col;
    tsize_t line_bytes;
    unsigned char *scanline = NULL;
    unsigned char *line_ptr;

/* allocating the scan-line buffer */
    line_bytes = img->width;
    scanline = (unsigned char *) _TIFFmalloc (line_bytes);
    if (!scanline)
	goto error;
    for (row = 0; row < img->height; row++)
      {
	  unsigned char *p_in = img->pixels + (row * img->scanline_width);
	  line_ptr = scanline;
	  for (col = 0; col < img->width; col++)
	      *line_ptr++ = *p_in++;	/* expected to be a PALETTE image */
	  if (TIFFWriteScanline (out, scanline, row, 0) < 0)
	      goto error;
      }

    _TIFFfree (scanline);
    return GGRAPH_OK;

  error:
    if (scanline)
	_TIFFfree (scanline);
    return GGRAPH_TIFF_CODEC_ERROR;
}

static int
common_tiff_palette_tile (TIFF * out, const gGraphImagePtr img, int tile_width,
			  int tile_height)
{
/* common function writing a tiff raster [palette/tile] */
    int row;
    int col;
    unsigned char *tile = NULL;
    unsigned char *line_ptr;
    uint32 tile_x;
    uint32 tile_y;
    int num_rows;
    int num_cols;

/* allocating the tile buffer */
    tile = (unsigned char *) _TIFFmalloc (tile_width * tile_height);
    if (!tile)
	goto error;
    for (tile_y = 0; tile_y < (uint32) (img->height); tile_y += tile_height)
      {
	  if (tile_y + tile_height > (uint32) (img->height))
	      num_rows = img->height - tile_y;
	  else
	      num_rows = tile_height;
	  for (tile_x = 0; tile_x < (uint32) (img->width); tile_x += tile_width)
	    {
		/* feeding a tile */
		if (tile_x + tile_width > (uint32) (img->width))
		    num_cols = img->width - tile_x;
		else
		    num_cols = tile_width;
		for (row = 0; row < num_rows; row++)
		  {
		      line_ptr = tile + (row * tile_width);
		      for (col = 0; col < num_cols; col++)
			{
			    unsigned char *p_in =
				img->pixels +
				((tile_y + row) * img->scanline_width) +
				(tile_x + col);
			    *line_ptr++ = *p_in;	/* expected to be a PALETTE image */
			}
		  }
		if (TIFFWriteTile (out, tile, tile_x, tile_y, 0, 0) < 0)
		    goto error;
	    }
      }
    _TIFFfree (tile);
    return GGRAPH_OK;

  error:
    if (tile)
	_TIFFfree (tile);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_tiff_palette_strip (const gGraphImagePtr img, void **mem_buf,
				int *mem_buf_size, FILE * file, int dest_type,
				int rows_per_strip, int compression)
{
/* exporting a palettte image [by strips] */
    TIFF *out = (TIFF *) 0;
    uint16 red[256];
    uint16 green[256];
    uint16 blue[256];
    int col;
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			      file_writeproc, file_seekproc, closeproc,
			      file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			      memory_writeproc, memory_seekproc, closeproc,
			      memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;

/* bulding the palette */
    for (col = 0; col < 256; col++)
      {
	  /* expected to be a PALETTE image */
	  if (col >= img->max_palette)
	    {
		red[col] = 0;
		green[col] = 0;
		blue[col] = 0;
	    }
	  else
	    {
		red[col] = (uint16) (img->palette_red[col] * 256);
		green[col] = (uint16) (img->palette_green[col] * 256);
		blue[col] = (uint16) (img->palette_blue[col] * 256);
	    }
      }
/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
    TIFFSetField (out, TIFFTAG_COLORMAP, red, green, blue);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_ROWSPERSTRIP, rows_per_strip);

    if (common_tiff_palette_strip (out, img) != GGRAPH_OK)
	goto error;

    TIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (out)
	TIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_tiff_palette_tile (const gGraphImagePtr img, void **mem_buf,
			       int *mem_buf_size, FILE * file, int dest_type,
			       int tile_width, int tile_height, int compression)
{
/* exporting a palettte image [by tiles] */
    TIFF *out = (TIFF *) 0;
    int col;
    uint16 red[256];
    uint16 green[256];
    uint16 blue[256];
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			      file_writeproc, file_seekproc, closeproc,
			      file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			      memory_writeproc, memory_seekproc, closeproc,
			      memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;

/* bulding the palette */
    for (col = 0; col < 256; col++)
      {
	  /* expected to be a PALETTE image */
	  if (col >= img->max_palette)
	    {
		red[col] = 0;
		green[col] = 0;
		blue[col] = 0;
	    }
	  else
	    {
		red[col] = (uint16) (img->palette_red[col] * 256);
		green[col] = (uint16) (img->palette_green[col] * 256);
		blue[col] = (uint16) (img->palette_blue[col] * 256);
	    }
      }
/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
    TIFFSetField (out, TIFFTAG_COLORMAP, red, green, blue);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_TILEWIDTH, tile_width);
    TIFFSetField (out, TIFFTAG_TILELENGTH, tile_height);

    if (common_tiff_palette_tile (out, img, tile_width, tile_height) !=
	GGRAPH_OK)
	goto error;

    TIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (out)
	TIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_geo_tiff_palette_strip (const gGraphImagePtr img, void **mem_buf,
				    int *mem_buf_size, FILE * file,
				    int dest_type, int rows_per_strip,
				    int compression)
{
/* exporting a palettte image [by strips] */
    TIFF *out = (TIFF *) 0;
    GTIF *gtif = (GTIF *) 0;
    double tiepoint[6];
    double pixsize[3];
    uint16 red[256];
    uint16 green[256];
    uint16 blue[256];
    int col;
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (img->is_georeferenced == 0)
	return GGRAPH_MISSING_GEO_DEFS;

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			       file_writeproc, file_seekproc, closeproc,
			       file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			       memory_writeproc, memory_seekproc, closeproc,
			       memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;
    gtif = GTIFNew (out);
    if (!gtif)
	goto error;

/* bulding the palette */
    for (col = 0; col < 256; col++)
      {
	  /* expected to be a PALETTE image */
	  if (col >= img->max_palette)
	    {
		red[col] = 0;
		green[col] = 0;
		blue[col] = 0;
	    }
	  else
	    {
		red[col] = (uint16) (img->palette_red[col] * 256);
		green[col] = (uint16) (img->palette_green[col] * 256);
		blue[col] = (uint16) (img->palette_blue[col] * 256);
	    }
      }
/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
    TIFFSetField (out, TIFFTAG_COLORMAP, red, green, blue);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_ROWSPERSTRIP, rows_per_strip);

/* setting up the GeoTIFF Tags */
    pixsize[0] = img->pixel_x_size;
    pixsize[1] = img->pixel_y_size;
    pixsize[2] = 0.0;
    TIFFSetField (out, GTIFF_PIXELSCALE, 3, pixsize);
    tiepoint[0] = 0.0;
    tiepoint[1] = 0.0;
    tiepoint[2] = 0.0;
    tiepoint[3] = img->upper_left_x;
    tiepoint[4] = img->upper_left_y;
    tiepoint[5] = 0.0;
    TIFFSetField (out, GTIFF_TIEPOINTS, 6, tiepoint);
    if (img->srs_name != NULL)
	TIFFSetField (out, GTIFF_ASCIIPARAMS, img->srs_name);
    if (img->proj4text != NULL)
	GTIFSetFromProj4 (gtif, img->proj4text);
    if (img->srs_name != NULL)
	GTIFKeySet (gtif, GTCitationGeoKey, TYPE_ASCII, 0, img->srs_name);
    if (is_projected_srs (img->proj4text))
	GTIFKeySet (gtif, ProjectedCSTypeGeoKey, TYPE_SHORT, 1, img->srid);
    GTIFWriteKeys (gtif);

    if (common_tiff_palette_strip (out, img) != GGRAPH_OK)
	goto error;

    GTIFFree (gtif);
    XTIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (gtif)
	GTIFFree (gtif);
    if (out)
	XTIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_geo_tiff_palette_tile (const gGraphImagePtr img, void **mem_buf,
				   int *mem_buf_size, FILE * file,
				   int dest_type, int tile_width,
				   int tile_height, int compression)
{
/* exporting a palettte image [by tiles] */
    TIFF *out = (TIFF *) 0;
    GTIF *gtif = (GTIF *) 0;
    double tiepoint[6];
    double pixsize[3];
    int col;
    uint16 red[256];
    uint16 green[256];
    uint16 blue[256];
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (img->is_georeferenced == 0)
	return GGRAPH_MISSING_GEO_DEFS;

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			       file_writeproc, file_seekproc, closeproc,
			       file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			       memory_writeproc, memory_seekproc, closeproc,
			       memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;
    gtif = GTIFNew (out);
    if (!gtif)
	goto error;

/* bulding the palette */
    for (col = 0; col < 256; col++)
      {
	  /* expected to be a PALETTE image */
	  if (col >= img->max_palette)
	    {
		red[col] = 0;
		green[col] = 0;
		blue[col] = 0;
	    }
	  else
	    {
		red[col] = (uint16) (img->palette_red[col] * 256);
		green[col] = (uint16) (img->palette_green[col] * 256);
		blue[col] = (uint16) (img->palette_blue[col] * 256);
	    }
      }
/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
    TIFFSetField (out, TIFFTAG_COLORMAP, red, green, blue);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_TILEWIDTH, tile_width);
    TIFFSetField (out, TIFFTAG_TILELENGTH, tile_height);

/* setting up the GeoTIFF Tags */
    pixsize[0] = img->pixel_x_size;
    pixsize[1] = img->pixel_y_size;
    pixsize[2] = 0.0;
    TIFFSetField (out, GTIFF_PIXELSCALE, 3, pixsize);
    tiepoint[0] = 0.0;
    tiepoint[1] = 0.0;
    tiepoint[2] = 0.0;
    tiepoint[3] = img->upper_left_x;
    tiepoint[4] = img->upper_left_y;
    tiepoint[5] = 0.0;
    TIFFSetField (out, GTIFF_TIEPOINTS, 6, tiepoint);
    if (img->srs_name != NULL)
	TIFFSetField (out, GTIFF_ASCIIPARAMS, img->srs_name);
    if (img->proj4text != NULL)
	GTIFSetFromProj4 (gtif, img->proj4text);
    if (img->srs_name != NULL)
	GTIFKeySet (gtif, GTCitationGeoKey, TYPE_ASCII, 0, img->srs_name);
    if (is_projected_srs (img->proj4text))
	GTIFKeySet (gtif, ProjectedCSTypeGeoKey, TYPE_SHORT, 1, img->srid);
    GTIFWriteKeys (gtif);

    if (common_tiff_palette_tile (out, img, tile_width, tile_height) !=
	GGRAPH_OK)
	goto error;

    GTIFFree (gtif);
    XTIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (gtif)
	GTIFFree (gtif);
    if (out)
	XTIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

static int
common_tiff_grayscale_strip (TIFF * out, const gGraphImagePtr img)
{
/* common function writing a tiff raster [grayscale/strip] */
    int row;
    int col;
    size_t line_bytes;
    unsigned char *scanline = NULL;
    unsigned char *line_ptr;

/* allocating the scan-line buffer */
    line_bytes = img->width;
    scanline = (unsigned char *) _TIFFmalloc (line_bytes);
    if (!scanline)
	goto error;
    for (row = 0; row < img->height; row++)
      {
	  unsigned char *p_in = img->pixels + (row * img->scanline_width);
	  line_ptr = scanline;
	  for (col = 0; col < img->width; col++)
	      *line_ptr++ = *p_in++;	/* expected to be a GRAYSCALE image */
	  if (TIFFWriteScanline (out, scanline, row, 0) < 0)
	      goto error;
      }
    _TIFFfree (scanline);
    return GGRAPH_OK;

  error:
    if (scanline)
	_TIFFfree (scanline);
    return GGRAPH_TIFF_CODEC_ERROR;
}

static int
common_tiff_grayscale_tile (TIFF * out, const gGraphImagePtr img,
			    int tile_width, int tile_height)
{
/* common function writing a tiff raster [grayscale/tile] */
    int row;
    int col;
    unsigned char *tile = NULL;
    unsigned char *line_ptr;
    uint32 tile_x;
    uint32 tile_y;
    int num_rows;
    int num_cols;

/* allocating the tile buffer */
    tile = (unsigned char *) _TIFFmalloc (tile_width * tile_height);
    if (!tile)
	goto error;
    for (tile_y = 0; tile_y < (uint32) (img->height); tile_y += tile_height)
      {
	  if (tile_y + tile_height > (uint32) (img->height))
	      num_rows = img->height - tile_y;
	  else
	      num_rows = tile_height;
	  for (tile_x = 0; tile_x < (uint32) (img->width); tile_x += tile_width)
	    {
		/* feeding a tile */
		if (tile_x + tile_width > (uint32) (img->width))
		    num_cols = img->width - tile_x;
		else
		    num_cols = tile_width;
		for (row = 0; row < num_rows; row++)
		  {
		      line_ptr = tile + (row * tile_width);
		      for (col = 0; col < num_cols; col++)
			{
			    unsigned char *p_in =
				img->pixels +
				((tile_y + row) * img->scanline_width) +
				(tile_x + col);
			    *line_ptr++ = *p_in;	/* expected to be a GRAYSCALE image */
			}
		  }
		if (TIFFWriteTile (out, tile, tile_x, tile_y, 0, 0) < 0)
		    goto error;
	    }
      }
    _TIFFfree (tile);
    return GGRAPH_OK;

  error:
    if (tile)
	_TIFFfree (tile);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_tiff_grayscale_strip (const gGraphImagePtr img, void **mem_buf,
				  int *mem_buf_size, FILE * file, int dest_type,
				  int rows_per_strip, int compression)
{
/* compressing a grayscale image [by strips] */
    TIFF *out = (TIFF *) 0;
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			      file_writeproc, file_seekproc, closeproc,
			      file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE
	      || compression == GGRAPH_TIFF_COMPRESSION_JPEG)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			      memory_writeproc, memory_seekproc, closeproc,
			      memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else if (compression == GGRAPH_TIFF_COMPRESSION_JPEG
	     && (rows_per_strip % 8) == 0)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_JPEG);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_ROWSPERSTRIP, rows_per_strip);

    if (common_tiff_grayscale_strip (out, img) != GGRAPH_OK)
	goto error;

    TIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (out)
	TIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_tiff_grayscale_tile (const gGraphImagePtr img, void **mem_buf,
				 int *mem_buf_size, FILE * file, int dest_type,
				 int tile_width, int tile_height,
				 int compression)
{
/* compressing a grayscale image [by tiles] */
    TIFF *out = (TIFF *) 0;
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			      file_writeproc, file_seekproc, closeproc,
			      file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE
	      || compression == GGRAPH_TIFF_COMPRESSION_JPEG)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			      memory_writeproc, memory_seekproc, closeproc,
			      memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else if (compression == GGRAPH_TIFF_COMPRESSION_JPEG)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_JPEG);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_TILEWIDTH, tile_width);
    TIFFSetField (out, TIFFTAG_TILELENGTH, tile_height);

    if (common_tiff_grayscale_tile (out, img, tile_width, tile_height) !=
	GGRAPH_OK)
	goto error;

    TIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (out)
	TIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_geo_tiff_grayscale_strip (const gGraphImagePtr img, void **mem_buf,
				      int *mem_buf_size, FILE * file,
				      int dest_type, int rows_per_strip,
				      int compression)
{
/* compressing a grayscale image [by strips] */
    TIFF *out = (TIFF *) 0;
    GTIF *gtif = (GTIF *) 0;
    double tiepoint[6];
    double pixsize[3];
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (img->is_georeferenced == 0)
	return GGRAPH_MISSING_GEO_DEFS;

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			       file_writeproc, file_seekproc, closeproc,
			       file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE
	      || compression == GGRAPH_TIFF_COMPRESSION_JPEG)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			       memory_writeproc, memory_seekproc, closeproc,
			       memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;
    gtif = GTIFNew (out);
    if (!gtif)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else if (compression == GGRAPH_TIFF_COMPRESSION_JPEG
	     && (rows_per_strip % 8) == 0)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_JPEG);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_ROWSPERSTRIP, rows_per_strip);

/* setting up the GeoTIFF Tags */
    pixsize[0] = img->pixel_x_size;
    pixsize[1] = img->pixel_y_size;
    pixsize[2] = 0.0;
    TIFFSetField (out, GTIFF_PIXELSCALE, 3, pixsize);
    tiepoint[0] = 0.0;
    tiepoint[1] = 0.0;
    tiepoint[2] = 0.0;
    tiepoint[3] = img->upper_left_x;
    tiepoint[4] = img->upper_left_y;
    tiepoint[5] = 0.0;
    TIFFSetField (out, GTIFF_TIEPOINTS, 6, tiepoint);
    if (img->srs_name != NULL)
	TIFFSetField (out, GTIFF_ASCIIPARAMS, img->srs_name);
    if (img->proj4text != NULL)
	GTIFSetFromProj4 (gtif, img->proj4text);
    if (img->srs_name != NULL)
	GTIFKeySet (gtif, GTCitationGeoKey, TYPE_ASCII, 0, img->srs_name);
    if (is_projected_srs (img->proj4text))
	GTIFKeySet (gtif, ProjectedCSTypeGeoKey, TYPE_SHORT, 1, img->srid);
    GTIFWriteKeys (gtif);

    if (common_tiff_grayscale_strip (out, img) != GGRAPH_OK)
	goto error;

    GTIFFree (gtif);
    XTIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (gtif)
	GTIFFree (gtif);
    if (out)
	XTIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_geo_tiff_grayscale_tile (const gGraphImagePtr img, void **mem_buf,
				     int *mem_buf_size, FILE * file,
				     int dest_type, int tile_width,
				     int tile_height, int compression)
{
/* compressing a grayscale image [by tiles] */
    TIFF *out = (TIFF *) 0;
    GTIF *gtif = (GTIF *) 0;
    double tiepoint[6];
    double pixsize[3];
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (img->is_georeferenced == 0)
	return GGRAPH_MISSING_GEO_DEFS;

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			       file_writeproc, file_seekproc, closeproc,
			       file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE
	      || compression == GGRAPH_TIFF_COMPRESSION_JPEG)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			       memory_writeproc, memory_seekproc, closeproc,
			       memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;
    gtif = GTIFNew (out);
    if (!gtif)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else if (compression == GGRAPH_TIFF_COMPRESSION_JPEG)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_JPEG);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_TILEWIDTH, tile_width);
    TIFFSetField (out, TIFFTAG_TILELENGTH, tile_height);

/* setting up the GeoTIFF Tags */
    pixsize[0] = img->pixel_x_size;
    pixsize[1] = img->pixel_y_size;
    pixsize[2] = 0.0;
    TIFFSetField (out, GTIFF_PIXELSCALE, 3, pixsize);
    tiepoint[0] = 0.0;
    tiepoint[1] = 0.0;
    tiepoint[2] = 0.0;
    tiepoint[3] = img->upper_left_x;
    tiepoint[4] = img->upper_left_y;
    tiepoint[5] = 0.0;
    TIFFSetField (out, GTIFF_TIEPOINTS, 6, tiepoint);
    if (img->srs_name != NULL)
	TIFFSetField (out, GTIFF_ASCIIPARAMS, img->srs_name);
    if (img->proj4text != NULL)
	GTIFSetFromProj4 (gtif, img->proj4text);
    if (img->srs_name != NULL)
	GTIFKeySet (gtif, GTCitationGeoKey, TYPE_ASCII, 0, img->srs_name);
    if (is_projected_srs (img->proj4text))
	GTIFKeySet (gtif, ProjectedCSTypeGeoKey, TYPE_SHORT, 1, img->srid);
    GTIFWriteKeys (gtif);

    if (common_tiff_grayscale_tile (out, img, tile_width, tile_height) !=
	GGRAPH_OK)
	goto error;

    GTIFFree (gtif);
    XTIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (gtif)
	GTIFFree (gtif);
    if (out)
	XTIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

static int
common_tiff_rgb_strip (TIFF * out, const gGraphImagePtr img)
{
/* common function writing a tiff raster [rgb/strip] */
    int row;
    int col;
    tsize_t line_bytes;
    unsigned char *scanline = NULL;
    unsigned char *line_ptr;

/* allocating the scan-line buffer */
    line_bytes = img->width * 3;
    scanline = (unsigned char *) _TIFFmalloc (line_bytes);
    if (!scanline)
	goto error;
    for (row = 0; row < img->height; row++)
      {
	  unsigned char *p_in = img->pixels + (row * img->scanline_width);
	  unsigned char r;
	  unsigned char g;
	  unsigned char b;
	  line_ptr = scanline;
	  for (col = 0; col < img->width; col++)
	    {
		if (img->pixel_format == GG_PIXEL_RGB)
		  {
		      r = *p_in++;
		      g = *p_in++;
		      b = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_RGBA)
		  {
		      r = *p_in++;
		      g = *p_in++;
		      b = *p_in++;
		      p_in++;	/* skipping Alpha */
		  }
		else if (img->pixel_format == GG_PIXEL_ARGB)
		  {
		      p_in++;	/* skipping Alpha */
		      r = *p_in++;
		      g = *p_in++;
		      b = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGR)
		  {
		      b = *p_in++;
		      g = *p_in++;
		      r = *p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_BGRA)
		  {
		      b = *p_in++;
		      g = *p_in++;
		      r = *p_in++;
		      p_in++;
		  }
		else if (img->pixel_format == GG_PIXEL_GDI)
		  {
		      b = *p_in++;
		      g = *p_in++;
		      r = *p_in++;
		  }
		*line_ptr++ = r;
		*line_ptr++ = g;
		*line_ptr++ = b;
	    }
	  if (TIFFWriteScanline (out, scanline, row, 0) < 0)
	      goto error;
      }
    _TIFFfree (scanline);
    return GGRAPH_OK;

  error:
    if (scanline)
	_TIFFfree (scanline);
    return GGRAPH_TIFF_CODEC_ERROR;
}

static int
common_tiff_rgb_tile (TIFF * out, const gGraphImagePtr img, int tile_width,
		      int tile_height)
{
/* common function writing a tiff raster [rgb/tile] */
    int row;
    int col;
    unsigned char *tile = NULL;
    unsigned char *line_ptr;
    uint32 tile_x;
    uint32 tile_y;
    int num_rows;
    int num_cols;

/* allocating the tile buffer */
    tile = (unsigned char *) _TIFFmalloc (tile_width * tile_height * 3);
    if (!tile)
	goto error;
    for (tile_y = 0; tile_y < (uint32) (img->height); tile_y += tile_height)
      {
	  if (tile_y + tile_height > (uint32) (img->height))
	      num_rows = img->height - tile_y;
	  else
	      num_rows = tile_height;
	  for (tile_x = 0; tile_x < (uint32) (img->width); tile_x += tile_width)
	    {
		/* feeding a tile */
		if (tile_x + tile_width > (uint32) (img->width))
		    num_cols = img->width - tile_x;
		else
		    num_cols = tile_width;
		for (row = 0; row < num_rows; row++)
		  {
		      unsigned char *p_in;
		      unsigned char r;
		      unsigned char g;
		      unsigned char b;
		      line_ptr = tile + (row * tile_width * 3);
		      for (col = 0; col < num_cols; col++)
			{
			    p_in =
				img->pixels +
				((tile_y + row) * img->scanline_width) +
				((tile_x + col) * img->pixel_size);
			    if (img->pixel_format == GG_PIXEL_RGB)
			      {
				  r = *p_in++;
				  g = *p_in++;
				  b = *p_in++;
			      }
			    else if (img->pixel_format == GG_PIXEL_RGBA)
			      {
				  r = *p_in++;
				  g = *p_in++;
				  b = *p_in++;
				  p_in++;	/* skipping Alpha */
			      }
			    else if (img->pixel_format == GG_PIXEL_ARGB)
			      {
				  p_in++;	/* skipping Alpha */
				  r = *p_in++;
				  g = *p_in++;
				  b = *p_in++;
			      }
			    else if (img->pixel_format == GG_PIXEL_BGR)
			      {
				  b = *p_in++;
				  g = *p_in++;
				  r = *p_in++;
			      }
			    else if (img->pixel_format == GG_PIXEL_BGRA)
			      {
				  b = *p_in++;
				  g = *p_in++;
				  r = *p_in++;
				  p_in++;	/* skipping Alpha */
			      }
			    else if (img->pixel_format == GG_PIXEL_GDI)
			      {
				  b = *p_in++;
				  g = *p_in++;
				  r = *p_in++;
			      }
			    *line_ptr++ = r;
			    *line_ptr++ = g;
			    *line_ptr++ = b;
			}
		  }
		if (TIFFWriteTile (out, tile, tile_x, tile_y, 0, 0) < 0)
		    goto error;
	    }
      }
    _TIFFfree (tile);
    return GGRAPH_OK;

  error:
    if (tile)
	_TIFFfree (tile);
    return GGRAPH_TIFF_CODEC_ERROR;

}

extern int
gg_image_to_tiff_rgb_strip (const gGraphImagePtr img, void **mem_buf,
			    int *mem_buf_size, FILE * file, int dest_type,
			    int rows_per_strip, int compression)
{
/* exporting an RGB image [by strips] */
    TIFF *out = (TIFF *) 0;
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			      file_writeproc, file_seekproc, closeproc,
			      file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE
	      || compression == GGRAPH_TIFF_COMPRESSION_JPEG)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			      memory_writeproc, memory_seekproc, closeproc,
			      memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 3);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else if (compression == GGRAPH_TIFF_COMPRESSION_JPEG
	     && (rows_per_strip % 8) == 0)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_JPEG);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_ROWSPERSTRIP, rows_per_strip);

    if (common_tiff_rgb_strip (out, img) != GGRAPH_OK)
	goto error;

    TIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (out)
	TIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_tiff_rgb_tile (const gGraphImagePtr img, void **mem_buf,
			   int *mem_buf_size, FILE * file, int dest_type,
			   int tile_width, int tile_height, int compression)
{
/* exporting an RGB image [by tiles] */
    TIFF *out = (TIFF *) 0;
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			      file_writeproc, file_seekproc, closeproc,
			      file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE
	      || compression == GGRAPH_TIFF_COMPRESSION_JPEG)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			      memory_writeproc, memory_seekproc, closeproc,
			      memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 3);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else if (compression == GGRAPH_TIFF_COMPRESSION_JPEG)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_JPEG);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_TILEWIDTH, tile_width);
    TIFFSetField (out, TIFFTAG_TILELENGTH, tile_height);

    if (common_tiff_rgb_tile (out, img, tile_width, tile_height) != GGRAPH_OK)
	goto error;

    TIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (out)
	TIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_geo_tiff_rgb_strip (const gGraphImagePtr img, void **mem_buf,
				int *mem_buf_size, FILE * file, int dest_type,
				int rows_per_strip, int compression)
{
/* exporting an RGB image [by strips] */
    TIFF *out = (TIFF *) 0;
    GTIF *gtif = (GTIF *) 0;
    double tiepoint[6];
    double pixsize[3];
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (img->is_georeferenced == 0)
	return GGRAPH_MISSING_GEO_DEFS;

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			       file_writeproc, file_seekproc, closeproc,
			       file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE
	      || compression == GGRAPH_TIFF_COMPRESSION_JPEG)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			       memory_writeproc, memory_seekproc, closeproc,
			       memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;
    gtif = GTIFNew (out);
    if (!gtif)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 3);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else if (compression == GGRAPH_TIFF_COMPRESSION_JPEG
	     && (rows_per_strip % 8) == 0)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_JPEG);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_ROWSPERSTRIP, rows_per_strip);

/* setting up the GeoTIFF Tags */
    pixsize[0] = img->pixel_x_size;
    pixsize[1] = img->pixel_y_size;
    pixsize[2] = 0.0;
    TIFFSetField (out, GTIFF_PIXELSCALE, 3, pixsize);
    tiepoint[0] = 0.0;
    tiepoint[1] = 0.0;
    tiepoint[2] = 0.0;
    tiepoint[3] = img->upper_left_x;
    tiepoint[4] = img->upper_left_y;
    tiepoint[5] = 0.0;
    TIFFSetField (out, GTIFF_TIEPOINTS, 6, tiepoint);
    if (img->srs_name != NULL)
	TIFFSetField (out, GTIFF_ASCIIPARAMS, img->srs_name);
    if (img->proj4text != NULL)
	GTIFSetFromProj4 (gtif, img->proj4text);
    if (img->srs_name != NULL)
	GTIFKeySet (gtif, GTCitationGeoKey, TYPE_ASCII, 0, img->srs_name);
    if (is_projected_srs (img->proj4text))
	GTIFKeySet (gtif, ProjectedCSTypeGeoKey, TYPE_SHORT, 1, img->srid);
    GTIFWriteKeys (gtif);

    if (common_tiff_rgb_strip (out, img) != GGRAPH_OK)
	goto error;

    GTIFFree (gtif);
    XTIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (gtif)
	GTIFFree (gtif);
    if (out)
	XTIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_geo_tiff_rgb_tile (const gGraphImagePtr img, void **mem_buf,
			       int *mem_buf_size, FILE * file, int dest_type,
			       int tile_width, int tile_height, int compression)
{
/* exporting an RGB image [by tiles] */
    TIFF *out = (TIFF *) 0;
    GTIF *gtif = (GTIF *) 0;
    double tiepoint[6];
    double pixsize[3];
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (img->is_georeferenced == 0)
	return GGRAPH_MISSING_GEO_DEFS;

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			       file_writeproc, file_seekproc, closeproc,
			       file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE
	      || compression == GGRAPH_TIFF_COMPRESSION_JPEG)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			       memory_writeproc, memory_seekproc, closeproc,
			       memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;
    gtif = GTIFNew (out);
    if (!gtif)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 3);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else if (compression == GGRAPH_TIFF_COMPRESSION_JPEG)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_JPEG);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_TILEWIDTH, tile_width);
    TIFFSetField (out, TIFFTAG_TILELENGTH, tile_height);

/* setting up the GeoTIFF Tags */
    pixsize[0] = img->pixel_x_size;
    pixsize[1] = img->pixel_y_size;
    pixsize[2] = 0.0;
    TIFFSetField (out, GTIFF_PIXELSCALE, 3, pixsize);
    tiepoint[0] = 0.0;
    tiepoint[1] = 0.0;
    tiepoint[2] = 0.0;
    tiepoint[3] = img->upper_left_x;
    tiepoint[4] = img->upper_left_y;
    tiepoint[5] = 0.0;
    TIFFSetField (out, GTIFF_TIEPOINTS, 6, tiepoint);
    if (img->srs_name != NULL)
	TIFFSetField (out, GTIFF_ASCIIPARAMS, img->srs_name);
    if (img->proj4text != NULL)
	GTIFSetFromProj4 (gtif, img->proj4text);
    if (img->srs_name != NULL)
	GTIFKeySet (gtif, GTCitationGeoKey, TYPE_ASCII, 0, img->srs_name);
    if (is_projected_srs (img->proj4text))
	GTIFKeySet (gtif, ProjectedCSTypeGeoKey, TYPE_SHORT, 1, img->srid);
    GTIFWriteKeys (gtif);

    if (common_tiff_rgb_tile (out, img, tile_width, tile_height) != GGRAPH_OK)
	goto error;

    GTIFFree (gtif);
    XTIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (gtif)
	GTIFFree (gtif);
    if (out)
	XTIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

static int
common_tiff_grid_strip (TIFF * out, const gGraphImagePtr img)
{
/* common function writing a tiff raster [grid/strip] */
    int row;
    int col;
    tsize_t line_bytes;
    unsigned char *scanline = NULL;
    unsigned char *p_out;
    int i;

/* allocating the scan-line buffer */
    line_bytes = img->width * img->pixel_size;
    scanline = (unsigned char *) _TIFFmalloc (line_bytes);
    if (!scanline)
	goto error;
    for (row = 0; row < img->height; row++)
      {
	  unsigned char *p_in = img->pixels + (row * img->scanline_width);
	  p_out = scanline;
	  for (col = 0; col < img->width; col++)
	    {
		for (i = 0; i < img->pixel_size; i++)
		    *p_out++ = *p_in++;
	    }
	  if (TIFFWriteScanline (out, scanline, row, 0) < 0)
	      goto error;
      }
    _TIFFfree (scanline);
    return GGRAPH_OK;

  error:
    if (scanline)
	_TIFFfree (scanline);
    return GGRAPH_TIFF_CODEC_ERROR;
}

static int
common_tiff_grid_tile (TIFF * out, const gGraphImagePtr img, int tile_width,
		       int tile_height)
{
/* common function writing a tiff raster [rgb/grid] */
    int row;
    int col;
    unsigned char *tile = NULL;
    unsigned char *p_out;
    int i;
    int y_row;
    uint32 tile_x;
    uint32 tile_y;
    int num_rows;
    int num_cols;
    int ret;

/* allocating the tile buffer */
    tile =
	(unsigned char *) _TIFFmalloc (tile_width * tile_height *
				       img->pixel_size);
    if (!tile)
      {
	  ret = GGRAPH_INSUFFICIENT_MEMORY;
	  goto error;
      }

    ret = GGRAPH_TIFF_CODEC_ERROR;
    for (tile_y = 0; tile_y < (uint32) (img->height); tile_y += tile_height)
      {
	  if (tile_y + tile_height > (uint32) (img->height))
	      num_rows = img->height - tile_y;
	  else
	      num_rows = tile_height;
	  for (tile_x = 0; tile_x < (uint32) (img->width); tile_x += tile_width)
	    {
		/* feeding a tile */
		if (tile_x + tile_width > (uint32) (img->width))
		    num_cols = img->width - tile_x;
		else
		    num_cols = tile_width;
		y_row = num_rows - 1;
		for (row = 0; row < num_rows; row++, y_row--)
		  {
		      unsigned char *p_in =
			  img->pixels + ((tile_y + row) * img->scanline_width) +
			  (tile_x * img->pixel_size);
		      p_out = tile + (row * tile_width * img->pixel_size);
		      for (col = 0; col < num_cols; col++)
			{
			    for (i = 0; i < img->pixel_size; i++)
				*p_out++ = *p_in++;
			}
		  }
		if (TIFFWriteTile (out, tile, tile_x, tile_y, 0, 0) < 0)
		    goto error;
	    }
      }
    _TIFFfree (tile);
    return GGRAPH_OK;

  error:
    if (tile)
	_TIFFfree (tile);
    return ret;
}

extern int
gg_image_to_tiff_grid_strip (const gGraphImagePtr img, void **mem_buf,
			     int *mem_buf_size, FILE * file, int dest_type,
			     int rows_per_strip, int compression)
{
/* exporting a GRID image [by strips] */
    TIFF *out = (TIFF *) 0;
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			      file_writeproc, file_seekproc, closeproc,
			      file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			      memory_writeproc, memory_seekproc, closeproc,
			      memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    if (img->sample_format == GGRAPH_SAMPLE_INT)
	TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    else if (img->sample_format == GGRAPH_SAMPLE_FLOAT)
	TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_IEEEFP);
    else
	TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_INT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, img->bits_per_sample);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_ROWSPERSTRIP, rows_per_strip);

    if (common_tiff_grid_strip (out, img) != GGRAPH_OK)
	goto error;

    TIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (out)
	TIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_tiff_grid_tile (const gGraphImagePtr img, void **mem_buf,
			    int *mem_buf_size, FILE * file, int dest_type,
			    int tile_width, int tile_height, int compression)
{
/* exporting a GRID image [by tiles] */
    TIFF *out = (TIFF *) 0;
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			      file_writeproc, file_seekproc, closeproc,
			      file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      TIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			      memory_writeproc, memory_seekproc, closeproc,
			      memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    if (img->sample_format == GGRAPH_SAMPLE_INT)
	TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_INT);
    else if (img->sample_format == GGRAPH_SAMPLE_FLOAT)
	TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_IEEEFP);
    else
	TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, img->bits_per_sample);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_TILEWIDTH, tile_width);
    TIFFSetField (out, TIFFTAG_TILELENGTH, tile_height);

    if (common_tiff_grid_tile (out, img, tile_width, tile_height) != GGRAPH_OK)
	goto error;

    TIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (out)
	TIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_geo_tiff_grid_strip (const gGraphImagePtr img, void **mem_buf,
				 int *mem_buf_size, FILE * file, int dest_type,
				 int rows_per_strip, int compression)
{
/* exporting a GRID image [by strips] */
    TIFF *out = (TIFF *) 0;
    GTIF *gtif = (GTIF *) 0;
    double tiepoint[6];
    double pixsize[3];
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (img->is_georeferenced == 0)
	return GGRAPH_MISSING_GEO_DEFS;

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			       file_writeproc, file_seekproc, closeproc,
			       file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			       memory_writeproc, memory_seekproc, closeproc,
			       memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;
    gtif = GTIFNew (out);
    if (!gtif)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    if (img->sample_format == GGRAPH_SAMPLE_INT)
	TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    else if (img->sample_format == GGRAPH_SAMPLE_FLOAT)
	TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_IEEEFP);
    else
	TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_INT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, img->bits_per_sample);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_ROWSPERSTRIP, rows_per_strip);

/* setting up the GeoTIFF Tags */
    pixsize[0] = img->pixel_x_size;
    pixsize[1] = img->pixel_y_size;
    pixsize[2] = 0.0;
    TIFFSetField (out, GTIFF_PIXELSCALE, 3, pixsize);
    tiepoint[0] = 0.0;
    tiepoint[1] = 0.0;
    tiepoint[2] = 0.0;
    tiepoint[3] = img->upper_left_x;
    tiepoint[4] = img->upper_left_y;
    tiepoint[5] = 0.0;
    TIFFSetField (out, GTIFF_TIEPOINTS, 6, tiepoint);
    if (img->srs_name != NULL)
	TIFFSetField (out, GTIFF_ASCIIPARAMS, img->srs_name);
    if (img->proj4text != NULL)
	GTIFSetFromProj4 (gtif, img->proj4text);
    if (img->srs_name != NULL)
	GTIFKeySet (gtif, GTCitationGeoKey, TYPE_ASCII, 0, img->srs_name);
    if (is_projected_srs (img->proj4text))
	GTIFKeySet (gtif, ProjectedCSTypeGeoKey, TYPE_SHORT, 1, img->srid);
    GTIFWriteKeys (gtif);

    if (common_tiff_grid_strip (out, img) != GGRAPH_OK)
	goto error;

    GTIFFree (gtif);
    XTIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (gtif)
	GTIFFree (gtif);
    if (out)
	XTIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_to_geo_tiff_grid_tile (const gGraphImagePtr img, void **mem_buf,
				int *mem_buf_size, FILE * file, int dest_type,
				int tile_width, int tile_height,
				int compression)
{
/* exporting a GRID image [by tiles] */
    TIFF *out = (TIFF *) 0;
    GTIF *gtif = (GTIF *) 0;
    double tiepoint[6];
    double pixsize[3];
    struct memfile clientdata;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (img->is_georeferenced == 0)
	return GGRAPH_MISSING_GEO_DEFS;

    if (dest_type == GG_TARGET_IS_FILE)
      {
	  /* writing to file */
	  clientdata.buffer = NULL;
	  clientdata.file = file;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, file_readproc,
			       file_writeproc, file_seekproc, closeproc,
			       file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* writing to memory */
	  if (compression == GGRAPH_TIFF_COMPRESSION_LZW
	      || compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	      clientdata.malloc_block = (img->width * img->height) / 100;
	  else
	      clientdata.malloc_block = (img->width * img->height) / 10;
	  if (clientdata.malloc_block < 65536)
	      clientdata.malloc_block = 65536;
	  clientdata.buffer = malloc (clientdata.malloc_block);
	  if (clientdata.buffer == NULL)
	      return GGRAPH_INSUFFICIENT_MEMORY;
	  clientdata.size = clientdata.malloc_block;
	  clientdata.eof = 0;
	  clientdata.current = 0;
	  out =
	      XTIFFClientOpen ("tiff", "w", &clientdata, memory_readproc,
			       memory_writeproc, memory_seekproc, closeproc,
			       memory_sizeproc, mapproc, unmapproc);
      }
    if (out == NULL)
	goto error;
    gtif = GTIFNew (out);
    if (!gtif)
	goto error;

/* setting up the TIFF headers */
    TIFFSetField (out, TIFFTAG_SUBFILETYPE, 0);
    TIFFSetField (out, TIFFTAG_IMAGEWIDTH, img->width);
    TIFFSetField (out, TIFFTAG_IMAGELENGTH, img->height);
    if (img->sample_format == GGRAPH_SAMPLE_INT)
	TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_INT);
    else if (img->sample_format == GGRAPH_SAMPLE_FLOAT)
	TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_IEEEFP);
    else
	TIFFSetField (out, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
    TIFFSetField (out, TIFFTAG_SAMPLESPERPIXEL, 1);
    TIFFSetField (out, TIFFTAG_BITSPERSAMPLE, img->bits_per_sample);
    TIFFSetField (out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField (out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField (out, TIFFTAG_XRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_YRESOLUTION, 300.0);
    TIFFSetField (out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
    TIFFSetField (out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
    if (compression == GGRAPH_TIFF_COMPRESSION_LZW)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
    else if (compression == GGRAPH_TIFF_COMPRESSION_DEFLATE)
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE);
    else
	TIFFSetField (out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    TIFFSetField (out, TIFFTAG_SOFTWARE, "GaiaGraphics-tools");
    TIFFSetField (out, TIFFTAG_TILEWIDTH, tile_width);
    TIFFSetField (out, TIFFTAG_TILELENGTH, tile_height);

/* setting up the GeoTIFF Tags */
    pixsize[0] = img->pixel_x_size;
    pixsize[1] = img->pixel_y_size;
    pixsize[2] = 0.0;
    TIFFSetField (out, GTIFF_PIXELSCALE, 3, pixsize);
    tiepoint[0] = 0.0;
    tiepoint[1] = 0.0;
    tiepoint[2] = 0.0;
    tiepoint[3] = img->upper_left_x;
    tiepoint[4] = img->upper_left_y;
    tiepoint[5] = 0.0;
    TIFFSetField (out, GTIFF_TIEPOINTS, 6, tiepoint);
    if (img->srs_name != NULL)
	TIFFSetField (out, GTIFF_ASCIIPARAMS, img->srs_name);
    if (img->proj4text != NULL)
	GTIFSetFromProj4 (gtif, img->proj4text);
    if (img->srs_name != NULL)
	GTIFKeySet (gtif, GTCitationGeoKey, TYPE_ASCII, 0, img->srs_name);
    if (is_projected_srs (img->proj4text))
	GTIFKeySet (gtif, ProjectedCSTypeGeoKey, TYPE_SHORT, 1, img->srid);
    GTIFWriteKeys (gtif);

    if (common_tiff_grid_tile (out, img, tile_width, tile_height) != GGRAPH_OK)
	goto error;

    GTIFFree (gtif);
    XTIFFClose (out);
    if (dest_type == GG_TARGET_IS_FILE)
	return GGRAPH_OK;

/* returning the memory buffer */
    if (clientdata.eof > 0)
      {
	  *mem_buf = clientdata.buffer;
	  *mem_buf_size = clientdata.eof;
	  return GGRAPH_OK;
      }

  error:
    if (gtif)
	GTIFFree (gtif);
    if (out)
	XTIFFClose (out);
    if (clientdata.buffer != NULL)
	free (clientdata.buffer);
    return GGRAPH_TIFF_CODEC_ERROR;
}

extern int
gg_image_infos_from_tiff (int size, const void *data, int source_type,
			  gGraphImageInfosPtr * infos_handle)
{
/* image infos from TIFF */
    gGraphImageInfosPtr infos;
    uint16 bits_per_sample;
    uint16 samples_per_pixel;
    uint16 photometric;
    uint16 compression;
    uint16 sample_format;
    uint16 planar_config;
    uint32 width = 0;
    uint32 height = 0;
    uint32 rows_strip = 0;
    struct memfile clientdata;
    int is_tiled;
    uint32 tile_width;
    uint32 tile_height;
    int type;
    int gg_sample_format;
    TIFF *in = (TIFF *) 0;
    *infos_handle = NULL;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (source_type == GG_TARGET_IS_FILE)
      {
	  /* reading from file */
	  clientdata.file = (FILE *) data;
	  in = TIFFClientOpen ("tiff", "r", &clientdata, file_readproc,
			       file_writeproc, file_seekproc, closeproc,
			       file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* reading from memory */
	  clientdata.buffer = (unsigned char *) data;
	  clientdata.size = size;
	  clientdata.eof = size;
	  clientdata.current = 0;
	  in = TIFFClientOpen ("tiff", "r", &clientdata, memory_readproc,
			       memory_writeproc, memory_seekproc, closeproc,
			       memory_sizeproc, mapproc, unmapproc);
      }
    if (in == NULL)
	return GGRAPH_TIFF_CODEC_ERROR;
    is_tiled = TIFFIsTiled (in);
/* retrieving the TIFF dimensions */
    TIFFGetField (in, TIFFTAG_IMAGELENGTH, &height);
    TIFFGetField (in, TIFFTAG_IMAGEWIDTH, &width);
    if (is_tiled)
      {
	  TIFFGetField (in, TIFFTAG_TILEWIDTH, &tile_width);
	  TIFFGetField (in, TIFFTAG_TILELENGTH, &tile_height);
      }
    else
	TIFFGetField (in, TIFFTAG_ROWSPERSTRIP, &rows_strip);
    TIFFGetField (in, TIFFTAG_BITSPERSAMPLE, &bits_per_sample);
    if (TIFFGetField (in, TIFFTAG_SAMPLESPERPIXEL, &samples_per_pixel) == 0)
      {
	  /* attempting to recover badly formatted TIFFs */
	  samples_per_pixel = 1;
      }
    TIFFGetField (in, TIFFTAG_PHOTOMETRIC, &photometric);
    TIFFGetField (in, TIFFTAG_COMPRESSION, &compression);
    if (TIFFGetField (in, TIFFTAG_PLANARCONFIG, &planar_config) == 0)
      {
	  /* attempting to recover badly formatted TIFFs */
	  planar_config = 1;
      }
    if (TIFFGetField (in, TIFFTAG_SAMPLEFORMAT, &sample_format) == 0)
	sample_format = SAMPLEFORMAT_UINT;
    if (planar_config == PLANARCONFIG_CONTIG)
      {
	  if (sample_format == SAMPLEFORMAT_UINT)
	    {
		if (bits_per_sample == 1 && samples_per_pixel == 1)
		    type = GG_PIXEL_PALETTE;
		else if (bits_per_sample == 8 && samples_per_pixel == 1
			 && photometric == 3)
		    type = GG_PIXEL_PALETTE;
		else if (bits_per_sample == 8 && samples_per_pixel == 1
			 && photometric < 2)
		    type = GG_PIXEL_GRAYSCALE;
		else if (bits_per_sample == 8 && samples_per_pixel == 3)
		    type = GG_PIXEL_RGB;
		else
		    type = GG_PIXEL_UNKNOWN;
	    }
	  else if (samples_per_pixel == 1)
	      type = GG_PIXEL_GRID;
	  else
	      type = GG_PIXEL_UNKNOWN;
      }
    else
	type = GG_PIXEL_UNKNOWN;
    switch (sample_format)
      {
      case SAMPLEFORMAT_UINT:
	  gg_sample_format = GGRAPH_SAMPLE_UINT;
	  break;
      case SAMPLEFORMAT_INT:
	  gg_sample_format = GGRAPH_SAMPLE_INT;
	  break;
      case SAMPLEFORMAT_IEEEFP:
	  gg_sample_format = GGRAPH_SAMPLE_FLOAT;
	  break;
      default:
	  gg_sample_format = GGRAPH_SAMPLE_UNKNOWN;
	  break;
      };
    infos =
	gg_image_infos_create (type, width, height, bits_per_sample,
			       samples_per_pixel, gg_sample_format, NULL, NULL);
    if (!infos)
	return GGRAPH_INSUFFICIENT_MEMORY;
    if (is_tiled)
      {
	  infos->tile_width = tile_width;
	  infos->tile_height = tile_height;
      }
    else
	infos->rows_per_strip = rows_strip;
    switch (compression)
      {
      case COMPRESSION_NONE:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_NONE;
	  break;
      case COMPRESSION_LZW:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_LZW;
	  break;
      case COMPRESSION_DEFLATE:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_DEFLATE;
	  break;
      case COMPRESSION_JPEG:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_JPEG;
	  break;
      case COMPRESSION_CCITTFAX3:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_CCITTFAX3;
	  break;
      case COMPRESSION_CCITTFAX4:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_CCITTFAX4;
	  break;
      default:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_UNKNOWN;
	  break;
      };
    TIFFClose (in);
    *infos_handle = infos;
    return GGRAPH_OK;
}

extern int
gg_image_infos_from_geo_tiff (int size, const void *data, int source_type,
			      gGraphImageInfosPtr * infos_handle)
{
/* image infos from GeoTIFF */
    gGraphImageInfosPtr infos;
    uint16 bits_per_sample;
    uint16 samples_per_pixel;
    uint16 photometric;
    uint16 compression;
    uint16 sample_format;
    uint16 planar_config;
    uint32 width = 0;
    uint32 height = 0;
    uint32 rows_strip = 0;
    struct memfile clientdata;
    int is_tiled;
    uint32 tile_width;
    uint32 tile_height;
    int type;
    int gg_sample_format;
    int epsg = -1;
    double cx;
    double cy;
    double upper_left_x;
    double upper_left_y;
    double upper_right_x;
    double upper_right_y;
    double lower_left_x;
    double lower_left_y;
    double lower_right_x;
    double lower_right_y;
    double pixel_x;
    double pixel_y;
    char srs_name[1024];
    char proj4text[1024];
    TIFF *in = (TIFF *) 0;
    GTIF *gtif = (GTIF *) 0;
    GTIFDefn definition;
    *infos_handle = NULL;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (source_type == GG_TARGET_IS_FILE)
      {
	  /* reading from file */
	  clientdata.file = (FILE *) data;
	  in = XTIFFClientOpen ("tiff", "r", &clientdata, file_readproc,
				file_writeproc, file_seekproc, closeproc,
				file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* reading from memory */
	  clientdata.buffer = (unsigned char *) data;
	  clientdata.size = size;
	  clientdata.eof = size;
	  clientdata.current = 0;
	  in = XTIFFClientOpen ("tiff", "r", &clientdata, memory_readproc,
				memory_writeproc, memory_seekproc, closeproc,
				memory_sizeproc, mapproc, unmapproc);
      }
    if (in == NULL)
	return GGRAPH_GEOTIFF_CODEC_ERROR;
    gtif = GTIFNew (in);
    if (gtif == NULL)
	goto error;

    if (!GTIFGetDefn (gtif, &definition))
	goto error;

/* retrieving the EPSG code */
    if (definition.PCS == 32767)
      {
	  if (definition.GCS != 32767)
	      epsg = definition.GCS;
      }
    else
	epsg = definition.PCS;
    *srs_name = '\0';
    if (definition.PCS == 32767)
      {
	  /* Get the GCS name if possible */
	  char *pszName = NULL;
	  GTIFGetGCSInfo (definition.GCS, &pszName, NULL, NULL, NULL);
	  if (pszName != NULL)
	      strcpy (srs_name, pszName);
	  CPLFree (pszName);
      }
    else
      {
	  /* Get the PCS name if possible */
	  char *pszPCSName = NULL;
	  GTIFGetPCSInfo (definition.PCS, &pszPCSName, NULL, NULL, NULL);
	  if (pszPCSName != NULL)
	      strcpy (srs_name, pszPCSName);
	  CPLFree (pszPCSName);
      }
/* retrieving the PROJ.4 params */
    strcpy (proj4text, GTIFGetProj4Defn (&definition));

    is_tiled = TIFFIsTiled (in);
/* retrieving the TIFF dimensions */
    TIFFGetField (in, TIFFTAG_IMAGELENGTH, &height);
    TIFFGetField (in, TIFFTAG_IMAGEWIDTH, &width);
    if (is_tiled)
      {
	  TIFFGetField (in, TIFFTAG_TILEWIDTH, &tile_width);
	  TIFFGetField (in, TIFFTAG_TILELENGTH, &tile_height);
      }
    else
	TIFFGetField (in, TIFFTAG_ROWSPERSTRIP, &rows_strip);
    TIFFGetField (in, TIFFTAG_BITSPERSAMPLE, &bits_per_sample);
    if (TIFFGetField (in, TIFFTAG_SAMPLESPERPIXEL, &samples_per_pixel) == 0)
      {
	  /* attempting to recover badly formatted TIFFs */
	  samples_per_pixel = 1;
      }
    TIFFGetField (in, TIFFTAG_PHOTOMETRIC, &photometric);
    TIFFGetField (in, TIFFTAG_COMPRESSION, &compression);
    if (TIFFGetField (in, TIFFTAG_PLANARCONFIG, &planar_config) == 0)
      {
	  /* attempting to recover badly formatted TIFFs */
	  planar_config = PLANARCONFIG_CONTIG;
      }
    if (TIFFGetField (in, TIFFTAG_SAMPLEFORMAT, &sample_format) == 0)
	sample_format = SAMPLEFORMAT_UINT;
    if (planar_config == PLANARCONFIG_CONTIG)
      {
	  if (sample_format == SAMPLEFORMAT_UINT)
	    {
		if (bits_per_sample == 1 && samples_per_pixel == 1)
		    type = GG_PIXEL_PALETTE;
		else if (bits_per_sample == 8 && samples_per_pixel == 1
			 && photometric == 3)
		    type = GG_PIXEL_PALETTE;
		else if (bits_per_sample == 8 && samples_per_pixel == 1
			 && photometric < 2)
		    type = GG_PIXEL_GRAYSCALE;
		else if (bits_per_sample == 8 && samples_per_pixel == 3)
		    type = GG_PIXEL_RGB;
		else
		    type = GG_PIXEL_UNKNOWN;
	    }
	  else if (samples_per_pixel == 1)
	      type = GG_PIXEL_GRID;
	  else
	      type = GG_PIXEL_UNKNOWN;
      }
    else
	type = GG_PIXEL_UNKNOWN;
    switch (sample_format)
      {
      case SAMPLEFORMAT_UINT:
	  gg_sample_format = GGRAPH_SAMPLE_UINT;
	  break;
      case SAMPLEFORMAT_INT:
	  gg_sample_format = GGRAPH_SAMPLE_INT;
	  break;
      case SAMPLEFORMAT_IEEEFP:
	  gg_sample_format = GGRAPH_SAMPLE_FLOAT;
	  break;
      default:
	  gg_sample_format = GGRAPH_SAMPLE_UNKNOWN;
	  break;
      };

/* computing the corners coords */
    cx = 0.0;
    cy = 0.0;
    GTIFImageToPCS (gtif, &cx, &cy);
    upper_left_x = cx;
    upper_left_y = cy;
    cx = 0.0;
    cy = height;
    GTIFImageToPCS (gtif, &cx, &cy);
    lower_left_x = cx;
    lower_left_y = cy;
    cx = width;
    cy = 0.0;
    GTIFImageToPCS (gtif, &cx, &cy);
    upper_right_x = cx;
    upper_right_y = cy;
    cx = width;
    cy = height;
    GTIFImageToPCS (gtif, &cx, &cy);
    lower_right_x = cx;
    lower_right_y = cy;
/* computing the pixel size */
    pixel_x = (upper_right_x - upper_left_x) / (double) width;
    pixel_y = (upper_left_y - lower_left_y) / (double) height;

    infos =
	gg_image_infos_create (type, width, height, bits_per_sample,
			       samples_per_pixel, gg_sample_format, srs_name,
			       proj4text);
    if (!infos)
	return GGRAPH_INSUFFICIENT_MEMORY;
    if (is_tiled)
      {
	  infos->tile_width = tile_width;
	  infos->tile_height = tile_height;
      }
    else
	infos->rows_per_strip = rows_strip;
    switch (compression)
      {
      case COMPRESSION_NONE:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_NONE;
	  break;
      case COMPRESSION_LZW:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_LZW;
	  break;
      case COMPRESSION_DEFLATE:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_DEFLATE;
	  break;
      case COMPRESSION_JPEG:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_JPEG;
	  break;
      case COMPRESSION_CCITTFAX3:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_CCITTFAX3;
	  break;
      case COMPRESSION_CCITTFAX4:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_CCITTFAX4;
	  break;
      default:
	  infos->compression = GGRAPH_TIFF_COMPRESSION_UNKNOWN;
	  break;
      };
    infos->is_georeferenced = 1;
    infos->srid = epsg;
    infos->upper_left_x = upper_left_x;
    infos->upper_left_y = upper_left_y;
    infos->pixel_x_size = pixel_x;
    infos->pixel_y_size = pixel_y;
    TIFFClose (in);
    GTIFFree (gtif);
    *infos_handle = infos;
    return GGRAPH_OK;

  error:
    XTIFFClose (in);
    if (gtif)
	GTIFFree (gtif);
    return GGRAPH_GEOTIFF_CODEC_ERROR;
}

static int
common_read_from_tiff_grid (TIFF * in, gGraphImagePtr img, uint32 width,
			    uint32 height, int is_tiled, int gg_sample_format,
			    uint16 bits_per_sample, uint32 tile_width,
			    uint32 tile_height, uint32 rows_strip)
{
/* common utility: decoding a TIFF raster [GRID] */
    unsigned char *raster = NULL;
    int x;
    int y;
    int i;
    int img_y;
    int strip_no;
    int effective_strip;
    int tile_x;
    int tile_y;
    unsigned char *p_in;
    unsigned char *p_out;
    int ret = GGRAPH_TIFF_CODEC_ERROR;

    if (is_tiled)
      {
	  if (gg_sample_format == GGRAPH_SAMPLE_FLOAT)
	    {
		if (bits_per_sample == 32)
		  {
		      /* float */
		      raster = malloc (4 * tile_width * tile_height);
		  }
		else
		  {
		      /* double */
		      raster = malloc (8 * tile_width * tile_height);
		  }
	    }
	  else
	    {
		if (bits_per_sample == 8)
		  {
		      /* 8bit Int */
		      raster = malloc (tile_width * tile_height);
		  }
		else if (bits_per_sample == 16)
		  {
		      /* 16bit Int */
		      raster = malloc (2 * tile_width * tile_height);
		  }
		else
		  {
		      /* 32bit Int */
		      raster = malloc (4 * tile_width * tile_height);
		  }
	    }
      }
    else
      {
	  if (gg_sample_format == GGRAPH_SAMPLE_FLOAT)
	    {
		if (bits_per_sample == 32)
		  {
		      /* float */
		      raster = malloc (4 * width * rows_strip);
		  }
		else
		  {
		      /* double */
		      raster = malloc (8 * width * rows_strip);
		  }
	    }
	  else
	    {
		if (bits_per_sample == 8)
		  {
		      /* 8bit Int */
		      raster = malloc (width * rows_strip);
		  }
		else if (bits_per_sample == 16)
		  {
		      /* 16bit Int */
		      raster = malloc (2 * width * rows_strip);
		  }
		else
		  {
		      /* 32bit Int */
		      raster = malloc (4 * width * rows_strip);
		  }
	    }
      }

    ret = GGRAPH_TIFF_CODEC_ERROR;
    if (is_tiled)
      {
	  for (tile_y = 0; tile_y < (int) height; tile_y += tile_height)
	    {
		/* scanning tiles by row */
		for (tile_x = 0; tile_x < (int) width; tile_x += tile_width)
		  {
		      /* reading a TIFF tile */
		      if (!TIFFReadTile (in, raster, tile_x, tile_y, 0, 0))
			  goto error;
		      for (y = 0; y < (int) tile_height; y++)
			{
			    img_y = tile_y + y;
			    if (img_y >= (int) height)
				continue;
			    p_in = raster + (tile_width * y * img->pixel_size);
			    p_out =
				img->pixels + (img_y * img->scanline_width) +
				(tile_x * img->pixel_size);
			    for (x = 0; x < (int) tile_width; x++)
			      {
				  if (tile_x + x >= (int) width)
				      break;
				  for (i = 0; i < img->pixel_size; i++)
				      *p_out++ = *p_in++;
			      }
			}
		  }
	    }
      }
    else
      {
	  for (strip_no = 0; strip_no < (int) height; strip_no += rows_strip)
	    {
		/* reading a TIFF strip */
		if (!TIFFReadScanline (in, raster, strip_no, 0))
		    goto error;
		effective_strip = rows_strip;
		if ((strip_no + rows_strip) > height)
		    effective_strip = height - strip_no;

		for (y = 0; y < effective_strip; y++)
		  {
		      img_y = strip_no + y;
		      p_in = raster;
		      p_out = img->pixels + (img_y * img->scanline_width);
		      for (x = 0; x < (int) width; x++)
			{
			    for (i = 0; i < img->pixel_size; i++)
				*p_out++ = *p_in++;
			}
		  }
	    }
      }
    free (raster);
    return GGRAPH_OK;

  error:
    if (raster)
	free (raster);
    return ret;
}

static int
common_read_from_tiff (TIFF * in, gGraphImagePtr img, uint32 width,
		       uint32 height, int is_tiled, int type,
		       int gg_sample_format, uint16 bits_per_sample,
		       uint32 tile_width, uint32 tile_height, uint32 rows_strip)
{
/* common utility: decoding a TIFF raster */
    uint32 *raster = NULL;
    uint32 *scanline;
    int x;
    int y;
    int img_y;
    int strip_no;
    int effective_strip;
    uint32 pixel;
    int tile_x;
    int tile_y;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char index;
    unsigned char *p_out;
    int ret = GGRAPH_TIFF_CODEC_ERROR;

    if (type == GG_PIXEL_GRID)
	return common_read_from_tiff_grid (in, img, width, height, is_tiled,
					   gg_sample_format, bits_per_sample,
					   tile_width, tile_height, rows_strip);

/* allocating read buffer [plain ordinary image] */
    if (is_tiled)
	raster = malloc (sizeof (uint32) * tile_width * tile_height);
    else
	raster = malloc (sizeof (uint32) * width * rows_strip);
    if (!raster)
      {
	  gg_image_destroy (img);
	  ret = GGRAPH_INSUFFICIENT_MEMORY;
	  goto error;
      }

    ret = GGRAPH_TIFF_CODEC_ERROR;
    if (is_tiled)
      {
	  for (tile_y = 0; tile_y < (int) height; tile_y += tile_height)
	    {
		/* scanning tiles by row */
		for (tile_x = 0; tile_x < (int) width; tile_x += tile_width)
		  {
		      /* reading a TIFF tile */
		      if (!TIFFReadRGBATile (in, tile_x, tile_y, raster))
			  goto error;
		      for (y = 0; y < (int) tile_height; y++)
			{
			    img_y = tile_y + (tile_height - y) - 1;
			    if (img_y >= (int) height)
				continue;
			    scanline = raster + (tile_width * y);
			    for (x = 0; x < (int) tile_width; x++)
			      {
				  if (tile_x + x >= (int) width)
				      break;
				  p_out =
				      img->pixels +
				      (img_y * img->scanline_width) +
				      ((tile_x + x) * img->pixel_size);
				  pixel = scanline[x];
				  red = TIFFGetR (pixel);
				  green = TIFFGetG (pixel);
				  blue = TIFFGetB (pixel);
				  if (img->pixel_format == GG_PIXEL_PALETTE)
				    {
					/* PALETTE image */
					index =
					    gg_match_palette (img, red, green,
							      blue);
					*p_out++ = index;
				    }
				  else if (img->pixel_format ==
					   GG_PIXEL_GRAYSCALE)
				    {
					/* GRAYSCALE  image */
					*p_out++ = red;
				    }
				  else
				    {
					/* should be an RGB image */
					*p_out++ = red;
					*p_out++ = green;
					*p_out++ = blue;
				    }
			      }
			}
		  }
	    }
      }
    else
      {
	  for (strip_no = 0; strip_no < (int) height; strip_no += rows_strip)
	    {
		/* reading a TIFF strip */
		if (!TIFFReadRGBAStrip (in, strip_no, raster))
		    goto error;
		effective_strip = rows_strip;
		if ((strip_no + rows_strip) > height)
		    effective_strip = height - strip_no;

		for (y = 0; y < effective_strip; y++)
		  {
		      img_y = strip_no + ((effective_strip - y) - 1);
		      scanline = raster + (width * y);
		      p_out = img->pixels + (img_y * img->scanline_width);
		      for (x = 0; x < (int) width; x++)
			{
			    pixel = scanline[x];
			    red = TIFFGetR (pixel);
			    green = TIFFGetG (pixel);
			    blue = TIFFGetB (pixel);
			    if (img->pixel_format == GG_PIXEL_PALETTE)
			      {
				  /* PALETTE image */
				  index =
				      gg_match_palette (img, red, green, blue);
				  *p_out++ = index;
			      }
			    else if (img->pixel_format == GG_PIXEL_GRAYSCALE)
			      {
				  /* GRAYSCALE  image */
				  *p_out++ = red;
			      }
			    else
			      {
				  /* should be an RGB image */
				  *p_out++ = red;
				  *p_out++ = green;
				  *p_out++ = blue;
			      }
			}
		  }
	    }
      }
    free (raster);
    return GGRAPH_OK;

  error:
    if (raster)
	free (raster);
    return ret;
}

extern int
gg_image_from_tiff (int size, const void *data, int source_type,
		    gGraphImagePtr * image_handle)
{
/* uncompressing a TIFF */
    gGraphImagePtr img = NULL;
    uint16 bits_per_sample;
    uint16 samples_per_pixel;
    uint16 photometric;
    uint16 compression;
    uint16 planar_config;
    uint16 sample_format;
    uint32 width = 0;
    uint32 height = 0;
    uint32 rows_strip = 0;
    struct memfile clientdata;
    int is_tiled;
    uint32 tile_width;
    uint32 tile_height;
    int type;
    int gg_sample_format;
    int ret = GGRAPH_TIFF_CODEC_ERROR;
    TIFF *in = (TIFF *) 0;
    *image_handle = NULL;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (source_type == GG_TARGET_IS_FILE)
      {
	  /* reading from file */
	  clientdata.file = (FILE *) data;
	  in = TIFFClientOpen ("tiff", "r", &clientdata, file_readproc,
			       file_writeproc, file_seekproc, closeproc,
			       file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* reading from memory */
	  clientdata.buffer = (unsigned char *) data;
	  clientdata.size = size;
	  clientdata.eof = size;
	  clientdata.current = 0;
	  in = TIFFClientOpen ("tiff", "r", &clientdata, memory_readproc,
			       memory_writeproc, memory_seekproc, closeproc,
			       memory_sizeproc, mapproc, unmapproc);
      }
    if (in == NULL)
	return GGRAPH_TIFF_CODEC_ERROR;
    is_tiled = TIFFIsTiled (in);
/* retrieving the TIFF dimensions */
    TIFFGetField (in, TIFFTAG_IMAGELENGTH, &height);
    TIFFGetField (in, TIFFTAG_IMAGEWIDTH, &width);
    if (is_tiled)
      {
	  TIFFGetField (in, TIFFTAG_TILEWIDTH, &tile_width);
	  TIFFGetField (in, TIFFTAG_TILELENGTH, &tile_height);
      }
    else
	TIFFGetField (in, TIFFTAG_ROWSPERSTRIP, &rows_strip);
    TIFFGetField (in, TIFFTAG_BITSPERSAMPLE, &bits_per_sample);
    if (TIFFGetField (in, TIFFTAG_SAMPLESPERPIXEL, &samples_per_pixel) == 0)
      {
	  /* attempting to recover badly formatted TIFFs */
	  samples_per_pixel = 1;
      }
    TIFFGetField (in, TIFFTAG_PHOTOMETRIC, &photometric);
    TIFFGetField (in, TIFFTAG_COMPRESSION, &compression);
    if (TIFFGetField (in, TIFFTAG_PLANARCONFIG, &planar_config) == 0)
      {
	  /* attempting to recover badly formatted TIFFs */
	  planar_config = PLANARCONFIG_CONTIG;
      }
    if (TIFFGetField (in, TIFFTAG_SAMPLEFORMAT, &sample_format) == 0)
	sample_format = SAMPLEFORMAT_UINT;
    if (planar_config == PLANARCONFIG_CONTIG)
      {
	  if (sample_format == SAMPLEFORMAT_UINT)
	    {
		if (bits_per_sample == 1 && samples_per_pixel == 1)
		    type = GG_PIXEL_PALETTE;
		else if (bits_per_sample == 8 && samples_per_pixel == 1
			 && photometric == 3)
		    type = GG_PIXEL_PALETTE;
		else if (bits_per_sample == 8 && samples_per_pixel == 1
			 && photometric < 2)
		    type = GG_PIXEL_GRAYSCALE;
		else if (bits_per_sample == 8 && samples_per_pixel == 3)
		    type = GG_PIXEL_RGB;
		else
		    type = GG_PIXEL_UNKNOWN;
	    }
	  else if (samples_per_pixel == 1)
	      type = GG_PIXEL_GRID;
	  else
	      type = GG_PIXEL_UNKNOWN;
      }
    else
	type = GG_PIXEL_UNKNOWN;
    switch (sample_format)
      {
      case SAMPLEFORMAT_UINT:
	  gg_sample_format = GGRAPH_SAMPLE_UINT;
	  break;
      case SAMPLEFORMAT_INT:
	  gg_sample_format = GGRAPH_SAMPLE_INT;
	  break;
      case SAMPLEFORMAT_IEEEFP:
	  gg_sample_format = GGRAPH_SAMPLE_FLOAT;
	  break;
      default:
	  gg_sample_format = GGRAPH_SAMPLE_UNKNOWN;
	  break;
      };
    if (type == GG_PIXEL_UNKNOWN)
      {
	  ret = GGRAPH_UNSUPPORTED_TIFF_LAYOUT;
	  goto error;
      }

    img =
	gg_image_create (type, width, height, bits_per_sample,
			 samples_per_pixel, gg_sample_format, NULL, NULL);
    if (!img)
      {
	  ret = GGRAPH_INSUFFICIENT_MEMORY;
	  goto error;
      }
    if (is_tiled)
      {
	  img->tile_width = tile_width;
	  img->tile_height = tile_height;
      }
    else
	img->rows_per_strip = rows_strip;

    ret =
	common_read_from_tiff (in, img, width, height, is_tiled, type,
			       gg_sample_format, bits_per_sample, tile_width,
			       tile_height, rows_strip);
    if (ret != GGRAPH_OK)
	goto error;

    TIFFClose (in);
    *image_handle = img;
    return GGRAPH_OK;

  error:
    TIFFClose (in);
    if (img)
	gGraphDestroyImage (img);
    return ret;
}

extern int
gg_image_from_geo_tiff (int size, const void *data, int source_type,
			gGraphImagePtr * image_handle)
{
/* uncompressing a GeoTIFF */
    gGraphImagePtr img = NULL;
    uint16 bits_per_sample;
    uint16 samples_per_pixel;
    uint16 photometric;
    uint16 compression;
    uint16 planar_config;
    uint16 sample_format;
    uint32 width = 0;
    uint32 height = 0;
    uint32 rows_strip = 0;
    struct memfile clientdata;
    int is_tiled;
    uint32 tile_width;
    uint32 tile_height;
    int type;
    int gg_sample_format;
    int epsg = -1;
    double cx;
    double cy;
    double upper_left_x;
    double upper_left_y;
    double upper_right_x;
    double upper_right_y;
    double lower_left_x;
    double lower_left_y;
    double lower_right_x;
    double lower_right_y;
    double pixel_x;
    double pixel_y;
    char srs_name[1024];
    char proj4text[1024];
    int ret = GGRAPH_TIFF_CODEC_ERROR;
    TIFF *in = (TIFF *) 0;
    GTIF *gtif = (GTIF *) 0;
    GTIFDefn definition;
    *image_handle = NULL;

/* suppressing TIFF warnings */
    TIFFSetWarningHandler (NULL);

    if (source_type == GG_TARGET_IS_FILE)
      {
	  /* reading from file */
	  clientdata.file = (FILE *) data;
	  in = XTIFFClientOpen ("tiff", "r", &clientdata, file_readproc,
				file_writeproc, file_seekproc, closeproc,
				file_sizeproc, mapproc, unmapproc);
      }
    else
      {
	  /* reading from memory */
	  clientdata.buffer = (unsigned char *) data;
	  clientdata.size = size;
	  clientdata.eof = size;
	  clientdata.current = 0;
	  in = XTIFFClientOpen ("tiff", "r", &clientdata, memory_readproc,
				memory_writeproc, memory_seekproc, closeproc,
				memory_sizeproc, mapproc, unmapproc);
      }
    if (in == NULL)
	return GGRAPH_GEOTIFF_CODEC_ERROR;
    gtif = GTIFNew (in);
    if (gtif == NULL)
	goto error;

    if (!GTIFGetDefn (gtif, &definition))
	goto error;

/* retrieving the EPSG code */
    if (definition.PCS == 32767)
      {
	  if (definition.GCS != 32767)
	      epsg = definition.GCS;
      }
    else
	epsg = definition.PCS;
    *srs_name = '\0';
    if (definition.PCS == 32767)
      {
	  /* Get the GCS name if possible */
	  char *pszName = NULL;
	  GTIFGetGCSInfo (definition.GCS, &pszName, NULL, NULL, NULL);
	  if (pszName != NULL)
	      strcpy (srs_name, pszName);
	  CPLFree (pszName);
      }
    else
      {
	  /* Get the PCS name if possible */
	  char *pszPCSName = NULL;
	  GTIFGetPCSInfo (definition.PCS, &pszPCSName, NULL, NULL, NULL);
	  if (pszPCSName != NULL)
	      strcpy (srs_name, pszPCSName);
	  CPLFree (pszPCSName);
      }
/* retrieving the PROJ.4 params */
    strcpy (proj4text, GTIFGetProj4Defn (&definition));

    is_tiled = TIFFIsTiled (in);
/* retrieving the TIFF dimensions */
    TIFFGetField (in, TIFFTAG_IMAGELENGTH, &height);
    TIFFGetField (in, TIFFTAG_IMAGEWIDTH, &width);
    if (is_tiled)
      {
	  TIFFGetField (in, TIFFTAG_TILEWIDTH, &tile_width);
	  TIFFGetField (in, TIFFTAG_TILELENGTH, &tile_height);
      }
    else
	TIFFGetField (in, TIFFTAG_ROWSPERSTRIP, &rows_strip);
    TIFFGetField (in, TIFFTAG_BITSPERSAMPLE, &bits_per_sample);
    if (TIFFGetField (in, TIFFTAG_SAMPLESPERPIXEL, &samples_per_pixel) == 0)
      {
	  /* attempting to recover badly formatted TIFFs */
	  samples_per_pixel = 1;
      }
    TIFFGetField (in, TIFFTAG_PHOTOMETRIC, &photometric);
    TIFFGetField (in, TIFFTAG_COMPRESSION, &compression);
    if (TIFFGetField (in, TIFFTAG_PLANARCONFIG, &planar_config) == 0)
      {
	  /* attempting to recover badly formatted TIFFs */
	  planar_config = PLANARCONFIG_CONTIG;
      }
    if (TIFFGetField (in, TIFFTAG_SAMPLEFORMAT, &sample_format) == 0)
	sample_format = SAMPLEFORMAT_UINT;
    if (planar_config == PLANARCONFIG_CONTIG)
      {
	  if (sample_format == SAMPLEFORMAT_UINT)
	    {
		if (bits_per_sample == 1 && samples_per_pixel == 1)
		    type = GG_PIXEL_PALETTE;
		else if (bits_per_sample == 8 && samples_per_pixel == 1
			 && photometric == 3)
		    type = GG_PIXEL_PALETTE;
		else if (bits_per_sample == 8 && samples_per_pixel == 1
			 && photometric < 2)
		    type = GG_PIXEL_GRAYSCALE;
		else if (bits_per_sample == 8 && samples_per_pixel == 3)
		    type = GG_PIXEL_RGB;
		else
		    type = GG_PIXEL_UNKNOWN;
	    }
	  else if (samples_per_pixel == 1)
	      type = GG_PIXEL_GRID;
	  else
	      type = GG_PIXEL_UNKNOWN;
      }
    else
	type = GG_PIXEL_UNKNOWN;
    switch (sample_format)
      {
      case SAMPLEFORMAT_UINT:
	  gg_sample_format = GGRAPH_SAMPLE_UINT;
	  break;
      case SAMPLEFORMAT_INT:
	  gg_sample_format = GGRAPH_SAMPLE_INT;
	  break;
      case SAMPLEFORMAT_IEEEFP:
	  gg_sample_format = GGRAPH_SAMPLE_FLOAT;
	  break;
      default:
	  gg_sample_format = GGRAPH_SAMPLE_UNKNOWN;
	  break;
      };

/* computing the corners coords */
    cx = 0.0;
    cy = 0.0;
    GTIFImageToPCS (gtif, &cx, &cy);
    upper_left_x = cx;
    upper_left_y = cy;
    cx = 0.0;
    cy = height;
    GTIFImageToPCS (gtif, &cx, &cy);
    lower_left_x = cx;
    lower_left_y = cy;
    cx = width;
    cy = 0.0;
    GTIFImageToPCS (gtif, &cx, &cy);
    upper_right_x = cx;
    upper_right_y = cy;
    cx = width;
    cy = height;
    GTIFImageToPCS (gtif, &cx, &cy);
    lower_right_x = cx;
    lower_right_y = cy;
/* computing the pixel size */
    pixel_x = (upper_right_x - upper_left_x) / (double) width;
    pixel_y = (upper_left_y - lower_left_y) / (double) height;
    if (type == GG_PIXEL_UNKNOWN)
      {
	  ret = GGRAPH_UNSUPPORTED_TIFF_LAYOUT;
	  goto error;
      }

    img =
	gg_image_create (type, width, height, bits_per_sample,
			 samples_per_pixel, gg_sample_format, srs_name,
			 proj4text);
    if (!img)
      {
	  ret = GGRAPH_INSUFFICIENT_MEMORY;
	  goto error;
      }
    if (is_tiled)
      {
	  img->tile_width = tile_width;
	  img->tile_height = tile_height;
      }
    else
	img->rows_per_strip = rows_strip;
    img->is_georeferenced = 1;
    img->srid = epsg;
    img->upper_left_x = upper_left_x;
    img->upper_left_y = upper_left_y;
    img->pixel_x_size = pixel_x;
    img->pixel_y_size = pixel_y;

    ret =
	common_read_from_tiff (in, img, width, height, is_tiled, type,
			       gg_sample_format, bits_per_sample, tile_width,
			       tile_height, rows_strip);
    if (ret != GGRAPH_OK)
	goto error;

    TIFFClose (in);
    GTIFFree (gtif);
    *image_handle = img;
    return GGRAPH_OK;

  error:
    XTIFFClose (in);
    if (gtif)
	GTIFFree (gtif);
    if (img)
	gGraphDestroyImage (img);
    return ret;
}
